/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * Dave Clemans, April 2003
 *
 * Copyright (C) 2003  Dave Clemans, Brian A. Lantz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * See the file "COPYING" for more details.
 */

#include "globals.h"
#include <string.h>


/* comment the line below to disable the comments in the logfile (makes smaller executable, too) */
#define DOCUMENTED_LOGFILE

static void mainAdmin (struct aMessage* msg);
static void submitMainAdmin (struct aMessage* msg, char *params);
static char *getParams (FILE *fp);
static void addPathAdmin (struct aMessage* msg);
static void addMethodAdmin (struct aMessage* msg);
static void submitAddPathAdmin (struct aMessage* msg, char *params);
static void saveCurrentConfig (void);

extern void _htmlFooter (struct aMessage *msg);
extern void htmlAdminHeader (struct aMessage* msg, char *url);


#ifdef DOCUMENTED_LOGFILE
static char * str_root = 	"#\n"
				"# The root directory... effectively the files we export\n"
				"#  Format: <keyword>: <TiVo Visible Name> <Disk pathname>\n"
				"#\n"
				"# Keywords:\n"
				"#  image:   \"publish\" a directory of images (JPEG files)\n"
				"#  audio:   \"publish\" a directory of music (MP3 files)\n"
				"#  virtual: \"publish\" a directory of music (MP3 files), organized by genre,\n"
				"#     artist & album information.  This depends on the ID3 tag information\n"
				"#     in the MP3 file being correct.\n"
				"#\n"
				"# You don't have to have both an \"audio\" and a \"virtual\" directory.  If the\n"
				"# tags are correct, a \"virtual\" directory should be sufficient.\n"
				"#\n"
				"# byRequest accesses mp3 tags using an in-memory database.  That database\n"
				"# is built automatically on the first access to the file.  If you have a very\n"
				"# large directory you might want to start byRequest with the \"-preload\" switch.\n"
				"# With the \"-preload\" switch, byRequest spends a few minutes indexing files\n"
				"# before it starts talking to the TiVo.  Pre-indexing files minimizes wait\n"
				"# times on later accesses.\n"
				"#\n";

#ifdef PHOTO_ORGANIZER
static char *str_photo =	"\n#\n"
				"# Optional service... a photo organizer similar to the music \"virtual\"\n"
				"# directory.  Depends on cameras that add EXIF tags.\n"
				"#\n"
				"# photo_organizer categorizes photos into year and month folders\n"
				"# based upon EXIF tags.  Requires libexif.so.  To build:\n"
				"#    $make PHOTO_ORGANIZER=y\n"
				"#\n";
#endif /* PHOTO_ORGANIZER */

static char *str_recorder =	"\n#\n"
				"# Optional service... a system where you can record internet radio audio streams\n"
				"#     on a specific date\n"
				"#     every day\n"
				"#     every weekday\n"
				"#     every weekend day\n"
				"#     a specific day of the week.\n"
				"#\n"
				"# Format:\n"
				"#  recorder: <TiVo Visible Name> <Disk directory> <schedule file>\n"
				"#              <Max directory size in gigs>\n"
				"#\n"
				"# If the directory gets bigger than its maximum size, it's shrunk by repeatedly\n"
				"# deleting the oldest file until the directory is small enough.\n"
				"# See the example \"byRequest.sched\" file for the format of the scheduling file.\n"
				"#\n";

#ifdef XMRADIO
static char *str_xm = 		"\n#\n"
				"# Optional service... Support for the XM PCR Satellite Radio.\n"
				"# Format:\n"
				"#     xmpcr: <USB serial port controling the radio> <DSP file to digitize audio>\n"
				"#\n"
				"# Note that the XM PCR is turned on only when needed.  If it's idle for\n"
				"# twenty minutes, it's turned off again.  There will be a few second delay\n"
				"# if you select a XM radio station while the receiver is turned off; e.g.\n"
				"# while it's turned back on again.\n"
				"#\n"
				"# Note: get the radio working first using the XM software, the OpenXM\n"
				"# software, etc. before trying to use the reciever with byRequest.  Fixing\n"
				"# any problems is much simpler that way.  Also, it's good to check that you\n"
				"# can read, and you do get data, from the sound dsp device.\n"
				"#\n"
				"# byRequest can be built to either drive the radio directly using the\n"
				"# USB serial port, or it can use the OpenXM network daemon protocol.\n"
				"# Look at the USE_XMDAEMON ifdef.\n"
				"#\n"
				"# byRequest will use the \"ossdsp\" and \"mp3high\" transcoding definitions\n"
				"# below to send the audio data to the TiVo.  You might have to experiment\n"
				"# with mixer levels on your computer, and with volume levels in the \"ossdsp\"\n"
				"# command (the \"-v 8\" switch to sox below) to get decent sound.\n"
				"# The command below seems to be mostly decent, as long as you don't have the\n"
				"# mixer volume level too high...  I use an external SB Extigy usb sound card\n"
				"# for digitizing the XM PCR audio.\n"
				"#\n"
				"# If you plug the XM PCR into the same sound card as your regular speakers,\n"
				"# you might want to mute the line in; otherwise you'll hear the XM PCR\n"
				"# through the speakers whenever it's on.  (Or you can use another sound\n"
				"# card; even an external USB sound card; that isn't connected to speakers).\n"
				"#\n";
#endif /*XMRADIO */

static char *str_cache = 	"\n#\n"
				"# Cache size for internet streams\n"
				"#  the bigger it is, the less likely you'll hear skips\n"
				"#  but larger caches will start up slower\n"
				"#\n";

static char *str_thumb =	"\n#\n"
				"# Cache directory for thumbnail images\n"
				"# NOTE: contents are deleted at startup\n"
				"#\n";

static char *str_logfile =	"\n#\n"
				"# Where to write log information.  The log is very important if you suspect a\n"
				"# problem.  If you report a problem to the mailing list\n"
				"# byrequest-devel@lists.sourceforge.net, you should include a portion of the\n"
				"# log file from a few minutes before the problem, to when the problem occurred.\n"
				"#\n";

static char *str_divider = 	"\n#\n"
				"# It's less likely that you'll have to modify anything after this point...\n"
				"#\n";

static char *str_port =		"\n#\n"
				"# Set a specific port to receive command requests on.  Optional; defaults\n"
				"# to the same port used for TiVo beacon broadcasts.\n"
				"#\n";

static char *str_network =	"\n#\n"
				"# Set a specific network to broadcast beacons on.  Optional; should only be\n"
				"# needed in systems with multiple network interfaces.\n"
				"#\n";

static char *str_filetypes =	"\n#\n"
				"# Recognized file suffixes\n"
				"#  mpeg type:     mp3 files\n"
				"#  jpeg type:     jpg files; jpeg, or a file that can be converted\n"
				"#  stream type:   audio files that aren't mp3 local files; could be\n"
				"#                 ogg-vorbis format, mpeg internet streams, or files or streams\n"
				"#                 supported by mplayer\n"
				"#  playlist:      audio playlists\n"
				"#  script:        script or application which dumps an image to stdout\n"
				"#  internet:      recognized internal protocols for streaming audio\n"
				"#\n";

static char *str_transcode =	"\n#\n"
				"# Define transcoding methods...\n"
				"#     Transcoding allows us to send streams of different types to the TiVo,\n"
				"#     which normally only accepts mp3 streams.  By transcoding, we mean\n"
				"#     that the source format ends up being converted to mp3 format.\n"
				"#\n"
				"# A normal transcoding method converts its inputs to \"wav\" format.\n"
				"# In a normal transcoding method line, the \"%%s\" argument is the name of\n"
				"# the input file or input url.\n"
				"#     %%in   - the input file or url\n"
				"#     %%out  - the output file (typically a fifo)\n"
				"#\n"
				"# There are two special \"transcoding\" methods... \"mp3high\" is a high-bit\n"
				"# rate mp3 encoder, for file or xmpcr sources.  \"mp3low\" is a low-bit\n"
				"# rate mp3 encoder, for internet radio streams.  The mp3 encoders should\n"
				"# expect a \"wav\" format input, and produce a \"mp3\" output.\n"
				"#\n";

#ifdef IGNORE_FILE
static char *str_ignore =	"\n#\n"
				"# A string defining a regular expression used to ignore files.\n"
				"#\n";
#endif /* IGNORE_FILE */

#endif /* DOCUMENTED_LOGFILE */



void adminInterface (int fd, FILE *fp, char *cp)
{
char *cp2;
struct aMessage*        msg;
int isCGI;
char *params = NULL;

	for (cp2 = cp; *cp2; cp2++) {
		if (*cp2 == ' ') {
			*cp2 = 0;
			break;
		}
	}
	addToLog( "Rcvd Admin Command:\n%s\n", cp );

	cp2 -= 4;
	isCGI = (strcmp (cp2, ".cgi") == 0);

	msg = newMessage();
	if (isCGI == 0)	{		/* this is a form page */
		/* if "admin", "admin/" or "admin/index.*" */		
		if (!strcmp (cp, "admin") || !strcmp (cp, "admin/") || !strncmp (cp, "admin/index.", 12))
			mainAdmin (msg);
		else	{
			write( fd, "404 file not found\r\n", 20 );
			msg = NULL;
		}
	} else {		/* this is a form being posted */
		params = getParams (fp);
		if (!strcmp (cp, "admin/submit.cgi")) {
			if (strstr (params, "addPathButton"))
				addPathAdmin (msg);
			else if (strstr (params, "addMethodButton"))
				addMethodAdmin (msg);
			else
				submitMainAdmin (msg, params);
		} else if (!strcmp (cp, "admin/add.cgi")) {
			submitAddPathAdmin (msg, params);
		} else	{
			write( fd, "404 file not found\r\n", 20 );
			msg = NULL;
		}
	}

	if (msg != NULL)	{
		htmlSend (fd, msg);
		freeMessage( msg );
	}
	if (params != NULL)
		free (params);
	close (fd);
}


static char *getParams (FILE *fp)
{
char buf[1024 * 4];
char *retval = NULL;
char *params = NULL;
int length;

	while (1)	{
		fgets (buf, 1024, fp);
		if (buf[0] == '\n' || buf[0] == '\r')
			break;
		if (!strncasecmp ("Content-Length:", buf, 15)) {
			length = atoi (&buf[15]);
		}
		addToLog ("recv: %s", buf);
	}
	params = (char *) calloc (1, length + 1);
	fread (params, length, 1, fp);
	retval = decodeString (params);
	free (params);
	for (params = retval; *params; params++)	{
		if (*params == '+')	{
			*params = ' ';
		}
	}
	addToLog ("params: %s\n", retval);
	return (retval);
}


static void addPathAdmin (struct aMessage* msg)
{
	htmlAdminHeader (msg, "/admin/add.cgi");

	appendToMsg (msg, "  <tr>\n    <td width=\"30%\" align=\"right\"><select size=\"1\" name=\"pathtype\">\n");
	appendToMsg (msg, "\t<option selected>Select Path Type</option>\n");
	appendToMsg (msg, "\t<option value=\"audio\">Audio Path</option>\n");
	appendToMsg (msg, "\t<option value=\"image\">Image Path</option>\n");
	appendToMsg (msg, "\t<option value=\"virtual\">Virtual Path</option>\n");
#ifdef PHOTO_ORGANIZER
	appendToMsg (msg, "\t<option value=\"photo_organizer\">Photo Organizer Path</option>\n");
#endif
	appendToMsg (msg, "</td>\n");
	appendToMsg (msg, "  <td width=\"35%\" align=\"left\">Name: \n");
	appendToMsg (msg, "    <input type=\"text\" name=\"name\" size=\"30\"></td>\n");
	appendToMsg (msg, "  <td width=\"35%\" align=\"left\">Path(s): \n");
	appendToMsg (msg, "    <input type=\"text\" name=\"path\" size=\"30\"></td>\n");
	appendToMsg (msg, "  </tr>\n");

	appendToMsg (msg, "</table>\n");
	appendToMsg (msg, "<p align=\"center\"><input type=\"submit\" value=\"Submit\" name=\"submitButton\">&nbsp;&nbsp;");
	appendToMsg (msg, "<input type=\"reset\" value=\"Reset\" name=\"reset\"></p>\n");
	appendToMsg (msg, "</form>\n\n");
	_htmlFooter (msg);
}

static void addMethodAdmin (struct aMessage* msg)
{
	htmlAdminHeader (msg, "/admin/add.cgi");

	appendToMsg (msg, "  <tr>\n<td width=\"70%\" align=\"left\">Command Line: \n");
	appendToMsg (msg, "    <input type=\"text\" name=\"name\" size=\"50\"></td>\n");
	appendToMsg (msg, "  <td width=\"30%\" align=\"left\">Filetype: \n");
	appendToMsg (msg, "    <input type=\"text\" name=\"path\" size=\"10\"></td>\n");
	appendToMsg (msg, "  </tr>\n");

	appendToMsg (msg, "</table>\n");
	appendToMsg (msg, "<input type=\"hidden\" name=\"pathtype\" value=\"transcode\">");
	appendToMsg (msg, "<p align=\"center\"><input type=\"submit\" value=\"Submit\" name=\"submitButton\">&nbsp;&nbsp;");
	appendToMsg (msg, "<input type=\"reset\" value=\"Reset\" name=\"reset\"></p>\n");
	appendToMsg (msg, "</form>\n\n");
	_htmlFooter (msg);
}

static void mainAdmin (struct aMessage* msg)
{
struct aDirectoryEntry* item;
char buf[1024];
int i;
char *tmp;
struct aTranscodingMethod *tm;
struct aRecorder *rec;

	htmlAdminHeader (msg, "/admin/submit.cgi");
	appendToMsg (msg, "  <tr>\n    <td width=\"30%\" align=\"right\">Port:</td>\n");
	appendToMsg (msg, "    <td width=\"20%\"><input type=\"text\" name=\"port\" size=\"7\" value=\"%d\"></td>\n", getDataPort());
	tmp = getLogfile ();
	if (tmp == NULL)
		tmp = "";
	appendToMsg (msg, "    <td width=\"20%\" align=\"right\">Log File:</td>\n");
	appendToMsg (msg, "    <td width=\"30%\"><input type=\"text\" name=\"logfile\" size=\"30\" value=\"%s\"></td>\n  </tr>\n", tmp);

	appendToMsg (msg, "  <tr>\n    <td width=\"30%\" align=\"right\">Cache Size:</td>\n");
	appendToMsg (msg, "    <td width=\"20%\"><input type=\"text\" name=\"cache\" size=\"7\" value=\"%s\"></td>\n", getCacheSize());
	tmp = getThumbCacheDir();
	if (tmp == NULL)
		tmp = "";
	appendToMsg (msg, "    <td width=\"20%\" align=\"right\">Thumbs Directory:</td>\n");
	appendToMsg (msg, "    <td width=\"30%\"><input type=\"text\" name=\"thumbdir\" size=\"30\" value=\"%s\"></td>\n  </tr>\n", tmp);

	buf[0] = 0;
	for( i = 0; getMpegEntry( i ) != NULL; i++ )
	{
		if (i > 0)
			strcat (buf, " ");
		strcat (buf, getMpegEntry(i));
	}
	appendToMsg (msg, "  <tr>\n    <td width=\"30%\" align=\"right\">Mpeg File Extensions:</td>\n");
	appendToMsg (msg, "    <td width=\"20%\"><input type=\"text\" name=\"mpeg\" size=\"20\" value=\"%s\"></td>\n", buf);

	buf[0] = 0;
	for( i = 0; getJpegEntry( i ) != NULL; i++ )
	{
		if (i > 0)
			strcat (buf, " ");
		strcat (buf, getJpegEntry(i));
	}
	appendToMsg (msg, "    <td width=\"20%\" align=\"right\">Jpeg File Extensions:</td>\n");
	appendToMsg (msg, "    <td width=\"30%\"><input type=\"text\" name=\"jpeg\" size=\"20\" value=\"%s\"></td>\n  </tr>\n", buf);

	buf[0] = 0;
	for( i = 0; getStreamEntry( i ) != NULL; i++ )
	{
		if (i > 0)
			strcat (buf, " ");
		strcat (buf, getStreamEntry(i));
	}
	appendToMsg (msg, "  <tr>\n    <td width=\"30%\" align=\"right\">Stream File Extensions:</td>\n");
	appendToMsg (msg, "    <td width=\"20%\"><input type=\"text\" name=\"stream\" size=\"20\" value=\"%s\"></td>\n", buf);

	buf[0] = 0;
	for( i = 0; getPlayEntry( i ) != NULL; i++ )
	{
		if (i > 0)
			strcat (buf, " ");
		strcat (buf, getPlayEntry(i));
	}
	appendToMsg (msg, "    <td width=\"20%\" align=\"right\">Playlist File Extensions:</td>\n");
	appendToMsg (msg, "    <td width=\"30%\"><input type=\"text\" name=\"playlist\" size=\"20\" value=\"%s\"></td>\n  </tr>\n", buf);

	buf[0] = 0;
	for( i = 0; getScriptEntry( i ) != NULL; i++ )
	{
		if (i > 0)
			strcat (buf, " ");
		strcat (buf, getScriptEntry(i));
	}
	appendToMsg (msg, "  <tr>\n    <td width=\"30%\" align=\"right\">Script File Extensions:</td>\n");
	appendToMsg (msg, "    <td width=\"20%\"><input type=\"text\" name=\"script\" size=\"20\" value=\"%s\"></td>\n", buf);

	buf[0] = 0;
	for( i = 0; getInternetEntry( i ) != NULL; i++ )
	{
		if (i > 0)
			strcat (buf, " ");
		strcat (buf, getInternetEntry(i));
	}
	appendToMsg (msg, "    <td width=\"20%\" align=\"right\">Internet File Extensions:</td>\n");
	appendToMsg (msg, "    <td width=\"30%\"><input type=\"text\" name=\"internet\" size=\"20\" value=\"%s\"></td>\n  </tr>\n", buf);

	tmp = getIFName ();
	appendToMsg (msg, "  <tr>\n    <td width=\"30%\" align=\"right\">Network Interface:</td>\n");
	appendToMsg (msg, "    <td width=\"20%\"><input type=\"text\" name=\"network\" size=\"20\" value=\"%s\"></td>\n", tmp);

#ifdef IGNORE_FILE
	tmp = getIgnoreString ();
	if (tmp == NULL)
		tmp = "";
	appendToMsg (msg, "    <td width=\"20%\" align=\"right\">Ignore String:</td>\n");
	appendToMsg (msg, "    <td width=\"30%\"><input type=\"text\" name=\"ignore\" size=\"30\" value=\"%s\"></td>\n", tmp);
#endif /* IGNORE_FILE */
	appendToMsg (msg, "  </tr>\n");

	appendToMsg (msg, "</table>\n<hr><table width=\"100%\">\n");

#ifdef XMRADIO
	appendToMsg (msg, "  <tr>\n    <td width=\"20%\" align=\"right\"><i><b>XMRadio:</b></i></td>\n");
	appendToMsg (msg, "    <td>&nbsp;</td>\n    <td>&nbsp;</td>\n  </tr>\n");
	appendToMsg (msg, "  <tr>\n    <td width=\"20%\" align=\"right\">&nbsp;</td>\n");
	tmp = getXMRadioControl ();
	if (tmp == NULL)
		tmp = "";
	appendToMsg (msg, "    <td width=\"30%\" align=\"left\">USB Port: \n");
	appendToMsg (msg, "      <input type=\"text\" name=\"xmradio_port\" size=\"30\" value=\"%s\"></td>\n", tmp);
	tmp = getXMRadio ();
	if (tmp == NULL)
		tmp = "";
	appendToMsg (msg, "    <td width=\"50%\" align=\"left\">DSP File: \n");
	appendToMsg (msg, "      <input type=\"text\" name=\"xmradio_dsp\" size=\"30\" value=\"%s\"></td>\n  </tr>\n", tmp);
#endif /* XMRADIO */

	appendToMsg (msg, "  <tr>\n    <td width=\"20%\" align=\"right\"><i><b>Audio Path(s):</b></i></td>\n");
	appendToMsg (msg, "    <td>&nbsp;</td>\n    <td>&nbsp;</td>\n  </tr>\n");
      	for( item = getRootdir(), i = 1; item; item = item->next)
	{
        	if( item->folderType == FILE_MPEG) {
			appendToMsg (msg, "  <tr>\n    <td width=\"20%\" align=\"right\">&nbsp;</td>\n");
			appendToMsg (msg, "    <td width=\"30%\" align=\"left\">Name: \n");
			appendToMsg (msg, "      <input type=\"text\" name=\"audio%dname\" size=\"30\" value=\"%s\"></td>\n", i, item->description);
			appendToMsg (msg, "    <td width=\"50%\" align=\"left\">Path(s): \n");
			appendToMsg (msg, "      <input type=\"text\" name=\"audio%dpath\" size=\"30\" value=\"%s\">", i, item->name);
			appendToMsg (msg, "&nbsp;<input type=\"submit\" value=\"Delete Path\" name=\"audio%ddelete\"></td>\n", i);
			appendToMsg (msg, "  </tr>\n");
			i++;
		}
        }


	appendToMsg (msg, "  <tr>\n    <td width=\"20%\" align=\"right\"><i><b>Image Path(s):</b></i></td>\n");
	appendToMsg (msg, "    <td>&nbsp;</td>\n    <td>&nbsp;</td>\n  </tr>\n");
      	for( item = getRootdir(), i = 1; item; item = item->next)
	{
        	if( item->folderType == FILE_JPEG) {
			appendToMsg (msg, "  <tr>\n    <td width=\"20%\" align=\"right\">&nbsp;</td>\n");
			appendToMsg (msg, "    <td width=\"30%\" align=\"left\">Name: \n");
			appendToMsg (msg, "      <input type=\"text\" name=\"image%dname\" size=\"30\" value=\"%s\"></td>\n", i, item->description);
			appendToMsg (msg, "    <td width=\"50%\" align=\"left\">Path(s): \n");
			appendToMsg (msg, "      <input type=\"text\" name=\"image%dpath\" size=\"30\" value=\"%s\">", i, item->name);
			appendToMsg (msg, "&nbsp;<input type=\"submit\" value=\"Delete Path\" name=\"image%ddelete\"></td>\n", i);
			appendToMsg (msg, "  </tr>\n");
			i++;
		}
	}

	tmp = getVirtualDirName();
	if (strncmp ("*No ", tmp, 4) != 0) {
		buf[0] = 0;
		for( item = getVirtualDirectories(); item; item = item->next )
		{
			if (buf[0] > 0)
				strcat (buf, " ");
			strcat (buf, item->name);
		}
		appendToMsg (msg, "  <tr>\n    <td width=\"20%\" align=\"right\"><i><b>Virtual Path(s):</b></i></td>\n");
		appendToMsg (msg, "    <td>&nbsp;</td>\n    <td>&nbsp;</td>\n  </tr>\n");
		appendToMsg (msg, "  <tr>\n    <td width=\"20%\" align=\"right\">&nbsp;</td>\n");
		appendToMsg (msg, "    <td width=\"30%\" align=\"left\">Name: \n");
		appendToMsg (msg, "      <input type=\"text\" name=\"virtual1name\" size=\"30\" value=\"%s\"></td>\n", tmp);
		appendToMsg (msg, "    <td width=\"50%\" align=\"left\">Path(s): \n");
		appendToMsg (msg, "      <input type=\"text\" name=\"virtual1path\" size=\"30\" value=\"%s\">", buf);
		appendToMsg (msg, "&nbsp;<input type=\"submit\" value=\"Delete Path\" name=\"virtual1delete\"></td>\n");
		appendToMsg (msg, "  </tr>\n");
	}

#ifdef PHOTO_ORGANIZER
	tmp = getPhotoOrganizerDirName();
	if (tmp != NULL) {
		buf[0] = 0;
		for( item = getPhotoOrganizerDirectories(); item; item = item->next )
		{
			if (buf[0] > 0)
				strcat (buf, " ");
			strcat (buf, item->name);
		}
		appendToMsg (msg, "  <tr>\n    <td width=\"20%\" align=\"right\"><i><b>Photo Organizer Path:</b></i></td>\n");
		appendToMsg (msg, "    <td>&nbsp;</td>\n    <td>&nbsp;</td>\n  </tr>\n");
		appendToMsg (msg, "  <tr>\n    <td width=\"20%\" align=\"right\">&nbsp;</td>\n");
		appendToMsg (msg, "    <td width=\"30%\" align=\"left\">Name: \n");
		appendToMsg (msg, "      <input type=\"text\" name=\"photo_organizer1name\" size=\"30\" value=\"%s\"></td>\n", tmp);
		appendToMsg (msg, "    <td width=\"50%\" align=\"left\">Path(s): \n");
		appendToMsg (msg, "      <input type=\"text\" name=\"photo_organizer1path\" size=\"30\" value=\"%s\">", buf);
		appendToMsg (msg, "&nbsp;<input type=\"submit\" value=\"Delete Path\" name=\"photo_organizer1delete\"></td>\n");
		appendToMsg (msg, "  </tr>\n");
	}
#endif

	appendToMsg (msg, "</table>\n<hr><table width=\"100%\">\n");
	appendToMsg (msg, "  <tr>\n    <td width=\"20%\" align=\"right\"><i><b>Recorder:</b></i></td>\n");
	appendToMsg (msg, "    <td>&nbsp;</td>\n    <td>&nbsp;</td>\n  </tr>\n");

	rec = getCurrentRecorder ();
	appendToMsg (msg, "  <tr>\n    <td width=\"20%\" align=\"right\">&nbsp;</td>\n");
	appendToMsg (msg, "    <td align=\"right\">Name: </td>\n");
	appendToMsg (msg, "    <td align=\"left\"><input type=\"text\" name=\"recordername\" size=\"40\" value=\"%s\"></td>\n", (rec) ? rec->name : "");
	appendToMsg (msg, "    <td align=\"right\">Directory: </td>\n");
	appendToMsg (msg, "    <td align=\"left\"><input type=\"text\" name=\"recorderdir\" size=\"30\" value=\"%s\"></td>\n  </tr>\n", (rec) ? rec->directory : "");

	appendToMsg (msg, "  <tr>\n    <td width=\"20%\" align=\"right\">&nbsp;</td>\n");
	appendToMsg (msg, "    <td align=\"right\">Schedule File: \n</td>");
	appendToMsg (msg, "    <td align-\"left\"><input type=\"text\" name=\"recorderschedule\" size=\"40\" value=\"%s\"></td>\n", (rec) ? rec->filename : "");
	appendToMsg (msg, "    <td align=\"right\">Size (in gigs): </td>\n");
	appendToMsg (msg, "    <td align=\"left\"><input type=\"text\" name=\"recordersize\" size=\"10\" value=\"%s\"></td>\n  </tr>\n", (rec) ? rec->containerSize : "");

	appendToMsg (msg, "</table>\n<hr><table width=\"100%\">\n");

	appendToMsg (msg, "  <tr>\n    <td width=\"20%\" align=\"right\"><i><b>Transcoding Methods:</b></i></td>\n");
	appendToMsg (msg, "    <td>&nbsp;</td>\n    <td>&nbsp;</td>\n  </tr>\n");

	for (tm = getTranscodingMethods(), i = 1; tm; tm = tm->next)
	{
		appendToMsg (msg, "  <tr>\n    <td width=\"20%\" align=\"right\">&nbsp;</td>\n");
		appendToMsg (msg, "    <td width=\"40%\" align=\"left\">Command: \n");
		appendToMsg (msg, "      <input type=\"text\" name=\"transcode%dname\" size=\"40\" value=\"%s\"></td>\n", i, tm->cmdLine);
		appendToMsg (msg, "    <td width=\"40%\" align=\"left\">Filetype: \n");
		appendToMsg (msg, "      <input type=\"text\" name=\"transcode%dpath\" size=\"10\" value=\"%s\">", i, tm->fileType);
		appendToMsg (msg, "&nbsp;<input type=\"submit\" value=\"Delete Transcode\" name=\"transcode%ddelete\"></td>\n", i);
		appendToMsg (msg, "  </tr>\n");
		i++;
	}

	appendToMsg (msg, "</table>\n");
	appendToMsg (msg, "<p align=\"center\"><input type=\"submit\" value=\"Submit\" name=\"submitButton\">&nbsp;&nbsp;");
	appendToMsg (msg, "<input type=\"submit\" value=\"Add New Path\" name=\"addPathButton\">&nbsp;&nbsp;");
	appendToMsg (msg, "<input type=\"submit\" value=\"Add New Transcode Method\" name=\"addMethodButton\">&nbsp;&nbsp;");
	appendToMsg (msg, "<input type=\"reset\" value=\"Reset\" name=\"reset\"></p>\n");
	appendToMsg (msg, "</form>\n\n");

	_htmlFooter (msg);

}


static void addSimpleParam (FILE *fp, char *params, char *find)
{
char *cp, *tmp;
char buf[64];

	sprintf (buf, "%s=", find);
	cp = strstr (params, buf);
	tmp = strchr (cp, '&');
	*tmp = 0;
	if (cp[strlen (buf)] == 0)
		fprintf (fp, "#%s:\n", find);
	else
		fprintf (fp, "%s: %s\n", find, &cp[strlen (buf)]);
	*tmp = '&';
}

static int addPathParam (FILE *fp, char *params, char *find, int counter)
{
char *cp, *tmp;
char buf[64];
char *desc;
char *path;

	sprintf (buf, "%s%ddelete=", find, counter);
	cp = strstr (params, buf);
	if (cp != NULL)
		return (1);

	sprintf (buf, "%s%dname=", find, counter);
	cp = strstr (params, buf);
	if (cp == NULL)
		return (0);

	tmp = strchr (cp, '&');
	*tmp = 0;
	desc = strdup (&cp[strlen (buf)]);
	*tmp = '&';

	sprintf (buf, "%s%dpath=", find, counter);
	cp = strstr (params, buf);

	tmp = strchr (cp, '&');
	if (tmp != NULL)
		*tmp = 0;
	path = strdup (&cp[strlen (buf)]);
	if (tmp != NULL)
		*tmp = '&';

	fprintf (fp, "%s: \"%s\" %s\n", find, desc, path);
	free (desc);
	free (path);
	return (1);
}


static void submitAddPathAdmin (struct aMessage* msg, char *params)
{
char *cp, *tmp;
char *desc;
char *path;
FILE *fp;

	// add a line to the config file
	fp = fopen (getConfigFile(), "a+");

	cp = strstr (params, "name=");
	tmp = strchr (cp, '&');
	*tmp = 0;
	desc = strdup (&cp[5]);
	*tmp = '&';

	cp = strstr (params, "path=");
	tmp = strchr (cp, '&');
	*tmp = 0;
	path = strdup (&cp[5]);
	*tmp = '&';

	cp = strstr (params, "pathtype=");
	tmp = strchr (cp, '&');
	*tmp = 0;

	fprintf (fp, "#\n# Recent addition\n#\n%s: \"%s\" %s\n", &cp[9], desc, path);
	free (desc);
	free (path);

	fclose (fp);
        readConfigFile( getConfigFile() );
	saveCurrentConfig ();
	mainAdmin (msg);
}


static void submitMainAdmin (struct aMessage* msg, char *params)
{
FILE *fp;
int more;
int counter;
char *tmp, *cp;
char *cp1, *cp2, *cp3, *cp4;

	fp = fopen (getConfigFile(), "w");

	fprintf (fp, "#\n# Configuration file for byRequest - 'Yet Another TiVo Server'\n");
	fprintf (fp, "# Automatically generated by admin interface\n#\n\n");

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_root);
#endif
	counter = 1;
	do {
		more = addPathParam (fp, params, "image", counter++);
	} while (more);

	counter = 1;
	do {
		more = addPathParam (fp, params, "audio", counter++);
	} while (more);

	counter = 1;
	do {
		more = addPathParam (fp, params, "virtual", counter++);
	} while (more);

#ifdef PHOTO_ORGANIZER
#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_photo);
#endif
	counter = 1;
	more = addPathParam (fp, params, "photo_organizer", counter);
	if (more == 0)
		fprintf (fp, "#photo_organizer: \"Photos\"    /home/dgc/photos\n");
#endif /* PHOTO_ORGANIZER */


#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_recorder);
#endif
	cp = strstr (params, "recordername=");
	tmp = strchr (cp, '&');
	*tmp = 0;
	cp1 = strdup (&cp[13]);
	*tmp = '&';
	cp = strstr (params, "recorderdir=");
	tmp = strchr (cp, '&');
	*tmp = 0;
	cp2 = strdup (&cp[12]);
	*tmp = '&';
	cp = strstr (params, "recorderschedule=");
	tmp = strchr (cp, '&');
	*tmp = 0;
	cp3 = strdup (&cp[17]);
	*tmp = '&';
	cp = strstr (params, "recordersize=");
	tmp = strchr (cp, '&');
	*tmp = 0;
	cp4 = strdup (&cp[13]);
	*tmp = '&';
	if (strlen (cp1) > 0 && strlen (cp2) > 0 && strlen (cp3) > 0 && strlen (cp4) > 0)
		fprintf (fp, "recorder: \"%s\" %s %s %s\n", cp1, cp2, cp3, cp4);
	else
		fprintf (fp, "#recorder: \"By Request...\" /d1/Recordings /mm/byRequest/byRequest.sched 10\n");
	free (cp1);
	free (cp2);
	free (cp3);
	free (cp4);

#ifdef XMRADIO
#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_xm);
#endif
	cp = strstr (params, "xmradio_port=");
	tmp = strchr (cp, '&');
	*tmp = 0;
	cp1 = strdup (&cp[13]);
	*tmp = '&';
	cp = strstr (params, "xmradio_dsp=");
	tmp = strchr (cp, '&');
	*tmp = 0;
	cp2 = strdup (&cp[12]);
	*tmp = '&';
	if (strlen (cp1) > 0 && strlen (cp2) > 0)
		fprintf (fp, "xmpcr: %s %s\n", cp1, cp2);
	else
		fprintf (fp, "#xmpcr: /dev/ttyUSB0 /dev/dsp1\n");
	free (cp1);
	free (cp2);
#endif /* XMRADIO */

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_cache);
#endif
	addSimpleParam (fp, params, "cache");

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_thumb);
#endif
	addSimpleParam (fp, params, "thumbdir");

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_logfile);
#endif
	addSimpleParam (fp, params, "logfile");

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_divider);

	fprintf (fp, str_port);
#endif
	addSimpleParam (fp, params, "port");

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_network);
#endif
	addSimpleParam (fp, params, "network");

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_filetypes);
#endif
	addSimpleParam (fp, params, "mpeg");
	addSimpleParam (fp, params, "jpeg");
	addSimpleParam (fp, params, "stream");
	addSimpleParam (fp, params, "playlist");
	addSimpleParam (fp, params, "script");
	addSimpleParam (fp, params, "internet");

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_ignore);
#endif
	addSimpleParam (fp, params, "ignore");

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_transcode);
#endif
	counter = 1;
	do {
		more = addPathParam (fp, params, "transcode", counter++);
	} while (more);

	fclose (fp);

        readConfigFile( getConfigFile() );

	mainAdmin (msg);
}

static void addValue (FILE *fp, char * label, char * value)
{
	if (value == NULL || value[0] == 0)
		fprintf (fp, "#%s:\n", label);
	else
		fprintf (fp, "%s: %s\n", label, value);
}


static void saveCurrentConfig ()
{
FILE *fp;
int more;
char buf[1024];
int i;
char *desc;
struct aDirectoryEntry* item;

	fp = fopen (getConfigFile(), "w");

	fprintf (fp, "#\n# Configuration file for byRequest - 'Yet Another TiVo Server'\n");
	fprintf (fp, "# Automatically generated by admin interface\n#\n\n");

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_root);
#endif
      	for( item = getRootdir(); item; item = item->next )
	{
        	if( item->folderType == FILE_JPEG)
			fprintf (fp, "image: \"%s\" %s\n", item->description, item->name);
	}
      	for( item = getRootdir(); item; item = item->next )
	{
        	if( item->folderType == FILE_MPEG)
			fprintf (fp, "audio: \"%s\" %s\n", item->description, item->name);
	}
	desc = getVirtualDirName();
	if (strncmp ("*No ", desc, 4) != 0) {
		buf[0] = 0;
		for( item = getVirtualDirectories(); item; item = item->next )
		{
			if (buf[0] > 0)
				strcat (buf, " ");
			strcat (buf, item->name);
		}
		fprintf (fp, "virtual: \"%s\" %s\n", desc, buf);
	}

#ifdef PHOTO_ORGANIZER
#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_photo);
#endif
	desc = getPhotoOrganizerDirName();
	if (desc != NULL) {
		buf[0] = 0;
		for( item = getPhotoOrganizerDirectories(); item; item = item->next )
		{
			if (buf[0] > 0)
				strcat (buf, " ");
			strcat (buf, item->name);
		}
		fprintf (fp, "photo_organizer: \"%s\" %s\n", desc, buf);
	} else
		fprintf (fp, "#photo_organizer: \"Photos\"    /home/dgc/photos\n");
#endif /* PHOTO_ORGANIZER */


#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_recorder);
#endif
	fprintf (fp, "#recorder: \"By Request...\" /d1/Recordings /mm/byRequest/byRequest.sched 10\n");

#ifdef XMRADIO
#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_xm);
#endif
	fprintf (fp, "#xmpcr: /dev/ttyUSB0 /dev/dsp1\n");
#endif /* XMRADIO */

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_cache);
#endif
	addValue (fp, "cache", getCacheSize ());

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_thumb);
#endif
	addValue (fp, "thumbdir", getThumbCacheDir());

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_logfile);
#endif
	addValue (fp, "logfile", getLogfile());

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_divider);

	fprintf (fp, str_port);
#endif
	sprintf (buf, "%d", getDataPort());
	addValue (fp, "port", buf);

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_network);
#endif
	addValue (fp, "network", getIFName());

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_filetypes);
#endif
	buf[0] = 0;
	for( i = 0; getMpegEntry( i ) != NULL; i++ )
	{
		if (i > 0)
			strcat (buf, " ");
		strcat (buf, getMpegEntry(i));
	}
	addValue (fp, "mpeg", buf);
	buf[0] = 0;
	for( i = 0; getJpegEntry( i ) != NULL; i++ )
	{
		if (i > 0)
			strcat (buf, " ");
		strcat (buf, getJpegEntry(i));
	}
	addValue (fp, "jpeg", buf);
	buf[0] = 0;
	for( i = 0; getStreamEntry( i ) != NULL; i++ )
	{
		if (i > 0)
			strcat (buf, " ");
		strcat (buf, getStreamEntry(i));
	}
	addValue (fp, "stream", buf);
	buf[0] = 0;
	for( i = 0; getPlayEntry( i ) != NULL; i++ )
	{
		if (i > 0)
			strcat (buf, " ");
		strcat (buf, getPlayEntry(i));
	}
	addValue (fp, "playlist", buf);
	buf[0] = 0;
	for( i = 0; getScriptEntry( i ) != NULL; i++ )
	{
		if (i > 0)
			strcat (buf, " ");
		strcat (buf, getScriptEntry(i));
	}
	addValue (fp, "script", buf);
	buf[0] = 0;
	for( i = 0; getInternetEntry( i ) != NULL; i++ )
	{
		if (i > 0)
			strcat (buf, " ");
		strcat (buf, getInternetEntry(i));
	}
	addValue (fp, "internet", buf);

#ifdef DOCUMENTED_LOGFILE
	fprintf (fp, str_transcode);
#endif
	fprintf (fp, "transcode: \"ogg123 -q -d wav -f %%out %%in\" ogg\n");
	fprintf (fp, "transcode: \"sox -r 48000 -t ossdsp -c 2 %%in -t wav -c 2 -v 8 %%out\" ossdsp\n");
	fprintf (fp, "transcode: \"sox -t au %%in -t wav %%out\" au\n");
	fprintf (fp, "transcode: \"mplayer -cache 128 -really-quiet -vc null -vo null -ao pcm -aofile %%out %%in\" wma asx\n");
	fprintf (fp, "transcode: \"mplayer -cache 128 -really-quiet -vc null -vo null -ao pcm -aop list=volume:volume=160 -aofile %%out %%in\" ra rm\n");
	fprintf (fp, "transcode: \"faad -o %%out %%in\" aac\n");
	fprintf (fp, "transcode: \"lame --silent %%in -\" mp3low\n");
	fprintf (fp, "transcode: \"lame --silent --cbr -b 192 %%in -\" mp3high\n");

	fclose (fp);
}



