/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * Dave Clemans, April 2003
 *
 * Copyright (C) 2003  Dave Clemans
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * See the file "COPYING" for more details.
 */

#include "globals.h"

#include "mp3tech.h"

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/poll.h>
#include <time.h>
#include <errno.h>

#define __USE_GNU
#ifdef NEED_HSEARCH_R
 #include "hsearch_r.h"
#else
 #include <search.h>
#endif

#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/poll.h>
#include <netdb.h>

static struct aTranscodingMethod* activeTranscodingMethods = NULL;
static int transcodeTableInited = 0;
static struct hsearch_data transcodeTable;

/*
 * Get information about a MP3 file
 */
void getMp3Info(
   struct aDirectory*   dirp,
   struct aDirectoryEntry* file
   )
{
   int                  foundSomething = 0;
   mp3info              mp3Data;
   char*                fullpath;
   char*                cp;
   char*                realFullpath;
   int                  force;
   struct stat          statbuf;
   char                 errbuf[ 512 ];

   if( !IS_MPEG( file ) )
      return;

   cp = fullPathname( dirp->path, file->name );
   fullpath = decodeString( cp );
   free( cp );

   realFullpath = fullpath;
   while (*realFullpath) {
       if (*realFullpath == '.' && foundSomething == 0) {
           foundSomething++;
           realFullpath++;
           continue;
       }
       if (*realFullpath == 'm' && foundSomething == 1) {
           foundSomething++;
           realFullpath++;
           continue;
       }
       if (*realFullpath == '3' && foundSomething == 2) {
           foundSomething++;
           realFullpath++;
           continue;
       }
       if (*realFullpath == 'u' && foundSomething == 3) {
           foundSomething++;
           realFullpath++;
           continue;
       }
       if (*realFullpath == '/' && foundSomething == 4) {
           realFullpath++;
           break;
       }
       realFullpath++;
   }
   if (*realFullpath == '\0') {
       realFullpath = fullpath;
   }

   force = 0;
   statbuf.st_mtime = 0;
   if( stat( realFullpath, &statbuf ) < 0 )
   {
      addToLog( "Unable to get mp3info for file %s: %s\n", realFullpath,
         strerror_r( errno, errbuf, sizeof errbuf ) );
      free( fullpath );
      return;
   }
   if( statbuf.st_mtime > file->modified || !file->infoValid )
      force = 1;
   if( force && file->lastInfoCheck > file->modified )
      force = 0;

   if( force )
   {
      memset( &mp3Data, 0, sizeof mp3Data );
      mp3Data.filename = realFullpath;

      if( getDebugMode() )
         addToLog( "Get mp3info for file %s\n", realFullpath );

      mp3Data.file = fopen( realFullpath, "r" );
      get_mp3_info( &mp3Data, SCAN_QUICK, 0 );
      fclose( mp3Data.file );

      file->rate = ( int ) mp3Data.vbr_average;
      if( file->rate > 0 )
         file->duration = ( ( statbuf.st_size - 128 ) * 8 ) / file->rate;
      else
         file->duration = 0;
      if( mp3Data.id3_isvalid )
      {
         file->infoValid = 1;
         file->title = ( char* ) malloc( strlen( mp3Data.id3.title ) + 1 );
         strcpy( file->title, mp3Data.id3.title );
         file->album = ( char* ) malloc( strlen( mp3Data.id3.album ) + 1 );
         strcpy( file->album, mp3Data.id3.album );
         file->artist = ( char* ) malloc( strlen( mp3Data.id3.artist ) + 1 );
         strcpy( file->artist, mp3Data.id3.artist );
         text_genre( mp3Data.id3.genre, errbuf );
         file->genre = ( char* ) malloc( strlen( errbuf ) + 1 );
         strcpy( file->genre, errbuf );
         file->year = (char *) 0;
	 if (mp3Data.id3.year[0])	{
		file->year = ( char* ) malloc (5);
		strcpy (file->year, mp3Data.id3.year);
	}
      }
      file->lastInfoCheck = time( NULL );
   }
   free( fullpath );
}

typedef void (*sigHandler)(int);

/*
 * Sends a MPEG file to the client...
 */
void sendMPEGFile(
   int                  fd,
   char*                path,
   struct aParameterList* params
   )
{
   int                  i;
   char*                cp;
   struct stat          statbuf;
   int                  filep;
   int                  duration;
   struct aDirectory*   dirp;
   struct aDirectoryEntry* items;
   int                  rc;
   long                 seekOffset;
   char                 buffer[ 1024 ];

   path = decodeString( path );

   if( stat( path, &statbuf ) < 0 )
   {
      addToLog( "Unable to find %s: %s\n", path,
         strerror_r( errno, buffer, sizeof buffer ) );
      write( fd, "404 file not found\r\n", 20 );
      free( path );
      return;
   }

   items = getFileFromDirCache( path );
   if( items == NULL )
   {
      addToLog( "Unable to find %s in cache\n", path );
      write( fd, "404 file not found\r\n", 20 );
      free( path );
      return;
   }
   if( !items->infoValid )
      getMp3Info( items->dirp, items );
   if( getDebugMode() )
      addToLog( "Got entry for %s, duration=%d\n", items->name, items->duration );
   duration = items->duration;

   filep = open( path, O_RDONLY );
   if( filep >= 0 )
   {
      addToLog( "Sending MPEG file %s, duration=%d, length=%d\n",
         path, duration, statbuf.st_size );

      write( fd, HTTP_HEADER, HTTP_HEADER_LEN );
      sprintf( buffer, "TiVoAccurateDuration: %d\r\n", duration );
      write( fd, buffer, strlen( buffer ) );

      cp = findParameter( params, "Seek" );
      if( cp != NULL )
      {
         seekOffset = atoi( cp );
         seekOffset = ( ( seekOffset * items->rate ) / 8 ) + 128;
         lseek( filep, seekOffset, SEEK_SET );
      }
      sprintf( buffer, "Content-Length: %d\r\n", statbuf.st_size );
      write( fd, buffer, strlen( buffer ) );
      write( fd, "Content-Type: audio/mpeg\r\n\r\n", 28 );
#ifdef PLAYLIST_EDITING
      createPeerTempFile (fd, path);
#endif

      for( ; ; )
      {
         rc = read( filep, buffer, sizeof buffer );
         if( rc <= 0 )
            break;
         rc = write( fd, buffer, rc );
         if( rc <= 0 )
            break;
      }
      close( filep );
   }
   else
   {
      addToLog( "Unable to open %s: %s\n", path,
         strerror_r( errno, buffer, sizeof buffer ) );
      write( fd, "404 file not found\r\n", 20 );
      close( fd );
   }
#ifdef PLAYLIST_EDITING
      deletePeerTempFile (fd);
#endif
   free( path );
}

/*
 * Send an internet stream (or other file type we don't handle directly)
 * using mplayer and lame.  We use the pcm audio output driver in mplayer,
 * and lame to convert the resulting .wav file into mp3 format for TiVo.
 *
 * MPEG streams (a Shoutcast server) are handled natively.  Other mpeg
 * stream servers might work.
 */
void sendStreamFile(
   int                  fd,
   char*                path,
   struct aParameterList* params,
   int                  addHeader,
   int                  duration
   )
{
   int                  rc;
   int                  pid1;
   int                  pid2;
   int                  convfd;
   int                  port;
   int                  isMpeg;
   int                  isXM;
   int                  isStream;
   char*                cp;
   char			tmpchar;
   char*                scp;
   char*                libpath;
   char*                fifo;
   struct sockaddr_in   streamaddr;
   struct hostent*      host;
   FILE*                streamp;
   int                  pipefds2[ 2 ];
   char                 errbuf[ 512 ];
   char                 buffer[ 1024 ];
   char                 hostname[1024];
   char			temppath[1024];
   struct aDirectoryEntry* item = NULL;
   struct anAudioStream* stream;
   struct sigaction     pipeSig;
   struct aTranscodingMethod* tm;
   struct aTranscodingMethod* etm;

   isXM = 0;
#ifdef XMRADIO
   if( getXMRadio() != NULL )
   {
      item = getFileFromDirCache( path );
      if( item && IS_XMDIR( item ) )
         isXM = 1;
   }
#endif /* XMRADIO */

   stream = audioStreamFindStream(
      ( isInternetRef( path ) && path[ 0 ] == '/' ) ? &path[ 1 ] : path );
   if( stream )
   {
      audioStreamAddClient( stream, fd, 1, 0, isXM );
      return;
   }

   /* We need to guess whether or not this is a mpeg stream */
   isMpeg = 1;
   isXM = 0;
   isStream = 0;
#ifdef XMRADIO
   if( getXMRadio() != NULL )
   {
      item = getFileFromDirCache( path );
      if( item && IS_XMDIR( item ) )
      {
         tm = transcodeMethodFind( "ossdsp" );
         if( !tm )
         {
            write( fd, "404 File not found.\r\n", 20 );
            addToLog( "Can't find ossdsp transcode method for XM Radio!\n" );
            close( fd );
            return;
         }
         if( !grabXMRadioStream() )
         {
            /* Try once more, just in case a previous stream is shutting down */
            pauseFor( 2, 0 );

            if( !grabXMRadioStream() )
            {
               addToLog( "XM PCR Radio in use!\n" );
               write( fd, "404 XM PCR Radio in use!\r\n", 26 );
               close( fd );
               return;
            }
         }
         isXM = 1;
         isStream = 1;
         changeXMRadioChannel( item );
      }
   }
   if( !isXM )
   {
      strcpy(temppath,path);   /* Copy path to a work buffer */
      cp = temppath;
      while( *cp )
         cp++;
      while( cp > temppath && *cp != '.' )
         cp--;
      if( cp > temppath && *cp == '.' )
      {
         cp++;
         tm = transcodeMethodFind( cp );
         if( tm )
         {
            isStream = 1;
            isMpeg = 0;
         }
      }
   }
#endif /* XMRADIO */
   
   if( isMpeg && isInternetRef( path ) )
   {
      for( cp = temppath; *cp; cp++ )
      {
         if( strncmp( cp, "http://", 7 ) == 0 )
         {
            cp = &cp[ 7 ];
            break;
         }
      }
      scp = cp;
      while( *cp && *cp != ':' && *cp != '/' )
         cp++;
      port = 80;
      if( *cp == ':' )
      {
         port = atoi( &cp[ 1 ] );
      }

      /* peg the end of the host name to copy it*/
      tmpchar = *cp;
      *cp = '\0';

      strncpy(hostname,scp,1024);
      addToLog( "Host name: '%s', Port number: %i\n", scp, port );

      /* remove the hostname peg */
      *cp = tmpchar;

      host = gethostbyname( hostname );
      if( host == NULL || host->h_length != sizeof( streamaddr.sin_addr ) )
      {
         write( fd, "404 host not found\r\n", 20 );
         return;         
      }
      
      addToLog( "Buffer: '%s'\n", cp );
      
      streamaddr.sin_family = AF_INET;
      memcpy( &streamaddr.sin_addr, host->h_addr, host->h_length );
      streamaddr.sin_port = htons( port );
      convfd = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP );
      if( convfd < 0 )
      {
         write( fd, "404 socket not opened\r\n", 23 );
         return;         
      }
      rc = connect( convfd, ( struct sockaddr* ) &streamaddr, sizeof( struct sockaddr_in ) );
      addToLog( "connect: Error errno =%i rc = %i \n", errno, rc );
      if( rc < 0 )
      {
         addToLog( "connect: Error errno =%i\n", errno );
         close( convfd );
         write( fd, "404 unable to connect to host\r\n", 31 );
         return;         
      }

      /* Find start of the path in the URL */
      for( ; *cp; cp++)
      {
          if( *cp == '/' )
          {
	      break;
          }
      }	

      /* Make sure scan stopped on the '/' at the begining of path */
      if ( *cp != '/' )
      {
	      /* If no '/', Munge path to add a '/' */
	      /* Its OK at this point we got host and port info */
	      cp--; *cp = '/';
      }

      streamp = fdopen( convfd, "r+" );
      addToLog( "fetching stream: %s\n", cp);
      fprintf( streamp, "GET %s HTTP/1.1\r\n", cp  );
      fprintf( streamp, "Host: %s \r\n", hostname ); 
      fprintf( streamp, "User-Agent: byRequest\r\n" );
      fprintf( streamp, "\r\n");  			/* Close out the http request */
      fflush( streamp );
      if( fgets( buffer, sizeof buffer, streamp ) == NULL )
      {
         write( fd, "404 connection to host closed\r\n", 31 );
         fclose( streamp );
         return;
      }
      for( cp = buffer; *cp; cp++ )
      {
         if( *cp == '\r' || *cp == '\n' )
            *cp = '\0';
      }
      addToLog( "Msg from stream host: '%s'\n", buffer );
      while( strlen( buffer ) > 0 )
      {
         if( fgets( buffer, sizeof buffer, streamp ) == NULL )
            break;
         for( cp = buffer; *cp; cp++ )
         {
            if( *cp == '\r' || *cp == '\n' )
               *cp = '\0';
         }
         addToLog( "Msg from stream host: '%s'\n", buffer );

	 if ( strncasecmp( buffer, "Location: ", strlen( "Location: " ) ) == 0 )
	 {
	    /* being redirected to different server */
	    addToLog( "Redirecting to %s'\n", buffer+strlen("Location: ") );
	    fclose( streamp );
	    sendStreamFile( fd, buffer+strlen("Location: "), params, 1, 0 );
	    return;
	 }
      }
      addToLog( "Sending MPEG stream file: %s\n",
         ( isInternetRef( path ) && path[ 0 ] == '/' ) ? &path[ 1 ] : path );

      stream = audioStreamCreateStream();
      audioStreamSetUrl( stream, path );
      audioStreamSetFd( stream, convfd );
      audioStreamAddClient( stream, fd, addHeader, 0, 0 );
      audioStreamProcessStream( stream, 0 );
      audioStreamDestroyStream( stream );

      fclose( streamp );
      return;
   }

   if( !tm )
   {
      write( fd, "404 file not found\r\n", 20 );

      addToLog( "Failed to find transcoding method.\n" );
#ifdef XMRADIO
      if( isXM )
         releaseXMRadioStream();
#endif /* XMRADIO */
      close( fd );
      return;         
   }

   if( isInternetRef( path ) )
      etm = transcodeMethodFind( "mp3low" );
   else
      etm = transcodeMethodFind( "mp3high" );
   if( !etm )
   {
      write( fd, "404 file not found\r\n", 20 );

      addToLog( "Failed to find mp3 encoding method.\n" );
#ifdef XMRADIO
      if( isXM )
         releaseXMRadioStream();
#endif /* XMRADIO */
      close( fd );
      return;
   }

   fifo = ( char* ) malloc( 32 );
   strcpy( fifo, "/tmp/fifo.XXXXXX" );
   rc = mkstemp( fifo );
   close( rc );
   unlink( fifo );
   rc = mkfifo( fifo, 0600 );

   pid1 = transcodeMethodStart( tm,
#ifdef XMRADIO
      ( isXM ) ? getXMRadio() : 
#endif /* XMRADIO */
      ( ( isInternetRef( path ) && path[ 0 ] == '/' ) ? &path[ 1 ] : path ),
      -1, fifo, -1 );
   if( pid1 == 0 )
   {
      write( fd, "404 file not found\r\n", 20 );

      addToLog( "Failed to start transcoding method '%s': %s\n",
         tm->cmdLine, strerror_r( errno, errbuf, sizeof errbuf ) );
#ifdef XMRADIO
      if( isXM )
         releaseXMRadioStream();
#endif /* XMRADIO */
      unlink( fifo );
      close( fd );
      return;
   }

   rc = pipe( pipefds2 );
   if( rc < 0 )
   {
      write( fd, "404 file  not found\r\n", 20 );

      addToLog( "Failed to create encoding method pipe: %s\n",
         strerror_r( errno, errbuf, sizeof errbuf ) );
#ifdef XMRADIO
      if( isXM )
         releaseXMRadioStream();
#endif /* XMRADIO */
      close( fd );
      unlink( fifo );
      kill( -pid1, SIGTERM );
      kill( -pid1, SIGKILL );
      wait4( -pid1, NULL, WNOHANG, NULL );
      return;
   }

   pid2 = transcodeMethodStart( etm, fifo, -1, NULL, pipefds2[ 1 ] );
   if( pid2 == 0 )
   {
      write( fd, "404 file  not found\r\n", 20 );

      addToLog( "Failed to start encoding method '%s': %s\n",
         etm->cmdLine, strerror_r( errno, errbuf, sizeof errbuf ) );
#ifdef XMRADIO
      if( isXM )
         releaseXMRadioStream();
#endif /* XMRADIO */
      close( fd );
      close( pipefds2[ 0 ] );
      close( pipefds2[ 1 ] );
      unlink( fifo );
      kill( -pid1, SIGTERM );
      kill( -pid1, SIGKILL );
      wait4( -pid1, NULL, WNOHANG, NULL );
      return;
   }

   convfd = pipefds2[ 0 ];
   close( pipefds2[ 1 ] );

   pauseFor( 0, 10000000 );

   stream = audioStreamCreateStream();
   audioStreamSetUrl( stream,
      ( isInternetRef( path ) && path[ 0 ] == '/' ) ? &path[ 1 ] : path );
   audioStreamSetFd( stream, convfd );
   audioStreamAddClient( stream, fd, addHeader, duration, isXM );
   audioStreamProcessStream( stream, 0 );
   audioStreamDestroyStream( stream );

   kill( -pid1, SIGTERM );
   kill( -pid2, SIGTERM );
   kill( -pid1, SIGKILL );
   kill( -pid2, SIGKILL );
   unlink( fifo );
   wait4( -pid1, NULL, WNOHANG, NULL );
   wait4( -pid2, NULL, WNOHANG, NULL );
}

/*
 * Create a new transcoding method
 */
struct aTranscodingMethod* transcodeMethodCreate(
   char*                   cmd,
   char*                   fileType
   )
{
   struct aTranscodingMethod *tm;
   int                     count;
   int                     i;
   char*                   cp;
   ENTRY                   item;
   ENTRY*                  hentry;

   if( getDebugMode() )
   {
      addToLog( "Create transcoding method '%s' for '%s' streams.\n",
         cmd, fileType );
   }

   if( !transcodeTableInited )
   {
      transcodeTableInited = 1;
      hcreate_r( 64, &transcodeTable );
   }

   tm = ( struct aTranscodingMethod* ) malloc( sizeof( *tm ) );
   tm->cmdLine = strdup( cmd );
   tm->cmdLineCopy = strdup( cmd );
   tm->fileType = strdup (fileType);
   count = 0;
   cp = tm->cmdLineCopy;
   while( *cp && isspace( *cp ) )
      cp++;
   for( ; *cp; cp++ )
   {
      if( isspace( *cp ) )
      {
         count++;
         while( *cp && isspace( *cp ) )
            cp++;
         cp--;
      }
   }
   count += 2;
   tm->parsedCmd = ( char** ) malloc( count * sizeof( char* ) );
   for( i = 0; i < count; i++ )
   {
      tm->parsedCmd[ i ] = NULL;
   }
   cp = tm->cmdLineCopy;
   while( *cp && isspace( *cp ) )
      cp++;
   tm->parsedCmd[ 0 ] = cp;
   i = 0;
   for( ; *cp; cp++ )
   {
      if( isspace( *cp ) )
      {
         *cp++ = '\0';
         while( *cp && isspace( *cp ) )
            cp++;
         if( *cp )
            tm->parsedCmd[ ++i ] = cp;
      }
   }

   tm->next = activeTranscodingMethods;
   activeTranscodingMethods = tm;

   item.key = strdup( fileType );
   item.data = tm;
   hsearch_r( item, ENTER, &hentry, &transcodeTable );

   return tm;
}

/*
 * Destroy a transcoding method
 *
 * Note: the hash table isn't updated...
 */
void transcodeMethodDestroy(
   struct aTranscodingMethod* tm
   )
{
   struct aTranscodingMethod* tmp;

   for( tmp = activeTranscodingMethods; tmp; tmp = tmp->next )
   {
      if( tmp == tm || tmp->next == tm )
         break;
   }
   if( !tmp )
   {
      /* not found... should never happen */
      return;
   }
   if( tmp == activeTranscodingMethods )
      activeTranscodingMethods = tmp->next;
   else
      tmp->next = tm->next;

   free( tm->cmdLine );
   free( tm->cmdLineCopy );
   free( tm->fileType );
   free( tm );
}

/*
 * Destroys all current transcoding methods (for reloading our configuration)
 */
void transcodeMethodDestroyAll( void )
{
   while( activeTranscodingMethods )
   {
      transcodeMethodDestroy( activeTranscodingMethods );
   }

   hdestroy_r( &transcodeTable );
   transcodeTableInited = 0;
}

/*
 * Get pointer to base transcoding methods
 */
struct aTranscodingMethod* getTranscodingMethods (void)
{
	return activeTranscodingMethods;
}


/*
 * Find a transcoding method for a file type
 */
struct aTranscodingMethod* transcodeMethodFind(
   char*                      fileType
   )
{
   ENTRY                      item;
   ENTRY*                     hentry;

   if( getDebugMode() )
      addToLog( "Try to find transcoding method '%s'\n", fileType );

   item.key = fileType;
   item.data = NULL;
   hsearch_r( item, FIND, &hentry, &transcodeTable );
   if( hentry )
   {
      if( getDebugMode() )
         addToLog( "Transcoding method found for stream '%s'\n", fileType );
            
      return ( struct aTranscodingMethod* ) hentry->data;
   }

   return NULL;
}

/*
 * Fork off and start to run a transcoding method...
 */
int transcodeMethodStart(
   struct aTranscodingMethod* tm,
   char*                      fileIn,
   int                        fdIn,
   char*                      fileOut,
   int                        fdOut
   )
{
   int                        pid;
   int                        rc;
   struct sigaction           pipeSig;
   int                        count;

   addToLog( "Starting transcoding method '%s'\n", tm->cmdLine );
   addToLog( "... For input url '%s', output file '%s'\n",
      ( fileIn ) ? fileIn : "-", ( fileOut ) ? fileOut : "-" );

   pid = fork();
   if( pid < 0 )
   {
      addToLog( "Fork to run transcode method '%s' failed.\n", tm->cmdLine );
      return 0;
   }
   if( pid == 0 )
   {
      for( count = 1; tm->parsedCmd[ count ]; count++ )
      {
         if( fileIn && ( strcmp( tm->parsedCmd[ count ], "%in" ) == 0 ) )
         {
            tm->parsedCmd[ count ] = fileIn;
         }
         if( fileOut && ( strcmp( tm->parsedCmd[ count ], "%out" ) == 0 ) )
         {
            tm->parsedCmd[ count ] = fileOut;
         }
      }

      setsid();
      if( fdIn != -1 )
         dup2( fdIn, 0 );
      else
         close( 0 );
      if( fdOut != -1 )
         dup2( fdOut, 1 );
      else
         close( 1 );
      if( fdIn != -1 )
         close( fdIn );
      if( fdOut != -1 )
         close( fdOut );
      if( getDebugMode() )
      {
         rc = open( "/tmp/byRequest_transcode.log", O_WRONLY | O_APPEND | O_CREAT );
         dup2( rc, 2 );
         close( rc );
      }
      else
      {
         rc = open( "/dev/null", O_RDWR );
         dup2( rc, 2 );
         close( rc );
      }

      pipeSig.sa_handler = SIG_DFL;
      sigemptyset( &pipeSig.sa_mask );
      pipeSig.sa_flags = 0;
      sigaction( SIGPIPE, &pipeSig, NULL );

      execv( getToolPath( tm->parsedCmd[ 0 ] ), tm->parsedCmd );

      addToLog( "Exec to run transcode method '%s' failed.\n", tm->cmdLine );

      exit( -1 );
   }

   return pid;
}
