/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * Dave Clemans, April 2003
 *
 * Copyright (C) 2003  Dave Clemans
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * See the file "COPYING" for more details.
 */

#include "globals.h"

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/poll.h>
#include <netdb.h>

static struct aBeacon* topBeacon = NULL;

/*
 * Send outgoing beacon to the specified address
 *    This will either be a specific or broadcast address
 */
void sendBeaconMessage(
   int                  fd,
   struct sockaddr_in*  client,
   int                  len,
   int                  broadcast
   )
{
   char*                hname;
   char*                cp;
   char                 outgoingBeacon[ 2048 ];

   hname = ( char* ) malloc( strlen( getHostname() ) + 1 );
   strcpy( hname, getHostname() );
   for( cp = hname; *cp; cp++ )
   {
      if( *cp == '.' )
         *cp = '\0';
   }

   sprintf( outgoingBeacon,
      "TiVoConnect=1\nswversion=%s\nmethod=%s\nidentity=%lu\nmachine=%s\nplatform=pc/linux\nservices=TiVoMediaServer:%d/http\n",
      SERVER_VERSION, ( broadcast ) ? "broadcast" : "connected",
      getHostid(), hname, getDataPort() );

   free( hname );

   ( void ) sendto( fd, outgoingBeacon, strlen( outgoingBeacon ), 0,
      ( struct sockaddr* ) client, len );

   if( getDebugMode() )
      addToLog( "Sent beacon:\n%s", outgoingBeacon );
}

/*
 * Parse and save a received beacon packet from another server, or from
 * a TiVo.
 *
 * Returns true if this was a broadcast beacon msg
 */
int rcvBeaconMessage(
   char*                beacon
   )
{
   char*                tivoConnect = NULL;
   char*                swVersion = NULL;
   char*                method = NULL;
   char*                identity = NULL;
   char*                machine = NULL;
   char*                platform = NULL;
   char*                services = NULL;
   char*                cp;
   char*                scp;
   char*                tokptr;
   struct aBeacon*      b;
   int                  len;
   time_t               current;
   char                 buf[ 32 ];
   static time_t        lastSummary = 0;

   cp = strtok_r( beacon, "=\r\n", &tokptr );
   while( cp != NULL )
   {
      scp = cp;
      cp = strtok_r( NULL, "=\r\n", &tokptr );
      if( strcasecmp( scp, "tivoconnect" ) == 0 )
         tivoConnect = cp;
      else if( strcasecmp( scp, "swversion" ) == 0 )
         swVersion = cp;
      else if( strcasecmp( scp, "method" ) == 0 )
         method = cp;
      else if( strcasecmp( scp, "identity" ) == 0 )
         identity = cp;
      else if( strcasecmp( scp, "machine" ) == 0 )
         machine = cp;
      else if( strcasecmp( scp, "platform" ) == 0 )
         platform = cp;
      else if( strcasecmp( scp, "services" ) == 0 )
         services = cp;
      cp = strtok_r( NULL, "=\r\n", &tokptr );
   }

   if( tivoConnect == NULL )
      return 0;

   current = time( NULL );
   for( b = topBeacon; b != NULL; b = b->next )
   {
      if( strcasecmp( machine, b->machine ) == 0 ||
         strcasecmp( identity, b->identity ) == 0 )
      {
         break;
      }
   }
   if( b == NULL )
   {
      b = ( struct aBeacon* ) calloc( 1, sizeof( *b ) );
      b->next = NULL;

      if ( machine )
      {
	 b->machine = ( char* ) malloc( strlen( machine ) + 1 );
	 strcpy( b->machine, machine );
      }
      if ( identity )
      {
	 b->identity = ( char* ) malloc( strlen( identity ) + 1 );
	 strcpy( b->identity, identity );
      }
      if ( swVersion )
      {
	 b->swversion = ( char* ) malloc( strlen( swVersion ) + 1 );
	 strcpy( b->swversion, swVersion );
      }
      if ( method )
      {
	 b->method = ( char* ) malloc( strlen( method ) + 1 );
	 strcpy( b->method, method );
      }
      if ( platform )
      {
	 b->platform = ( char* ) malloc( strlen( platform ) + 1 );
	 strcpy( b->platform, platform );
      }
      if ( services )
      {
	 b->services = ( char* ) malloc( strlen( services ) + 1 );
	 strcpy( b->services, services );
      }

      b->next = topBeacon;
      topBeacon = b;

      addToLog( "Received new beacon: machine(%s) platform(%s) services(%s)\n", 
		b->machine ? b->machine : "-",
		b->platform ? b->platform : "-", 
		b->services ? b->services : "-" );
   }
   b->lastSeen = current;

   if( lastSummary == 0 )
      lastSummary = current;
   if( lastSummary + 1800 < current )
   {  /* Give a summary of received server beacons every half hour or so */
      len = 0;
      for( b = topBeacon; b != NULL; b = b->next )
      {
         len += strlen( b->machine ) + 32;
      }
      scp = ( char* ) malloc( len + 128 );
      strcpy( scp, "Known servers: " );
      for( b = topBeacon; b != NULL; b = b->next )
      {
         strcat( scp, b->machine );
         sprintf( buf, "(%d)", current - b->lastSeen );
         strcat( scp, buf );
         if( b->next != NULL )
            strcat( scp, "," );
      }
      strcat( scp, "\n" );
      addToLog( scp );
      free( scp );
      lastSummary = current;
   }

   if( strcasecmp( method, "broadcast" ) == 0 )
      return 1;
   return 0;
}
