/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * Dave Clemans, April 2003
 *
 * Copyright (C) 2003  Dave Clemans
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * See the file "COPYING" for more details.
 */

#include "globals.h"

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <stdlib.h>
#define  __USE_GNU         1
#include <search.h>
#include <time.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/wait.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/poll.h>
#include <netdb.h>

/*
 * Generally unimplemented parameters:
 *    Format      {text/xml,text/http}
 *                we always do text/xml
 *    Session     we logically have one session
 */

/*
 * Process an incoming datagram from the client, typically TiVo.
 * This will usually be the beacon advertising what services are provided.
 */
void processDatagram(
   int                  fd
   )
{
   int                  rc;
   char*                cp;
   int                  len;
   struct sockaddr_in   client;
   struct in_addr       client_address;
   char                 datagram[ 2048 ];
   char                 buffer[ 128 ];
   
   /* We only expect to see beacon msgs from TiVo's and possibly other tivo servers */

   client.sin_family = AF_INET;
   client_address.s_addr = htonl( INADDR_ANY );
   client.sin_addr = client_address;
   client.sin_port = htons( 2190 );
   len = sizeof( struct sockaddr_in );

   rc = recvfrom( fd, datagram, sizeof datagram, 0,
      ( struct sockaddr* ) &client, &len );
   if( rc > 0 )
      datagram[ rc ] = '\0';
   for( cp = datagram; *cp; cp++ )
      /* do nothing */;
   if( cp[ -1 ] == '\r' || cp[ -1 ] == '\n' )
      *--cp = '\0';
   if( cp[ -1 ] == '\r' || cp[ -1 ] == '\n' )
      *--cp = '\0';

   if( rcvBeaconMessage( datagram ) )
      sendBeaconMessage( fd, &client, len, 0 );
}

/*
 * Process a stream command from the client (e.g., from the TiVo)
 * Currently supported are:
 *   Get the contents of a file or directory by pathname
 *   QueryContainer (both normal and root directories)
 *   QueryServer
 */
void processCommand(
   int                  fd
   )
{
   int                  length;
   int                  rc;
   char*                ptr;
   char*                cp;
   char*                cmdstart;
   char*                command;
   char*                lasts;
   struct aParameterList* params;
   int                  cmdlineLength;
   char*                cmdline;
   FILE*		fp;

   fp = fdopen (fd, "r");
   cmdline = ( char* ) malloc( 1024 );
   cmdlineLength = 1024;
#if 0
   length = read( fd, cmdline, cmdlineLength - 1 );
   if( length <= 0 )
#else
   ptr = fgets( cmdline, cmdlineLength - 1, fp );
   length = strlen (cmdline);
   if( ptr == NULL )
#endif
   {
      close( fd );
      return;
   }

   if ( (length == 4) || ((cmdline[0] == 0) && (cmdline[1] == 0)) )
   {
      char*                hname;
      char*                cp;
      char                 outgoingBeacon[ 2048 ];
      long                 len;

      hname = ( char* ) malloc( strlen( getHostname() ) + 1 );
      strcpy( hname, getHostname() );
      for( cp = hname; *cp; cp++ )
      {
	 if( *cp == '.' )
	    *cp = '\0';
      }

      sprintf( outgoingBeacon,
	       "TiVoConnect=1\nswversion=%s\nmethod=connected\nidentity=%lu\nmachine=%s\nplatform=pc/linux\nservices=TiVoMediaServer:%d/http\n",
	       SERVER_VERSION, getHostid(), hname, getDataPort() );

      free( hname );
      free(cmdline);
      len = htonl( strlen(outgoingBeacon) );
      write( fd, &len, sizeof(len) );
      write( fd, outgoingBeacon, strlen( outgoingBeacon ) );

      addToLog( "Sent beacon/TCP:\n%s", outgoingBeacon );

      close(fd);
      return;
   }
   cmdline[ length ] = '\0';

   ptr = cmdline;
   while( *ptr )
   {
      ptr++;
   }
   if( ptr > cmdline )
      ptr--;
   if( ptr > cmdline && *ptr == '\n' && ptr[ -1 ] == '\\' )
      ptr--;
#if 0
   while( *ptr == '\\' || *ptr != '\n' )
   {
      if( *ptr == '\\' )
         *ptr = ' ';
      if( cmdlineLength - length < 1024 )
      {
         cmdlineLength += 1024;
         cp = ( char* ) malloc( cmdlineLength );
         strcpy( cp, cmdline );
         free( cmdline );
         cmdline = cp;
      }
      ptr = cmdline;
      ptr = &ptr[ length ];
      rc = read( fd, ptr, cmdlineLength - length - 1 );
      if( rc < 0 )
	 break;
      ptr[ rc ] = '\0';
      length += rc;
      while( *ptr )
      {
         ptr++;
      }
      ptr--;
      if( ptr > cmdline && *ptr == '\n' && ptr[ -1 ] == '\\' )
         ptr--;
   }
#endif
   
   /*
    * Isolate the relevant parts of the command string
    */
   for( cp = cmdline; *cp; cp++ )
   {
      if( *cp == '\n' || *cp == '\r' )
         *cp = '\0';
   }
   addToLog( "Rcvd Command:\n%s\n", cmdline );
   for( cp = cmdline; *cp; cp++ )
   {
#ifdef WEB_ADMIN
      if (strncmp ("admin", cp, 5) == 0)
      {
	adminInterface (fd, fp, cp);
	free (cmdline);
	return;
      }
#endif /* WEB_ADMIN */
      if( strncasecmp( "TiVoConnect", cp, 11 ) == 0 )
         break;
   }
   cmdstart = &cp[ 11 ];
   for( cp = cmdstart; *cp; cp++ )
      /* do nothing */;
   while( cp > cmdstart && *cp != ' ' )
      cp--;
   if( *cp == ' ' )
      *cp = '\0';
   
   /*
    * Parse any command parameters
    */
   params = ( struct aParameterList* ) malloc( sizeof( *params ) );
   params->top = NULL;
   params->bottom = NULL;
   lasts = NULL;
   cp = strtok_r( cmdline, "?&", &lasts );
   while( cp != NULL )
   {
      if( strlen( cp ) == 0 )
      {
         cp = strtok_r( NULL, "?&", &lasts );
         continue;
      }
      for( ptr = cp; *ptr; ptr++ )
      {
         if( *ptr == '=' )
         {
            *ptr++ = '\0';
            break;
         }
      }
      storeParameter( params, cp, ptr );
      cp = strtok_r( NULL, "?&", &lasts );
   }
   
   command = findParameter( params, "command" );
   if( command == NULL && strlen( cmdstart ) == 0 )
   {
      queryContainer( fd, params );
   }
   else if( command == NULL )
   {
      sendFile( fd, cmdstart, params );
   }
   else if( strcasecmp( command, "QueryContainer" ) == 0 )
   {
#ifdef PLAYLIST_EDITING
      char *plist;
      plist = findParameter (params, "PlaylistEdit");
      if (plist != NULL) {
	 if (modifyPlaylist (fd, params, plist))
	    queryContainer (fd, params);
      } else
#endif
      queryContainer( fd, params );
   }
   else if( strcasecmp( command, "QueryServer" ) == 0 )
   {
      queryServer( fd, params );
   }
   else if( strcasecmp( command, "ResetServer" ) == 0 )
   {
      resetServer( fd, params );
   }
   else if( strcasecmp( command, "FlushServer" ) == 0 )
   {
      freeParameters( params );
      addToLog( "FlushServer command... ignored.\n" );
      close( fd );
   }
   else if( strcasecmp( command, "QueryItem" ) == 0 )
   {
      queryItem( fd, params );
   }
   else
   {
      freeParameters( params );
      addToLog( "Unknown command: %s\n", command );
      close( fd );
   }

   free( params );
   free( cmdline );

   return;
}

typedef void (*sigHandler)(int);

/*
 * Send the contents of a container to the caller as an html message
 */
static void sendContainerAsHTML(
   int                     fd,
   char*                   path,
   struct aDirectoryEntry* dirItem,
   struct aDirectoryEntry** itemArray,
   int                     anchorItem,
   int                     itemsToSend,
   int                     itemCount,
   int                     totalduration,
   int                     fileType
   )
{
   int                     i;
   int                     len;
   char*                   fullpath;
   char*                   encoded;
   char*                   addr;
   struct aMessage*        msg;
   int                     isRoot;
   char*                   cp;
   struct aDirectoryEntry* rootdir;
   char                    buffer[ 256 ];

   isRoot = strcmp( dirItem->dirp->path, "/" ) == 0 &&
      strcmp( path, "/" ) == 0;

   addr = getHostaddress();

   msg = newMessage();

   htmlHeader (msg, isRoot, path);

/*
  <li><a
 href="http://192.168.4.8:%d/TIVoConnect?Command=QueryContainer&amp;Container=
%2Fdata%2Farchive%2FInternetRadio&amp;Format=text/html">Internet
Radio</a>&nbsp;&nbsp;  &nbsp;&nbsp;  &nbsp;&nbsp;  &nbsp;&nbsp; 
&nbsp;&nbsp;  &nbsp;&nbsp;  &nbsp;&nbsp;  &nbsp;&nbsp;  audio<br>
  </li>
*/
   
#ifdef PLAYLIST_EDITING
   if( dirItem && IS_PLAYLIST( dirItem )) {
	if (anchorItem == 0)	{
//		if (itemsToSend != itemCount)
			itemsToSend--;
		if (strstr (dirItem->name, ".m3u")) {
         	  appendToMsg( msg, " <li><a\n" );
         	  appendToMsg( msg, "  href=\"http://%s:2190/TiVoConnect?Command=QueryContainer&Container=%s&PlaylistEdit=Modify&",
            		addr, path );
		  cp = strrchr (path, '/');
         	  appendToMsg( msg, "Format=text/html\">Modify Playlist - '%s'</a><br>\n", ++cp);
         	  appendToMsg( msg, " </li>\n" );
		}
	} else {
		anchorItem--;
	}
   }
#endif

   for( i = anchorItem; i < anchorItem + itemsToSend; i++ )
   {
      if( IS_FOLDER( itemArray[ i ] ) )
      {
         appendToMsg( msg, " <li><a\n" );
         appendToMsg( msg, "  href=\"http://%s:%d/TiVoConnect?Command=QueryContainer&Container=",
            addr, getDataPort() );
         if( isRoot )
         {
            fullpath = encodeString( itemArray[ i ]->name );
         }
         else
         {
            cp = fullPathname( path, itemArray[ i ]->name );
            fullpath = encodeString( cp );
            free( cp );
         }
         appendToMsg( msg, "%s", fullpath );
         free( fullpath );
         appendToMsg( msg, "&Format=text/html\">%s</a><br>\n",
            ( isRoot ) ? itemArray[ i ]->description : itemArray[ i ]->name );
         appendToMsg( msg, " </li>\n" );
      }
      else if( IS_PLAYLIST( itemArray[ i ] ) )
      {
         appendToMsg( msg, " <li><a\n" );
         appendToMsg( msg, "  href=\"http://%s:%d/TiVoConnect?Command=QueryContainer&Container=",
            addr, getDataPort() );
         if( isRoot )
         {
            fullpath = encodeString( itemArray[ i ]->name );
         }
         else
         {
            cp = fullPathname( path, itemArray[ i ]->name );
            fullpath = encodeString( cp );
            free( cp );
         }
         appendToMsg( msg, "%s", fullpath );
         free( fullpath );
         appendToMsg( msg, "&Format=text/html\">%s</a><br>\n",
            ( isRoot ) ? itemArray[ i ]->description : itemArray[ i ]->name );
         appendToMsg( msg, " </li>\n" );
      }
      else if( IS_MPEG( itemArray[ i ] ) || IS_JPEG( itemArray[ i ] ) ||
         IS_STREAM( itemArray[ i ] ) )
      {
         appendToMsg( msg, " <li><a\n" );
         appendToMsg( msg, "  href=\"http://%s:%d/TiVoConnect",
            addr, getDataPort() );
         if( dirItem && IS_PLAYLIST( dirItem ) )
         {
            if( itemArray[ i ]->name[ 0 ] == '/' )
               fullpath = encodeString( itemArray[ i ]->name );
            else
            {
               cp = ( char* ) malloc( strlen( itemArray[ i ]->name ) + 2 );
               sprintf( cp, "/%s", itemArray[ i ]->name );
               fullpath = encodeString( cp );
               free( cp );
            }
         }
         else
         {
            cp = fullPathname( itemArray[ i ]->dirp->path, itemArray[ i ]->name );
            fullpath = encodeString( cp );
            free( cp );
         }
         appendToMsg( msg, fullpath );
         free( fullpath );
         if( !IS_MPEG( itemArray[ i ] ) )
         {
            appendToMsg( msg, "\">%s</a><br>\n",
               ( isRoot ) ? itemArray[ i ]->description : itemArray[ i ]->name );
         }
         else
         {
            appendToMsg( msg, "\">%s</a><br>\n",
               itemArray[ i ]->infoValid ? itemArray[ i ]->title :
               itemArray[ i ]->name );
         }
         appendToMsg( msg, " </li>\n" );
      }
      else
      {
         addToLog( "Unknown file type for %s\n", itemArray[ i ] );
      }
   }

   htmlFooter (msg);
   htmlSend (fd, msg);

   addToLog( "Directory contents, length=%d:\n%s", msg->size, msg->msg );

   free( addr );
   freeMessage( msg );
}

/*
 * Append info about a url pathname to a xml message
 */
static void addUrlToMsg(
   struct aMessage*        msg,
   struct aDirectoryEntry* item,
   struct aDirectoryEntry* dirItem
   )
{
   char*                   cp;
   char*                   fullpath;

   if( dirItem && IS_PLAYLIST( dirItem ) )
   {
      if( item->name[ 0 ] == '/' )
      {
         saveUrlEntry( item->name, item );
         fullpath = encodeString( item->name );
      }
      else
      {
         cp = ( char* ) malloc( strlen( item->name ) + 2 );
         sprintf( cp, "/%s", item->name );
         saveUrlEntry( cp, item );
         fullpath = encodeString( cp );
         free( cp );
      }
   }
   else
   {
      cp = fullPathname( item->dirp->path, item->name );
      saveUrlEntry( cp, item );
      fullpath = encodeString( cp );
      free( cp );
   }
   cp = encodeString( fullpath );
   appendToMsg( msg, "    <Url>/TiVoConnect%s</Url>\n", cp );
   free( cp );
   free( fullpath );
}

/*
 * Append info about a directory entry to a xml message
 */
static void addItemToMsg(
   struct aMessage*        msg,
   struct aDirectoryEntry* item,
   int                     isRoot,
   struct aDirectoryEntry* dirItem,
   char*                   path
   )
{
   char*                   cp;
   char*                   fullpath;
   char*                   encoded;
   char*                   url;

   appendToMsg( msg, " <Item>\n" );
   if( IS_FOLDER( item ) )
   {
      appendToMsg( msg, "  <Details>\n" );
      if( isRoot )
      {
         if( item->folderType & FILE_MPEG )
            cp = "tivo-music";
         else
            cp = "tivo-photos";
      }
      else
         cp = "folder";
      appendToMsg( msg, "   <ContentType>x-container/%s</ContentType>\n",
         cp );
      appendToMsg( msg, "   <SourceFormat>x-container/folder</SourceFormat>\n" );
      appendToMsg( msg, "   <Title>%s</Title>\n",
         ( isRoot ) ? item->description : item->name );
      appendToMsg( msg, "  </Details>\n" );
      appendToMsg( msg, "  <Links>\n" );
      appendToMsg( msg, "   <Content>\n" );
      if( isRoot )
      {
         fullpath = encodeString( item->name );
      }
      else
      {
         cp = fullPathname( path, item->name );
         fullpath = encodeString( cp );
         free( cp );
      }
      appendToMsg( msg, "    <Url>/TiVoConnect?Command=QueryContainer&amp;Container=%s",
         fullpath );
      free( fullpath );
      appendToMsg( msg, "</Url>\n" );
      appendToMsg( msg, "   </Content>\n" );
      appendToMsg( msg, "  </Links>\n" );
   }
   if( IS_PLAYLIST( item ) )
   {
      appendToMsg( msg, "  <Details>\n" );
      appendToMsg( msg, "   <ContentType>x-container/playlist</ContentType>\n" );
      appendToMsg( msg, "   <SourceFormat>x-container/playlist</SourceFormat>\n" );
      appendToMsg( msg, "   <Title>%s</Title>\n", item->name );
      appendToMsg( msg, "  </Details>\n" );
      appendToMsg( msg, "  <Links>\n" );
      appendToMsg( msg, "   <Content>\n" );
      fullpath = fullPathname( item->dirp->path, item->name );
      appendToMsg( msg, "    <Url>/TiVoConnect?Command=QueryContainer&amp;Container=%s</Url>\n",
         fullpath );
      free( fullpath );
      appendToMsg( msg, "   </Content>\n" );
      appendToMsg( msg, "  </Links>\n" );
   }
   else if( IS_MPEG( item ) )
   {
      appendToMsg( msg, "  <Details>\n" );
      appendToMsg( msg, "   <MusicGenre>%s</MusicGenre>\n",
         ( item->genre ) ? item->genre : "" );
      appendToMsg( msg, "   <ArtistName>%s</ArtistName>\n",
         ( item->artist ) ? item->artist : "" );
      appendToMsg( msg, "   <ContentType>audio/*</ContentType>\n" );
      appendToMsg( msg, "   <SourceFormat>audio/mpeg</SourceFormat>\n" );
      appendToMsg( msg, "   <Title>%s</Title>\n",
         ( item->title && strlen( item->title ) > 0 ) ?
         item->title : item->name );
      appendToMsg( msg, "   <SongTitle>%s</SongTitle>\n",
         ( item->title && strlen( item->title ) > 0 ) ?
         item->title : item->name );
      appendToMsg( msg, "   <AlbumTitle>%s</AlbumTitle>\n",
         ( item->album ) ? item->album : "" );
      appendToMsg( msg, "   <Duration>%d</Duration>\n", item->duration );
         appendToMsg( msg, "   <AlbumYear>%s</AlbumYear>\n", 
		(item->year == (char *)0) ? "unknown" : item->year);
      appendToMsg( msg, "  </Details>\n" );
      appendToMsg( msg, "  <Links>\n" );
      appendToMsg( msg, "   <Content>\n" );
      addUrlToMsg( msg, item, dirItem );
      appendToMsg( msg, "    <Seekable>Yes</Seekable>\n" );
      appendToMsg( msg, "   </Content>\n" );
      appendToMsg( msg, "  </Links>\n" );
   }
   else if( IS_STREAM( item ) )
   {
      appendToMsg( msg, "  <Details>\n" );
      appendToMsg( msg, "   <ContentType>audio/*</ContentType>\n" );
      appendToMsg( msg, "   <SourceFormat>audio/mpeg</SourceFormat>\n" );
      appendToMsg( msg, "   <Title>%s</Title>\n",
         ( item->title && strlen( item->title ) > 0 ) ?
         item->title : item->name );
      cp = dirLeafName( item->name );
      appendToMsg( msg, "   <SongTitle>%s</SongTitle>\n", cp );
	   free( cp );
      appendToMsg( msg, "   <AlbumTitle>%s</AlbumTitle>\n",
         ( item->title && strlen( item->title ) > 0 ) ?
         item->title : item->name );
      appendToMsg( msg, "   <Duration>%d</Duration>\n", 0 );
      appendToMsg( msg, "  </Details>\n" );
      appendToMsg( msg, "  <Links>\n" );
      appendToMsg( msg, "   <Content>\n" );
      addUrlToMsg( msg, item, dirItem );
      appendToMsg( msg, "    <Seekable>No</Seekable>\n" );
      appendToMsg( msg, "   </Content>\n" );
      appendToMsg( msg, "  </Links>\n" );
   }
   else if( IS_JPEG( item ) )
   {
      appendToMsg( msg, "  <Details>\n" );
      appendToMsg( msg, "   <ContentType>image/*</ContentType>\n" );
      appendToMsg( msg, "   <SourceFormat>image/jpeg</SourceFormat>\n" );
      appendToMsg( msg, "   <LastChangeDate>0x%x</LastChangeDate>\n",
         item->modified );
      appendToMsg( msg, "   <Title>%s</Title>\n",
         item->name );
      appendToMsg( msg, "  </Details>\n" );
      appendToMsg( msg, "  <Links>\n" );
      appendToMsg( msg, "   <Content>\n" );
      addUrlToMsg( msg, item, dirItem );
      appendToMsg( msg, "   </Content>\n" );
      appendToMsg( msg, "  </Links>\n" );
   }
   else if( IS_CGI( item ) )
   {
      appendToMsg( msg, "  <Details>\n" );
      appendToMsg( msg, "   <ContentType>image/*</ContentType>\n" );
      appendToMsg( msg, "   <SourceFormat>image/jpeg</SourceFormat>\n" );
#if 0
      appendToMsg( msg, "   <LastChangeDate>0x%x</LastChangeDate>\n",
         item->modified );
#endif
      appendToMsg( msg, "   <Title>%s</Title>\n",
         item->name );
      appendToMsg( msg, "  </Details>\n" );
      appendToMsg( msg, "  <Links>\n" );
      appendToMsg( msg, "   <Content>\n" );
      addUrlToMsg( msg, item, dirItem );
      appendToMsg( msg, "   </Content>\n" );
      appendToMsg( msg, "  </Links>\n" );
   }
   appendToMsg( msg, " </Item>\n" );
}

/*
 * Send the contents of a container to the caller as an xml message
 */
static void sendContainerAsXML(
   int                     fd,
   char*                   path,
   struct aDirectoryEntry* dirItem,
   struct aDirectoryEntry** itemArray,
   int                     anchorItem,
   int                     itemsToSend,
   int                     itemCount,
   int                     totalduration,
   int                     fileType,
   int                     singleSongInfo
   )
{
   int                     i;
   char*                   cp;
   struct aMessage*        msg;
   int                     ftype;
   int                     isRoot;

   msg = newMessage();
   
   appendToMsg( msg, "%s\n", XML_START );
   appendToMsg( msg, "<TiVoContainer>\n" );
   appendToMsg( msg, " <Details>\n" );
   isRoot = strcmp( dirItem->dirp->path, "/" ) == 0 &&
      strcmp( path, "/" ) == 0;
   ftype = ( isRoot ) ? dirItem->folderType : fileType;
   if( isRoot )
      cp = "tivo-server";
   else if( ftype & ( FILE_MPEG ) )
      cp = "tivo-music";
   else if( ftype & ( FILE_JPEG ) )
      cp = "tivo-server";
   else cp = "folder";
   appendToMsg( msg, "  <ContentType>x-container/%s</ContentType>\n", cp );
   appendToMsg( msg, " <SourceFormat>x-container/folder</SourceFormat>\n" );
   appendToMsg( msg, " <TotalDuration>%d</TotalDuration>\n", totalduration );
   appendToMsg( msg, " <TotalItems>%d</TotalItems>\n", itemCount - singleSongInfo);
   appendToMsg( msg, " <Title>%s</Title>\n", getHostname() );
   appendToMsg( msg, " </Details>\n" );
   appendToMsg( msg, " <ItemStart>%d</ItemStart>\n", anchorItem - singleSongInfo);
   appendToMsg( msg, " <ItemCount>%d</ItemCount>\n", itemsToSend );

   /*
    * In general we don't provide the "Details" items below;
    * e.g. they are unimplemented
    *    SourceLocation
    *    SourceSize
    *    CreationDate
    *    LastAccessDate
    *    Copyright
    *
    *    SourceBitRate        audio specific
    *    SourceSampleRate
    *    RecordLabel
    *
    *    SourceWidth          image specific
    *    SourceHeight
    *    SourceColors
    *    SourceResolution
    *
    * In general we don't provide the "Url" items below:
    * e.g. they are unimplemented
    *    AcceptsParams={yes,no}
    *
    *    LoopCount=n          audio specific
    *
    *    RefreshInterval=n    image specific
    *    ShowStyle=style
    *    ShowDelay=n
    */
   
#ifdef PLAYLIST_EDITING
   if( dirItem && IS_PLAYLIST( dirItem ) && itemsToSend ) {
	if (anchorItem == 0)	{
//		if (itemsToSend != itemCount)
			itemsToSend--;

      		if ( strstr(dirItem->name, ".m3u") ) {
		  appendToMsg( msg, " <Item>\n" );
		  appendToMsg( msg, "  <Details>\n" );
         	  appendToMsg( msg, "   <ContentType>x-container/folder</ContentType>\n");
         	  appendToMsg( msg, "   <SourceFormat>x-container/folder</SourceFormat>\n" );
		  cp = strrchr (path, '/');
         	  appendToMsg( msg, "   <Title>Modify Playlist - '%s'</Title>\n", ++cp);
         	  appendToMsg( msg, "  </Details>\n" );
         	  appendToMsg( msg, "  <Links>\n" );
         	  appendToMsg( msg, "   <Content>\n" );
         	  appendToMsg( msg, "    <Url>/TiVoConnect?Command=QueryContainer&Container=%s&PlaylistEdit=Modify", path);
         	  appendToMsg( msg, "</Url>\n" );
         	  appendToMsg( msg, "   </Content>\n" );
         	  appendToMsg( msg, "  </Links>\n" );
      		  appendToMsg( msg, " </Item>\n" );
		}
	} else {
		anchorItem--;
	}
   }
#endif


   for( i = anchorItem; i < anchorItem + itemsToSend; i++ )
   {
      addItemToMsg( msg, itemArray[ i ], isRoot, dirItem, path );
   }
   appendToMsg( msg, "</TiVoContainer>\n" );
   appendToMsg( msg, "%s\n", XML_END );

   sendXmlMessage( msg, fd );

   addToLog( "Directory contents, length=%d:\n%s", msg->size, msg->msg );

   freeMessage( msg );
}

/*
 * Generate and send the description of a specific directory.
 */
void queryContainer(
   int                  fd,
   struct aParameterList* params
   )
{
   int                  i;
   char*                cp;
   struct aDirectoryEntry* topItem;
   struct aDirectoryEntry* nextItem;
   struct aDirectoryEntry* dirItem;
   struct aDirectoryEntry** itemArray;
   int                  itemCount = 0;
   int                  folder;
   int                  totalduration;
   int                  anchorItem;
   char*                anchorItemName;
   char*                anchorItemEntry;
   char*                anchorItemBase;
   char*                base;
   int                  itemsToSend;
   char*                newpath = NULL;
   char*                path;
   int                  fileType;
   int                  singleSongInfo;

   path = findParameter( params, "Container" );
   if( path == NULL )
      path = "/";
   path = decodeString( path );
   newpath = isValidFileRef( path );
   if( newpath == NULL )
   {
      write( fd, "404 file not found\r\n", 20 );
      free( path );
      freeParameters( params );

      close( fd );

      return;
   }

   addToLog( "Sending directory %s\n", path );

   topItem = getFilteredDirectory( path, params );
   if( !topItem )
   {
      addToLog( "Couldn't find directory %s\n", path );

      write( fd, "404 file not found\r\n", 20 );
      free( path );
      if( newpath )
         free( newpath );
      freeParameters( params );

      close( fd );

      return;
   }
   itemCount = 0;
   totalduration = 0;
   fileType = 0;
   for( nextItem = topItem; nextItem; nextItem = nextItem->next )
   {
      if( IS_MPEG( nextItem ) )
         getMp3Info( nextItem->dirp, nextItem );
      itemCount++;
      totalduration += nextItem->duration;

      fileType |= nextItem->fileType;
   }

   dirItem = getFileFromDirCache( path );
   if( dirItem == NULL )
   {
      dirItem = getFileFromDirCache( topItem->dirp->path );
   }

   if ( dirItem->helperCmd )
   {
      doHelperCmd( dirItem );
   }      

   if( !IS_PLAYLIST( dirItem ) )
       itemArray = sortDirectoryContents( topItem, topItem->dirp, &itemCount, params );
    else
   {
      itemCount = 0;
      for( nextItem = topItem; nextItem != NULL; nextItem = nextItem->next )
         itemCount++;
      itemArray = ( struct aDirectoryEntry** ) malloc( itemCount *
         sizeof( *nextItem ) );
      itemCount = 0;
      for( nextItem = topItem; nextItem != NULL; nextItem = nextItem->next )
         itemArray[ itemCount++ ] = nextItem;
#ifdef PLAYLIST_EDITING
      itemCount++;
#endif
   }

   itemsToSend = itemCount;
   cp = findParameter( params, "ItemCount" );
   if( cp != NULL )
   {
      itemsToSend = atoi( cp );
   }
   anchorItem = 0;
   anchorItemName = findParameter( params, "AnchorItem" );
#ifdef PLAYLIST_EDITING
   if (anchorItemName != NULL && strstr (anchorItemName, "PlaylistEdit"))
      anchorItemName = NULL;
#endif
   if( anchorItemName != NULL )
   {
      anchorItemName = decodeString( anchorItemName );
      cp = decodeString( anchorItemName );
      free( anchorItemName );
      anchorItemName = cp;
      cp = decodeString( anchorItemName );
      free( anchorItemName );
      anchorItemName = cp;

      if ( !IS_PLAYLIST(dirItem) )
      {
	 anchorItemEntry = dirLeafName( anchorItemName );
	 base = dirBaseName( anchorItemName );
	 cp = dirLeafName( base );
	 free( base );
	 anchorItemBase = cp;
      }
      else
      {
	 int  offset = 0;

	 anchorItemEntry = malloc( strlen(anchorItemName) );
	 anchorItemBase = NULL;

	 if ( strncmp( "/TiVoConnect", anchorItemName, strlen("/TiVoConnect") ) == 0 )
	    offset = strlen( "/TiVoConnect" );

         if( isInternetRef( &anchorItemName[ offset + 1 ] ) )
            offset++;

	 strcpy( anchorItemEntry, anchorItemName + offset );
      }

      for( i = 0; i < itemCount; i++ )
      {
	 /*addToLog( "DEC: %d %s %s %s ", IS_PLAYLIST(dirItem), anchorItemName, anchorItemEntry, itemArray[i]->name );*/
         if( strcmp( anchorItemEntry, itemArray[ i ]->name ) == 0 )
         {
            if( IS_FOLDER( itemArray[ i ] ) )
               break;

	    if ( IS_PLAYLIST(dirItem) )
	       break;

            base = dirLeafName( itemArray[ i ]->dirp->path );
            if( strcmp( anchorItemBase, base ) == 0 )
            {
               free( base );
               break;
            }
         }
      }
      free( anchorItemEntry );
      if ( anchorItemBase )
	 free( anchorItemBase );

      if( i < itemCount )
      {
         anchorItem = i;
         if( itemsToSend > 0 )
            anchorItem++;
      }
      else
         addToLog( "Failed to find anchor %s\n", anchorItemName );
#ifdef PLAYLIST_EDITING
      if (IS_PLAYLIST(dirItem))
         anchorItem++;
#endif
      if( getDebugMode() )
         addToLog( "Find anchor %s at %d\n", anchorItemName, anchorItem );
      free( anchorItemName );
   }
   cp = findParameter( params, "AnchorOffset" );
   if( cp != NULL )
   {
      anchorItem += atoi( cp );
      if( anchorItem < 0 )
         anchorItem = 0;
   }
   if( itemsToSend < 0 )
   {
      i = anchorItem + itemsToSend;
      if( i < 0 )
         i = 0;
      anchorItem = i;
      itemsToSend = -itemsToSend;
   }
   if( anchorItem + itemsToSend > itemCount )
      itemsToSend -= ( ( anchorItem + itemsToSend ) - itemCount );
   if( itemsToSend < 0 )
      itemsToSend = 0;

   singleSongInfo = 0;
#ifdef PLAYLIST_EDITING
   if (IS_PLAYLIST(dirItem)) {
         cp = findParameter (params, "Recurse");
         if (cp && !strcasecmp (cp, "Yes") && itemsToSend == 1)
		singleSongInfo = 1;
   }
#endif
   cp = decodeString( findParameter( params, "Format" ) );
   if( cp == NULL || strcasecmp( cp, "text/xml" ) == 0 )
   {
      sendContainerAsXML( fd, path, dirItem, itemArray, anchorItem, itemsToSend,
         itemCount, totalduration, fileType, singleSongInfo );
   }
   else if( strcasecmp( cp, "text/html" ) == 0 )
   {
      sendContainerAsHTML( fd, path, dirItem, itemArray, anchorItem, itemsToSend,
         itemCount, totalduration, fileType );
   }
   else
   {
      write( fd, "404 file not found\r\n", 20 );
   }
   free( cp );

   free( path );
   if( newpath )
      free( newpath );
   free( itemArray );
   freeParameters( params );

   close( fd );
}

/*
 * Return information about this server
 */
void queryServer(
   int                  fd,
   struct aParameterList* params
   )
{
   struct aMessage*     msg;

   msg = newMessage();

   appendToMsg( msg, "%s\n", XML_START );
   appendToMsg( msg, "<TiVoServer>\n" );
   appendToMsg( msg, " <Version>1</Version>\n" );
   appendToMsg( msg, " <InternalName>%s</InternalName>\n", SERVER_NAME );
   appendToMsg( msg, " <InternalVersion>%s</InternalVersion>\n", SERVER_VERSION );
   appendToMsg( msg, " <Organization>%s</Organization>\n", SERVER_ORGANIZATION );
   appendToMsg( msg, " <Comment>%s</Comment>\n", SERVER_COMMENT );
   appendToMsg( msg, "</TiVoServer>\n" );
   appendToMsg( msg, "%s\n", XML_END );

   sendXmlMessage( msg, fd );
   
   freeMessage( msg );
   freeParameters( params );

   close( fd );
}

/*
 * Generate and send the description of a specific item.
 *
 * uses parameter URL; identifies what item to return
 *
 * Should return output very similarto QueryContainer, except that
 * enclosing term is TiVoItem.
 */
void queryItem(
   int                  fd,
   struct aParameterList* params
   )
{
   char*                url;
   char*                newurl;
   char*                cp;
   struct aDirectoryEntry *item;
   struct aDirectoryEntry *dirItem;
   struct aMessage*     msg;
   
   url = findParameter( params, "Url" );
   if( url == NULL )
   {
      write( fd, "404 No url parameter found.\r\n", 29 );
      close( fd );
      freeParameters( params );
      
      return;
   }
   url = decodeString( url );
   newurl = decodeString( url );
   free( url );
   url = newurl;
   newurl = decodeString( url );
   free( url );
   url = newurl;

   cp = url;
   if( strncasecmp( cp, "/TiVoConnect", 12 ) == 0 )
      cp = &cp[ 12 ];
/*
   if( strncasecmp( cp, "http://", 7 ) == 0 )
   {
      cp = &cp[ 7 ];
      while( *cp && *cp != '/' )
         cp++;
   }
   if( strncasecmp( cp, "/TiVoConnect/", 13 ) == 0 )
      cp = &cp[ 13 ];
*/
   if( getDebugMode() )
   {
      addToLog( "Try to find item for url %s\n", url );
   }

   item = getUrlEntry( cp );
   if( !item && *cp == '/' )
      item = getUrlEntry( &cp[ 1 ] );
   if( !item )
      item = getFileFromDirCache( cp );
   if( !item )
   {
      addToLog( "Can't find url for QueryItem; url=%s\n", url );
      write( fd, "404 Can't find url.\r\n", 21 );
      free( url );
      close( fd );
      freeParameters( params );
      
      return;
   }

   msg = newMessage();
   
   appendToMsg( msg, "%s\n", XML_START );
   appendToMsg( msg, "<TiVoItem>\n" );
   dirItem = getFileFromDirCache( item->dirp->path );
   addItemToMsg( msg, item, 0, dirItem, cp );
   appendToMsg( msg, "</TiVoItem>\n" );

   sendXmlMessage( msg, fd );

   addToLog( "Item Details, length=%d:\n%s", msg->size, msg->msg );

   freeMessage( msg );

   /*sendFile( fd, cp, params );*/

   freeParameters( params );

   close( fd );
}

/*
 * Reset the server after a connection; has no meaning for us
 */
void resetServer(
   int                  fd,
   struct aParameterList* params
   )
{
   freeParameters( params );
   close( fd );
}

/*
 * Unimplemented command: QueryFormats
 *    uses parameter SourceFormat: what format to query
 *       parameter can include wildcard; audio/*
 *
 * Output format:
 *    <TiVoFormats> ... </TiVoFormats>
 *       <Format> ... </Format>
 *          <Description> text </Description>
 *          <ContentType> mimetype </ContentType>
 */

/*
 * Special type to package file sending info for threads
 */
struct aTask
{
   int                  fd;
   char*                path;
   struct aParameterList* params;
};

/*
 * Create a task object
 */
static struct aTask* newTask(
   int                  fd,
   char*                path,
   struct aParameterList* params
   )
{
   struct aTask*        task;

   task = ( struct aTask* ) malloc( sizeof( *task ) );
   task->fd = fd;
   task->params = copyParameters( params );
   task->path = ( char* ) malloc( strlen( path ) + 1 );
   strcpy( task->path, path );

   return task;
}

/*
 * Free a task object
 */
static void freeTask(
   struct aTask*        task
   )
{
   free( task->path );
   if( task->fd != -1 )
      close( task->fd );
   freeParameters( task->params );
   free( task );
}

/*
 * Wrapper for sendMPEGFile
 */
static void doSendMPEGFile(
   struct aTask*        task
   )
{
   sigset_t             threadMask;

   sigemptyset( &threadMask );
   sigaddset( &threadMask, SIGHUP );
   sigaddset( &threadMask, SIGINT );
   sigaddset( &threadMask, SIGQUIT );
   sigaddset( &threadMask, SIGTERM );
   pthread_sigmask( SIG_BLOCK, &threadMask, NULL );

   sendMPEGFile( task->fd, task->path, task->params );

   freeTask( task );
}

/*
 * Wrapper for sendJPEGFile
 */
static void doSendJPEGFile(
   struct aTask*        task
   )
{
   sigset_t             threadMask;

   sigemptyset( &threadMask );
   sigaddset( &threadMask, SIGHUP );
   sigaddset( &threadMask, SIGINT );
   sigaddset( &threadMask, SIGQUIT );
   sigaddset( &threadMask, SIGTERM );
   pthread_sigmask( SIG_BLOCK, &threadMask, NULL );

   sendJPEGFile( task->fd, task->path, task->params );

   freeTask( task );
}

/*
 * Wrapper for sendCGI
 */
static void doSendCGIImage(
   struct aTask*        task
   )
{
   sigset_t             threadMask;

   sigemptyset( &threadMask );
   sigaddset( &threadMask, SIGHUP );
   sigaddset( &threadMask, SIGINT );
   sigaddset( &threadMask, SIGQUIT );
   sigaddset( &threadMask, SIGTERM );
   pthread_sigmask( SIG_BLOCK, &threadMask, NULL );

   sendCGIImage( task->fd, task->path, task->params );

   freeTask( task );
}

/*
 * Wrapper for sendStreamFile
 */
static void doSendStreamFile(
   struct aTask*        task
   )
{
   sigset_t             threadMask;

   sigemptyset( &threadMask );
   sigaddset( &threadMask, SIGHUP );
   sigaddset( &threadMask, SIGINT );
   sigaddset( &threadMask, SIGQUIT );
   sigaddset( &threadMask, SIGTERM );
   pthread_sigmask( SIG_BLOCK, &threadMask, NULL );

   sendStreamFile( task->fd, task->path, task->params, 1, 0 );

   task->fd = -1;

   freeTask( task );
}

/*
 * Sends a container description or a file to the client
 *
 * Unimplemented parameters:
 *    PixelShape=m:n       for photos
 */
void sendFile(
   int                  fd,
   char*                path,
   struct aParameterList* params
   )
{
   int                  i;
   char*                cp;
   char*                newpath = NULL;
   struct stat          statbuf;
   struct aDirectory*   dirp;
   struct aDirectoryEntry* items;
   int                  internetRef;
#if USE_THREADS
   struct aTask*        task;
#endif

   path = decodeString( path );
   cp = decodeString( path );
   free( path );
   path = cp;

   if( !path || strlen( path  ) <= 1 )
   {
      queryContainer( fd, params );
      free( path );

      return;
   }

   addToLog( "Path:\n%s\n", path );

   newpath = isValidFileRef( path );
   if( newpath == NULL )
   {
      addToLog( "Invalid path in sendFile: %s\n", path );
      write( fd, "404 file not found\r\n", 20 );
      free( path );

      close( fd );

      return;
   }

   i = stat( path, &statbuf );
   if( i >= 0 && S_ISDIR( statbuf.st_mode ) )
   {
      free( path );
      if( newpath )
         free( newpath );
      queryContainer( fd, params );
      return;
   }

   internetRef = isInternetRef( path );
   if( internetRef && path[ 0 ] == '/' )
      items = getFileFromDirCache( &path[ 1 ] );
   else
      items = getFileFromDirCache( path );
   if( items == NULL  && !internetRef )
   {
      addToLog( "File not found: %s\n", path );
      write( fd, "404 file not found\r\n", 20 );
      free( path );
      if( newpath )
         free( newpath );

      close( fd );

      return;
   }

   if ( items && items->helperCmd )
   {
      doHelperCmd( items );
   }      

#if USE_THREADS
   task = newTask( fd, path, params );
   if( internetRef )
      newThread( (void*(*)(void*))doSendStreamFile, task );
   else if( items )
   {
      if( IS_MPEG( items ) )
         newThread( (void*(*)(void*))doSendMPEGFile, task );
      else if( IS_JPEG( items ) )
         newThread( (void*(*)(void*))doSendJPEGFile, task );
      else if( IS_CGI( items ) )
         newThread( (void*(*)(void*))doSendCGIImage, task );
      else if( IS_STREAM( items ) )
         newThread( (void*(*)(void*))doSendStreamFile, task );
      else
      {
         addToLog( "Unknown filetype in sendFile: %d\n", items->fileType );
         freeTask( task );
         close( fd );
      }
   }
   else
   {
      addToLog( "Unknown path in sendFile: %s\n", path );
      freeTask( task );
      close( fd );
   }
#else
   if( internetRef )
   {
      sendStreamFile( fd, path, params, 1, 0 );
      fd = -1;
   }
   else
   {
      if( IS_MPEG( items ) )
         sendMPEGFile( fd, path, params );
      else if( IS_JPEG( items ) )
         sendJPEGFile( fd, path, params );
      else if( IS_STREAM( items ) )
      {
         sendStreamFile( fd, path, params, 1, 0 );
         fd = -1;
      }
      else
      {
         addToLog( "Unknown filetype in sendFile: %d\n", items->fileType );
      }
   }
   else
   {
      addToLog( "Unknown path in sendFile: %s\n", path );
   }
   freeParameters( params );
   if( fd != -1 )
      close( fd );
#endif
   free( path );
   if( newpath )
      free( newpath );
}
