/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * Dave Clemans, April 2003
 *
 * Copyright (C) 2003  Dave Clemans
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * See the file "COPYING" for more details.
 */

#include "globals.h"

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <netdb.h>
#include <sys/ioctl.h>

/*
 * Types of configuration file lines...
 */
#define  CFG_LOGFILE    0
#define  CFG_PORT       1
#ifdef IGNORE_FILE
#define  CFG_IGNORE     2
#endif
#define  CFG_THUMBDIR   3
#define  CFG_MPEG       4
#define  CFG_JPEG       5
#define  CFG_STREAM     6
#define  CFG_PLAYLIST   7
#define  CFG_SCRIPT     8
#define  CFG_INTERNET   9
#define  CFG_CACHE      10
#define  CFG_VSOUND     11
#define  CFG_SPLAY      12
#define  CFG_AUDIO      13
#define  CFG_IMAGE      14
#ifdef XMRADIO
#define  CFG_XMPCR      15
#endif /* XMRADIO */
#define  CFG_VIRTUAL    16
#define  CFG_NETWORK    17
#ifdef PHOTO_ORGANIZER
#define  CFG_PHOTO_ORGANIZER  18
#endif /* PHOTO_ORGANIZER */
#define  CFG_TRANSCODE  19
#define  CFG_RECORDER   20

struct configMap
{
   char* keyword;
   int   keytype;
};

static struct configMap configCmds[] =
{
   { "logfile:",        CFG_LOGFILE    },
   { "port:",           CFG_PORT       },
#ifdef IGNORE_FILE
   { "ignore:",         CFG_IGNORE     },
#endif
   { "thumbdir:",       CFG_THUMBDIR   },
   { "mpeg:",           CFG_MPEG       },
   { "jpeg:",           CFG_JPEG       },
   { "stream:",         CFG_STREAM     },
   { "playlist:",       CFG_PLAYLIST   },
   { "script:",         CFG_SCRIPT     },
   { "internet:",       CFG_INTERNET   },
   { "cache:",          CFG_CACHE      },
   { "vsound:",         CFG_VSOUND     },
   { "splay:",          CFG_SPLAY      },
   { "audio:",          CFG_AUDIO      },
   { "image:",          CFG_IMAGE      },
#ifdef XMRADIO
   { "xmpcr:",          CFG_XMPCR      },
#endif /* XMRADIO */
   { "virtual:",        CFG_VIRTUAL    },
   { "network:",        CFG_NETWORK    },
   { "byartist:",       CFG_VIRTUAL    },
   { "bygenre:",        CFG_VIRTUAL    },
#ifdef PHOTO_ORGANIZER
   { "photo_organizer:", CFG_PHOTO_ORGANIZER },
#endif /* PHOTO_ORGANIZER */
   { "transcode:",      CFG_TRANSCODE  },
   { "recorder:",       CFG_RECORDER   },
   { NULL,              -1             }
};

/*
 * Current configuration "database"
 */
static int currentDebugMode = 0;
static char* currentConfigFile = NULL;
static char* currentLogfile = NULL;
static char* currentThumbCacheDir = NULL;
static struct aFileType* currentMpegList = NULL;
static struct aFileType* currentJpegList = NULL;
static struct aFileType* currentStreamList = NULL;
static struct aFileType* currentPlayList = NULL;
static struct aFileType* currentScriptList = NULL;
static struct aFileType* currentInternetList = NULL;
static struct aDirectoryEntry* currentRootdir = NULL;
static char* currentCacheSize = NULL;
static struct aDirectoryEntry* currentVirtualDirectories = NULL;
#ifdef PHOTO_ORGANIZER
static struct aDirectoryEntry* currentPhotoOrganizerDirectories = NULL;
#endif /* PHOTO_ORGANIZER */
#ifdef XMRADIO
static char* currentXMRadioPath = NULL;
static char* currentXMRadioControlPath = NULL;
#endif /* XMRADIO */
static char* currentVirtualDirName = NULL;
static uint32_t bcastAddress = INADDR_BROADCAST;
static char *ifname = "eth0";
static int currentDataPort = 2190;

#define max_interfaces 8

struct iface_struct
{
	char name[128];
	struct in_addr ip;
	struct in_addr netmask;
};
static struct iface_struct ifaces[ max_interfaces ];

static void freeFileTypes( struct aFileType** );
static void addFileType( struct aFileType**, char* );
static int getInterfaces( void );

/*
 * Adds a new entry to our "root" directory
 *
 * Note: we assume all root directory entries are folders...
 */
static void newRootDirEntry(
   char*                   name,
   char*                   description,
   int                     foldertype
   )
{
   struct aDirectoryEntry* item;
   char*                   newname = NULL;
   int                     newctr = 0;
   int                     len;

   do
   {
      for( item = currentRootdir; item; item = item->next )
      {
         if( strcasecmp( name, item->name ) == 0 )
            break;
      }
      if( item )
      {
         /* Are we trying to add a duplicate name? */
         len = strlen( name );
         if( !newctr )
            newname = ( char* ) malloc( len + 16 );
         else
            newname = ( char* ) malloc( len + 1 );
         sprintf( newname, "%s (%d)", name, ++newctr );
         if( newctr != 1 )
            free( name );
         name = newname;
      }
   } while( item );

   item = newDirectoryEntry();
   item->folderType = foldertype;
   item->fileType = FILE_FOLDER;
   item->description = strdup( description );
   item->name = strdup( name );
   addToLog( "New rootdir entry: %s/*,%s,%s\n",
      ( foldertype == FILE_JPEG ) ? "image" : "audio",
      name, description );

   item->next = currentRootdir;
   currentRootdir = item;

   if( newname )
      free( newname );
}

/*
 * Get the next word from a configuration file line
 *
 * Normally words are separated by white space...
 * You can put spaces in a word by surrounding it with double quotes
 */
char* getNextWord(
   char*                   startPtr,
   char**                  endPtr
   )
{
   char*                   scp;
   char*                   ecp;

   scp = startPtr;
   while( *scp && isspace( *scp ) )
      scp++;
   if( *scp == '"' )
   {
      scp++;
      ecp = scp;
      while( *ecp && *ecp != '"' )
         ecp++;
      if( *ecp )
         *ecp++ = '\0';
   }
   else
   {
      ecp = scp;
      while( *ecp && !isspace( *ecp ) )
         ecp++;
      if( isspace( *ecp ) )
         *ecp++ = '\0';
   }
   *endPtr = ecp;

   return scp;
}

/*
 * Read in our configuration file, replacing any information from
 * previous calls.
 */
void readConfigFile(
   char*                configFile
   )
{
   FILE*                filep;
   char*                cp;
   char*                scp;
   char*                tokPtr;
   char*                filen;
   struct aDirectoryEntry* item;
   char                 buffer[ 1024 ];
   int                  total;
   int                  i;
   int                  cmd;
   struct in_addr       addr;

   addToLog( "Read configuration file: %s\n", configFile );

   filep = fopen( configFile, "r" );
   if( filep == NULL )
   {
      addToLog( "Unable to read configuration file: %s\n", configFile );
      return;
   }

   if( currentConfigFile != NULL && configFile != currentConfigFile )
      free( currentConfigFile );
   if( currentConfigFile != configFile )
   {
      currentConfigFile = ( char* ) malloc( strlen( configFile ) + 1 );
      strcpy( currentConfigFile, configFile );
   }

   freeFileTypes( &currentMpegList );
   freeFileTypes( &currentJpegList );
   freeFileTypes( &currentStreamList );
   freeFileTypes( &currentPlayList );
   freeFileTypes( &currentInternetList );
   if( currentVirtualDirectories )
   {
      freeDirectoryEntries( currentVirtualDirectories );
      currentVirtualDirectories = NULL;
   }
   if( currentVirtualDirName )
   {
      free( currentVirtualDirName );
      currentVirtualDirName = NULL;
   }

#ifdef XMRADIO
   if( currentXMRadioPath || currentXMRadioControlPath )
      shutdownXMRadio();
   if( currentXMRadioPath )
   {
      free( currentXMRadioPath );
      currentXMRadioPath = NULL;
   }
   if( currentXMRadioControlPath )
   {
      free( currentXMRadioControlPath );
      currentXMRadioControlPath = NULL;
   }
#endif /* XMRADIO */
   
   bcastAddress = INADDR_BROADCAST;

   if( currentRootdir )
   {
      if( currentRootdir->dirp != NULL )
         freeDirectory( currentRootdir->dirp );
      /* freeDirectoryEntry( currentRootdir ); not needed; done above */
      currentRootdir = NULL;
   }

   transcodeMethodDestroyAll();

   while( fgets( buffer, sizeof buffer, filep ) != NULL )
   {
      for( cp = buffer; *cp; cp++ )
      {
         if( *cp == '\n' )
            *cp = '\0';
      }
      if( strlen( buffer ) == 0 )
         continue;
      if( buffer[ 0 ] == '#' )
         continue;
      cp = getNextWord( buffer, &tokPtr );
      for( cmd = 0; configCmds[ cmd ].keyword; cmd++ )
      {
         if( strcasecmp( cp, configCmds[ cmd ].keyword ) == 0 )
            break;
      }
      if( !configCmds[ cmd ].keyword )
      {
         addToLog( "Unable to understand config line: %s %s\n", cp, tokPtr );
         continue;
      }
      switch( configCmds[ cmd ].keytype )
      {
         case CFG_LOGFILE:
            cp = getNextWord( tokPtr, &tokPtr );
            if( *cp )
               setLogfile( cp );
            else
               addToLog( "Bad config file line: %s\n", buffer );
            break;

         case CFG_PORT:
            cp = getNextWord( tokPtr, &tokPtr );
            if( *cp )
               setDataPort( atoi( cp ) );
            else
               addToLog( "Bad config file line: %s\n", buffer );
            break;

#ifdef IGNORE_FILE
         case CFG_IGNORE:
            /* Check for a regular expression of files to ignore */
            cp = getNextWord( tokPtr, &tokPtr );
            if( *cp )
            {
               initIgnore( cp );
            }
            break;
#endif /* IGNORE_FILE */

         case CFG_THUMBDIR:
            cp = getNextWord( tokPtr, &tokPtr );
            if( *cp )
            {
               struct stat statbuf;
               int   rc;

               rc = stat( cp, &statbuf );
               if ( rc < 0 )
               {
                  addToLog( "Bad cache directory %s \n", cp );
               }
               else
               {
                  char t[128];

                  setThumbCacheDir( cp );
                  sprintf( t, "rm -rf %s/*", cp );
                  system( t );
               }   
            }
            break;

         case CFG_MPEG:
            cp = getNextWord( tokPtr, &tokPtr );
            while( *cp )
            {
               addFileType( &currentMpegList, cp );
               cp = getNextWord( tokPtr, &tokPtr );
            }
            break;

         case CFG_JPEG:
            cp = getNextWord( tokPtr, &tokPtr );
            while( *cp )
            {
               addFileType( &currentJpegList, cp );
               cp = getNextWord( tokPtr, &tokPtr );
            }
            break;

         case CFG_STREAM:
            cp = getNextWord( tokPtr, &tokPtr );
            while( *cp )
            {
               addFileType( &currentStreamList, cp );
               cp = getNextWord( tokPtr, &tokPtr );
            }
            break;

         case CFG_PLAYLIST:
            cp = getNextWord( tokPtr, &tokPtr );
            while( *cp )
            {
               addFileType( &currentPlayList, cp );
               cp = getNextWord( tokPtr, &tokPtr );
            }
            break;

         case CFG_SCRIPT:
            cp = getNextWord( tokPtr, &tokPtr );
            while( *cp )
            {
               addFileType( &currentScriptList, cp );
               cp = getNextWord( tokPtr, &tokPtr );
            }
            break;

         case CFG_INTERNET:
            cp = getNextWord( tokPtr, &tokPtr );
            while( *cp )
            {
               addFileType( &currentInternetList, cp );
               cp = getNextWord( tokPtr, &tokPtr );
            }
            break;

         case CFG_CACHE:
            cp = getNextWord( tokPtr, &tokPtr );
            if( *cp )
            {
               if( currentCacheSize != NULL )
                  free( currentCacheSize );
               currentCacheSize = ( char* ) malloc( strlen( cp ) + 1 );
               strcpy( currentCacheSize, cp );
            }
            break;

         case CFG_VSOUND:
            addToLog( "Configuration option 'vsound:' no longer used.\n" );
            break;

         case CFG_SPLAY:
            addToLog( "Configuration option 'splay:' no longer used.\n" );
            break;

         case CFG_AUDIO:
            scp = getNextWord( tokPtr, &tokPtr );
            cp = getNextWord( tokPtr, &tokPtr );
            newRootDirEntry( cp, scp, FILE_MPEG );
            break;

         case CFG_IMAGE:
            scp = getNextWord( tokPtr, &tokPtr );
            cp = getNextWord( tokPtr, &tokPtr );
            newRootDirEntry( cp, scp, FILE_JPEG );
            break;

#ifdef XMRADIO
         case CFG_XMPCR:
            scp = getNextWord( tokPtr, &tokPtr );
            cp = getNextWord( tokPtr, &tokPtr );
   
            setXMRadioControl( scp );
            setXMRadio( cp );
   
            if( !initXMRadio() )
            {
               addToLog( "Unable to get XM Radio Channel info; is XMDaemon running?\n" );
               continue;
            }
   
            addToLog( "Accessing XM Satellite Radio stream via %s, %s\n",
               getXMRadioControl(), getXMRadio() );

            newRootDirEntry( XMPCR_NAME, XMPCR_NAME, FILE_MPEG | FILE_XMDIR );
            startupXMThread();
            break;
#endif /* XMRADIO */

#ifdef PHOTO_ORGANIZER
         case CFG_PHOTO_ORGANIZER:
            addToLog( "Enabling support for virtual directories; year,month,.. etc.\n" );
   
	    currentPhotoOrganizerDirectories = NULL;

            cp = getNextWord( tokPtr, &tokPtr );
            setPhotoOrganizerDirName( cp );
   
            cp = getNextWord( tokPtr, &tokPtr );
            while( *cp )
            {
               for( item = currentPhotoOrganizerDirectories; item; item = item->next )
               {
                  if( strcasecmp( cp, item->name ) == 0 )
                     break;
               }

               if( !item )
               {
                  item = newDirectoryEntry();
                  item->folderType = FILE_JPEG;
                  item->fileType = FILE_FOLDER;
                  item->description = strdup( cp );
                  item->name = strdup( cp );
                  item->next = currentPhotoOrganizerDirectories;
                  currentPhotoOrganizerDirectories = item;
               }
               cp = getNextWord( tokPtr, &tokPtr );
            }

            newRootDirEntry( getPhotoOrganizerDirName(), getPhotoOrganizerDirName(), FILE_JPEG | FILE_XMDIR );
            break;
#endif /* PHOTO_ORGANIZER */

         case CFG_VIRTUAL:
            addToLog( "Enabling support for virtual directories; artist, genre, etc.\n" );
   
            currentVirtualDirectories = NULL;

            cp = getNextWord( tokPtr, &tokPtr );
            setVirtualDirName( cp );
   
            cp = getNextWord( tokPtr, &tokPtr );
            while( *cp )
            {
               for( item = currentVirtualDirectories; item; item = item->next )
               {
                  if( strcasecmp( cp, item->name ) == 0 )
                     break;
               }
               if( !item )
               {
                  item = newDirectoryEntry();
                  item->folderType = FILE_MPEG;
                  item->fileType = FILE_FOLDER;
                  item->description = strdup( cp );
                  item->name = strdup( cp );
                  item->next = currentVirtualDirectories;
                  currentVirtualDirectories = item;
               }
               cp = getNextWord( tokPtr, &tokPtr );
            }

            newRootDirEntry( getVirtualDirName(), getVirtualDirName(), FILE_MPEG | FILE_XMDIR );
            break;

         case CFG_NETWORK:
            cp = getNextWord( tokPtr, &tokPtr );
            if( *cp )
            {
#if 0
               total = getInterfaces();
               for( i = 0; i < total; i++ )
               {
                  if( strcasecmp( cp, ifaces[ i ].name ) == 0 )
                     break;
                  if( inet_aton( cp, &addr ) != 0 )
                  {
                     if( ( addr.s_addr & ifaces[ i ].netmask.s_addr ) ==
                        ( ifaces[ i ].ip.s_addr & ifaces[ i ].netmask.s_addr ) )
                     {
                        break;
                     }
                  }
               }
               if( i < total )
               {
                  addToLog( "Broadcasting on interface: %s\n",
                     ifaces[ i ].name );
                  bcastAddress = ntohl( ifaces[ i ].ip.s_addr );
                  addToLog( "Broadcasting on interface: %x\n", bcastAddress );
               }
               else
               {
                  addToLog( "Unable to find interface: %s\n", cp );
               }
#else
               ifname = malloc( strlen(cp) + 1 );
               strcpy( ifname, cp );
#endif
            }
            break;

         case CFG_TRANSCODE:
            scp = getNextWord( tokPtr, &tokPtr );
            while( cp = getNextWord( tokPtr, &tokPtr ) )
            {
               if( !*cp )
                  break;
               ( void ) transcodeMethodCreate( scp, cp );
            }
            break;

         case CFG_RECORDER:
            scp = getNextWord( tokPtr, &tokPtr );
            cp = getNextWord( tokPtr, &tokPtr );
            filen = getNextWord( tokPtr, &tokPtr );
            newRootDirEntry( cp, scp, FILE_MPEG );
            recorderInit( scp, cp, filen, getNextWord( tokPtr, &tokPtr ) );
            break;

         default:
            addToLog( "Unable to understand config line: %s %s\n", cp, buffer );
            break;
      }
   }
   fclose( filep );

   if( currentRootdir->dirp == NULL )
   {
      currentRootdir->dirp = getFromDirCache( "/" );
      if( currentRootdir->dirp != NULL )
      {
         freeDirectoryEntries( currentRootdir->dirp->items );
      }
      else
      {
         currentRootdir->dirp = newDirectory();
         currentRootdir->dirp->path = ( char* ) malloc( 2 );
         strcpy( currentRootdir->dirp->path, "/" );
         addDirToDirCache( currentRootdir->dirp );
      }
      currentRootdir->dirp->items = currentRootdir;

      for( item = currentRootdir->next; item != NULL; item = item->next )
         item->dirp = currentRootdir->dirp;
   }
}

/*
 * Is debug mode on?
 */
int getDebugMode( void )
{
   return currentDebugMode;
}

/*
 * Change debug mode
 */
void setDebugMode(
   int                  flag
   )
{
   currentDebugMode = flag;
}

/*
 * Current config file name?
 */
char* getConfigFile( void )
{
   return currentConfigFile;
}

/*
 * Get directories to virtualize...
 */
struct aDirectoryEntry* getVirtualDirectories( void )
{
   return currentVirtualDirectories ;
}

#ifdef XMRADIO
/*
 * Save path to XM PCR Satellite Radio dsp file
 */
void setXMRadio(
   char*                file
   )
{
   if( currentXMRadioPath != NULL )
      free( currentXMRadioPath );
   currentXMRadioPath = ( char* ) malloc( strlen( file ) + 1 );
   strcpy( currentXMRadioPath, file );
}

/*
 * Current XM Satellite Radio dsp name
 */
char* getXMRadio( void )
{
   return currentXMRadioPath;
}
#endif /* XMRADIO */

/*
 * Save name of virtual music directory
 */
void setVirtualDirName(
   char*                file
   )
{
   if( currentVirtualDirName != NULL )
      free( currentVirtualDirName );
   currentVirtualDirName = strdup( file );
}

/*
 * Current virtual music directory name
 */
char* getVirtualDirName( void )
{
   if( !currentVirtualDirName )
      return "*No Virtual Directory*";

   return currentVirtualDirName;
}

#ifdef PHOTO_ORGANIZER

char * photoOrganizerDirName = NULL;

void setPhotoOrganizerDirName( char *file )
{
   if ( !photoOrganizerDirName )
      free( photoOrganizerDirName );

   photoOrganizerDirName = strdup( file );
}

char *getPhotoOrganizerDirName( void )
{
   return photoOrganizerDirName;
}

/*
 * Get directories to virtualize...
 */
struct aDirectoryEntry* getPhotoOrganizerDirectories( void )
{
   return currentPhotoOrganizerDirectories;
}
#endif /* PHOTO_ORGANIZER */

#ifdef XMRADIO
/*
 * Save path to XM PCR Satellite Radio control file
 */
void setXMRadioControl(
   char*                file
   )
{
   if( currentXMRadioControlPath != NULL )
      free( currentXMRadioControlPath );
   currentXMRadioControlPath = ( char* ) malloc( strlen( file ) + 1 );
   strcpy( currentXMRadioControlPath, file );
}

/*
 * Current XM Satellite Radio control name
 */
char* getXMRadioControl( void )
{
   return currentXMRadioControlPath;
}
#endif /* XMRADIO */

/*
 * Change our current logfile name
 */
void setLogfile(
   char*                file
   )
{
   if( currentLogfile != NULL )
      free( currentLogfile );
   currentLogfile = ( char* ) malloc( strlen( file ) + 1 );
   strcpy( currentLogfile, file );
}

/*
 * Get our current logfile name
 */
char* getLogfile( void )
{
   return currentLogfile;
}

/*
 * Store which dataport we'll use
 */
void setDataPort(
   int                  port
   )
{
   currentDataPort = port;
}

/*
 * Get our current dataport
 */
int getDataPort( void )
{
   return currentDataPort;
}

/*
 * Free a filetype list
 */
static void freeFileTypes(
   struct aFileType**   types
   )
{
   struct aFileType*    ft;
   int                  i;

   if( types == NULL )
      return;
   ft = *types;
   if( ft == NULL )
      return;
   for( i = 0; i < ft->size; i++ )
      free( ft->types[ i ] );
   free( ft->types );
   free( ft );
   *types = NULL;
}

/*
 * Add an entry to a filetype list
 */
static void addFileType(
   struct aFileType**   types,
   char*                name
   )
{
   struct aFileType*    ft;
   struct aFileType*    newft;
   int                  i;

   if( types == NULL )
      return;
   ft = *types;
   if( ft == NULL )
   {
      ft = ( struct aFileType* ) malloc( sizeof( *ft ) );
      ft->size = 1;
      ft->types = ( char** ) malloc( sizeof( char* ) );
      ft->types[ 0 ] = ( char* ) malloc( strlen( name ) + 1 );
      strcpy( ft->types[ 0 ], name );
      *types = ft;
      return;
   }

   newft = ( struct aFileType* ) malloc( sizeof( *newft ) );
   newft->size = ft->size + 1;
   newft->types = ( char** ) malloc( sizeof( char* ) * newft->size );
   memcpy( newft->types, ft->types, sizeof( char* ) * ft->size );
   newft->types[ ft->size ] = ( char* ) malloc( strlen( name ) + 1 );
   strcpy( newft->types[ ft->size ], name );
   free( ft->types );
   free( ft );
   *types = newft;
}

/*
 * A particular mpeg file entry
 */
char* getMpegEntry(
   int                  i
   )
{
   if( currentMpegList == NULL )
      return NULL;
   if( i >= currentMpegList->size )
      return NULL;
   return currentMpegList->types[ i ];
}

/*
 * A particular jpeg file entry
 */
char* getJpegEntry(
   int                  i
   )
{
   if( currentJpegList == NULL )
      return NULL;
   if( i >= currentJpegList->size )
      return NULL;
   return currentJpegList->types[ i ];
}

/*
 * A particular stream file entry
 */
char* getStreamEntry(
   int                  i
   )
{
   if( currentStreamList == NULL )
      return NULL;
   if( i >= currentStreamList->size )
      return NULL;
   return currentStreamList->types[ i ];
}

/*
 * A particular ...
 */
char* getScriptEntry(
   int                  i
   )
{
   if( currentScriptList == NULL )
      return NULL;
   if( i >= currentScriptList->size )
      return NULL;
   return currentScriptList->types[ i ];
}

/*
 * A particular playlist file entry
 */
char* getPlayEntry(
   int                  i
   )
{
   if( currentPlayList == NULL )
      return NULL;
   if( i >= currentPlayList->size )
      return NULL;
   return currentPlayList->types[ i ];
}

/*
 * A particular internet file entry
 */
char* getInternetEntry(
   int                  i
   )
{
   if( currentInternetList == NULL )
      return NULL;
   if( i >= currentInternetList->size )
      return NULL;
   return currentInternetList->types[ i ];
}

/*
 * Return the current rootdirectory
 */
struct aDirectoryEntry* getRootdir( void )
{
   return currentRootdir;
}

/*
 * Path to mogrify
 */
char* getConvertPath( void )
{
   static char*         path = NULL;

   if( path == NULL )
      path = getToolPath( "convert" );

   return path;
}

/*
 * Returns the current cache size
 */
char* getCacheSize( void )
{
   if( currentCacheSize == NULL )
      return "16";
   return currentCacheSize;
}

/*
 * Get a list of network interfaces on this system; stored in file static "ifaces"
 *    based on similar code from samba
 */
static int getInterfaces( void )
{
   struct ifconf        ifc;
   char                 buff[8192];
   int                  fd;
   int                  i;
   int                  n;
   struct ifreq         *ifr=NULL;
   int                  total = 0;
   struct in_addr       ipaddr;
   struct in_addr       nmask;
   char                 *iname;
   uint32_t             mask;
   uint32_t             ip;

	if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
		return -1;
	}
  
	ifc.ifc_len = sizeof(buff);
	ifc.ifc_buf = buff;

	if (ioctl(fd, SIOCGIFCONF, &ifc) != 0) {
		close(fd);
		return -1;
	} 

	ifr = ifc.ifc_req;
  
	n = ifc.ifc_len / sizeof(struct ifreq);

	/* Loop through interfaces, looking for given IP address */
	for (i=n-1;i>=0 && total < max_interfaces;i--)
   {
		if (ioctl(fd, SIOCGIFADDR, &ifr[i]) != 0)
      {
			continue;
		}

		iname = ifr[i].ifr_name;
		ipaddr = (*(struct sockaddr_in *)&ifr[i].ifr_addr).sin_addr;

		if (ioctl(fd, SIOCGIFFLAGS, &ifr[i]) != 0)
      {
			continue;
		}  

		if (!(ifr[i].ifr_flags & IFF_UP))
      {
			continue;
		}

		if (ioctl(fd, SIOCGIFNETMASK, &ifr[i]) != 0)
      {
			continue;
		}  

		nmask = ((struct sockaddr_in *)&ifr[i].ifr_addr)->sin_addr;

		strncpy(ifaces[total].name, iname, sizeof(ifaces[total].name)-1);
		ifaces[total].name[sizeof(ifaces[total].name)-1] = 0;
		ifaces[total].ip = ipaddr;
		ifaces[total].netmask = nmask;
		total++;
	}

	close(fd);

	for (i=0; i<total; i++)
   {
		ip = ntohl(ifaces[i].ip.s_addr);
		mask = ntohl(ifaces[i].netmask.s_addr);
		addToLog("Network interface %s: ip=%d.%d.%d.%d, mask=%d.%d.%d.%d\n",
			ifaces[i].name,
			(ip >> 24)&0xff, (ip >> 16)&0xff,
			(ip >> 8)&0xff, (ip)&0xff,
			(mask >> 24)&0xff, (mask >> 16)&0xff,
			(mask >> 8)&0xff, (mask)&0xff);
	}
   
   return total;
}

/*
 * Returns the interface broadcast address to be used for beacons
 */
uint32_t getBcastAddress( void )
{
   struct ifreq        ifr;
   struct sockaddr_in *sin;
   int                 s, rval;
   static int          ifacePrinted = 0;

   s = socket( AF_INET, SOCK_DGRAM, 0 );
   if ( s < 0 )
   {
      return INADDR_BROADCAST;
   }

   strcpy( ifr.ifr_name, ifname );
   rval = ioctl( s, SIOCGIFBRDADDR, &ifr );
   if ( rval < 0 )
   {
      close(s);
      return INADDR_BROADCAST;
   }

   sin = (struct sockaddr_in *)&ifr.ifr_broadaddr;
   if( !ifacePrinted || getDebugMode() )
   {
      addToLog( "Interface: %s broadcast addr %s \n", ifname, inet_ntoa(sin->sin_addr) );
      ifacePrinted = 1;
   }
   
   close(s);

   return ntohl((uint32_t)(sin->sin_addr.s_addr));
}

/*
 * Returns the current thumbnail image cache directory
 */
char * getThumbCacheDir( void )
{
   return currentThumbCacheDir;
}

/*
 * Set the current thumbnail image cache directory
 */
void setThumbCacheDir( char* dir )
{
   if( currentThumbCacheDir != NULL )
      free( currentThumbCacheDir );

   currentThumbCacheDir = ( char* ) malloc( strlen( dir ) + 1 );
   strcpy( currentThumbCacheDir, dir );
}

char *getIFName (void)
{
	return (ifname);
}
