/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * Dave Clemans, April 2003
 *
 * Copyright (C) 2003  Dave Clemans
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * See the file "COPYING" for more details.
 */

#include "globals.h"

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <time.h>

/*
 * Compare two directory items
 */
static int cmpItemsBySortList(
   const struct aDirectoryEntry** i1,
   const struct aDirectoryEntry** i2
   )
{
   struct aDirectory*      dirp;
   int                     i;
   int                     rc;

   dirp = (*i1)->dirp;

   for( i = 0; i < dirp->sortCount; i++ )
   {
      switch( dirp->sortList[ i ] )
      {
         case SORT_TYPE:
            if( IS_FOLDER( *i1 ) && !IS_FOLDER( *i2 ) )
               return -1;
            if( !IS_FOLDER( *i1 ) && IS_FOLDER( *i2 ) )
               return 1;
            if( IS_PLAYLIST( *i1 ) && !IS_PLAYLIST( *i2 ) )
               return -1;
            if( !IS_PLAYLIST( *i1 ) && IS_PLAYLIST( *i2 ) )
               return 1;
            break;
         case -SORT_TYPE:
            if( IS_FOLDER( *i1 ) && !IS_FOLDER( *i2 ) )
               return 1;
            if( !IS_FOLDER( *i1 ) && IS_FOLDER( *i2 ) )
               return -1;
            if( IS_PLAYLIST( *i1 ) && !IS_PLAYLIST( *i2 ) )
               return 1;
            if( !IS_PLAYLIST( *i1 ) && IS_PLAYLIST( *i2 ) )
               return -1;
            break;
         case SORT_TITLE:
            rc = strcmp( ( (*i1)->description ) ? (*i1)->description : (*i1)->name,
               ( (*i2)->description ) ? (*i2)->description : (*i2)->name );
            if( rc != 0 )
               return rc;
            break;
         case -SORT_TITLE:
            rc = strcmp( ( (*i2)->description ) ? (*i2)->description : (*i2)->name,
               ( (*i1)->description ) ? (*i1)->description : (*i1)->name );
            if( rc != 0 )
               return rc;
            break;
         case SORT_CREATED:
            if( (*i1)->created < (*i2)->created )
               return -1;
            if( (*i1)->created > (*i2)->created )
               return 1;
            break;
         case -SORT_CREATED:
            if( (*i1)->created < (*i2)->created )
               return 1;
            if( (*i1)->created > (*i2)->created )
               return -1;
            break;
         case SORT_MODIFIED:
            if( (*i1)->modified < (*i2)->modified )
               return -1;
            if( (*i1)->modified > (*i2)->modified )
               return 1;
            break;
         case -SORT_MODIFIED:
            if( (*i1)->modified < (*i2)->modified )
               return 1;
            if( (*i1)->modified > (*i2)->modified )
               return -1;
            break;
         case SORT_RANDOM:
            if( (*i1)->randomValue < (*i2)->randomValue )
               return -1;
            if( (*i1)->randomValue > (*i2)->randomValue )
               return 1;
            break;
         case -SORT_RANDOM:
            if( (*i1)->randomValue < (*i2)->randomValue )
               return 1;
            if( (*i1)->randomValue > (*i2)->randomValue )
               return -1;
            break;
      }
   }

   return 0;
}

/*
 * Return a sorted array of directory items for this directory
 */
struct aDirectoryEntry** sortDirectoryContents(
   struct aDirectoryEntry* items,
   struct aDirectory*      dirp,
   int*                    size,
   struct aParameterList*  params
   )
{
   struct aDirectoryEntry** itemArray;
   struct aDirectoryEntry* nextItem;
   struct aDirectoryEntry* np;
   int                     itemCount;
   int                     i;
   char*                   cp;
   char*                   scp;
   int                     sortList[ 16 ];
   int                     sortCount;

   cp = findParameter( params, "RandomSeed" );
   if( cp != NULL )
      i = atoi( cp );
   else
      i = 0;
   srandom( i );

   itemCount = 0;
   for( nextItem = items; nextItem != NULL; nextItem = nextItem->next )
   {
      nextItem->randomValue = random();
      itemCount++;
   }

   cp = findParameter( params, "RandomStart" );
   if( cp != NULL )
   {
      scp = decodeString( cp );
      if( strncasecmp( "/TiVoConnect", scp, 12 ) == 0 )
         cp = &scp[ 12 ];
      else
         cp = scp;
      np = getFileFromDirCache( cp );
      if( np != NULL )
      {
         for( nextItem = items; nextItem != NULL; nextItem = nextItem->next )
         {
            if( strcmp( nextItem->dirp->path, np->dirp->path ) == 0 &&
               strcmp( nextItem->name, np->name ) == 0 )
            {
               nextItem->randomValue = -1;
               break;
            }
         }
      }

      free( scp );
   }

   itemArray = ( struct aDirectoryEntry** ) malloc( itemCount * sizeof( nextItem ) );
   nextItem = items;
   for( i = 0; i < itemCount; i++ )
   {
      itemArray[ i ] = nextItem;
      nextItem = nextItem->next;
   }

   scp = findParameter( params, "SortOrder" );
   if( scp == NULL )
   {
      sortList[ 0 ] = SORT_TYPE;
      sortList[ 1 ] = SORT_TITLE;
      sortCount = 2;
   }
   else
   {
      sortCount = 0;
      do
      {
         cp = scp;
         while( *cp && *cp != ',' )
            cp++;
         if( *cp == ',' )
            *cp++ = '\0';
         i = 1;
         if( *scp == '!' )
               i = -1;
         if( strcasecmp( scp, "Type" ) == 0 )
         {
            sortList[ sortCount++ ] = SORT_TYPE * i;
         }
         else if( strcasecmp( scp, "Title" ) == 0 )
         {
            sortList[ sortCount++ ] = SORT_TITLE * i;
         }
         else if( strcasecmp( scp, "CreationDate" ) == 0 )
         {
            sortList[ sortCount++ ] = SORT_CREATED * i;
         }
         else if( strcasecmp( scp, "LastChangeDate" ) == 0 )
         {
            sortList[ sortCount++ ] = SORT_MODIFIED * i;
         }
         else if( strcasecmp( scp, "Random" ) == 0 )
         {
            sortList[ sortCount++ ] = SORT_RANDOM * i;
         }
         scp = cp;
         if( sortCount > 16 )
            break;
      } while( *cp );
   }

   dirp->sortList = sortList;
   dirp->sortCount = sortCount;

   for( nextItem = items; nextItem != NULL; nextItem = nextItem->next )
   {
      nextItem->dirp->sortList = sortList;
      nextItem->dirp->sortCount = sortCount;
   }

   qsort( itemArray, itemCount, sizeof( nextItem ),
      ( int(*)(const void*, const void*) ) cmpItemsBySortList );

   if( size != NULL )
      *size = itemCount;
   return itemArray;
}

/*
 * Return the "file type" of this file
 */
int guessFileType(
   char*                name
   )
{
   int                  i;
   char*                suffix;

   if( isInternetRef( name ) )
      return FILE_STREAM;

   for( suffix = name; *suffix; suffix++ )
      /* do nothing */;
   while( suffix > name && *suffix != '.' )
      suffix--;
   if( *suffix == '.' )
      suffix++;

   for( i = 0; getMpegEntry( i ) != NULL; i++ )
   {
      if( strcasecmp( suffix, getMpegEntry( i ) ) == 0 )
         return FILE_MPEG;
   }
   for( i = 0; getJpegEntry( i ) != NULL; i++ )
   {
      if( strcasecmp( suffix, getJpegEntry( i ) ) == 0 )
         return FILE_JPEG;
   }
   for( i = 0; getStreamEntry( i ) != NULL; i++ )
   {
      if( strcasecmp( suffix, getStreamEntry( i ) ) == 0 )
         return FILE_STREAM;
   }
   for( i = 0; getPlayEntry( i ) != NULL; i++ )
   {
      if( strcasecmp( suffix, getPlayEntry( i ) ) == 0 )
         return FILE_PLAYLIST;
   }
   for( i = 0; getScriptEntry( i ) != NULL; i++ )
   {
      if( strcasecmp( suffix, getScriptEntry( i ) ) == 0 )
         return FILE_CGI;
   }
   
   return 0;
}

/*
 * Get which filetypes correspond to the user filter
 */
int getFiletypeFilter(
   struct aParameterList* params
   )
{
   int                  filter;
   char*                cp;
   char*                scp;
   char*                filterString;

   filter = FILE_FOLDER | FILE_MPEG | FILE_JPEG |
      FILE_PLAYLIST | FILE_STREAM | FILE_CGI;
   cp = findParameter( params, "Filter" );
   if( cp != NULL )
   {
      filterString = decodeString( cp );
      cp = filterString;
      filter = 0;
      scp = cp;
      do
      {
         while( *cp && *cp != ',' )
            cp++;
         if( *cp == ',' )
            *cp++ = '\0';
         if( strcasecmp( scp, "x-container/folder" ) == 0 )
            filter |= FILE_FOLDER;
         else if( strcasecmp( scp, "audio/*" ) == 0 )
            filter |= ( FILE_MPEG | FILE_STREAM | FILE_CGI );
         else if( strcasecmp( scp, "image/*" ) == 0 )
            filter |= FILE_JPEG | FILE_CGI;
         else if( strcasecmp( scp, "x-container/playlist" ) == 0 )
            filter |= FILE_PLAYLIST;
         scp = cp;
      } while( *cp );
      free( filterString );
   }

   return filter;
}

/*
 * Get a filtered copy of the contents of a directory
 */
struct aDirectoryEntry* getFilteredDirectory(
   char*                path,
   struct aParameterList* params
   )
{
   struct aDirectory*   dirp;
   struct aDirectoryEntry* top = NULL;
   struct aDirectoryEntry* bottom = NULL;
   struct aDirectoryEntry* src;
   struct aDirectoryEntry* dest;
   struct aDirectoryEntry* dp;
   struct aDirectoryEntry* sp;
   char*                cp;
   char*                cp1;
   char*                leaf;
   char*                base;
   int                  filter;
   int                  recursion;

   char *execute = malloc (strlen (path) + 10);
   sprintf (execute, "%s%sfetch.sh", path, (path[strlen(path) - 1] == '/') ? "" : "/");
   if (!access (execute, X_OK)) {
   	addToLog( "Fetching files for directory: %s\n", path);
   	system (execute);
   }
   free (execute);

   filter = getFiletypeFilter( params );

   dirp = addToDirCache( path );
   if( !dirp )
      return NULL;

   cp = findParameter( params, "Recurse" );
   if( cp != NULL && strcasecmp( cp, "Yes" ) == 0 )
      recursion = 1;
   else
      recursion = 0;

   for( src = dirp->items; src != NULL; src = src->next )
   {
      if( recursion && IS_FOLDER( src ) )
      {
         cp = fullPathname( path, src->name );
         cp1 = decodeString( cp );
         free( cp );
         cp = cp1;
         dp = getFilteredDirectory( cp, params );
         free( cp );
         if( dp == NULL )
            continue;
         for( sp = dp; sp != NULL; sp = sp->next )
         {
            dest = copyDirectoryEntry( sp );

            if( !top )
               top = dest;
            else
               bottom->next = dest;
            bottom = dest;
         }
         continue;
      }

      if( !( src->fileType & filter ) )
         continue;

      if (!strcmp (src->name, "fetch.sh"))
	 continue;

      if( IS_MPEG( src ) && !src->infoValid )
         getMp3Info( src->dirp, src );
      dest = copyDirectoryEntry( src );

      if( !top )
         top = dest;
      else
         bottom->next = dest;
      bottom = dest;
   }

   return top;
}
