/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * Dave Clemans, April 2003
 *
 * Copyright (C) 2003  Dave Clemans
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * See the file "COPYING" for more details.
 */

#include "globals.h"

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#define __USE_GNU
#ifdef NEED_HSEARCH_R
 #include "hsearch_r.h"
#else
 #include <search.h>
#endif
#include <time.h>
#include <libgen.h>

static int dirCacheInited = 0;
static struct hsearch_data hashTable;

static int urlTableInited = 0;
static struct hsearch_data urlTable;

#if USE_THREADS
static pthread_mutex_t dirCacheMutex = PTHREAD_MUTEX_INITIALIZER;
#endif

/*
 * Create a directory object
 */
struct aDirectory* newDirectory( void )
{
   struct aDirectory*      dirp;

   dirp = ( struct aDirectory* ) malloc( sizeof( *dirp ) );
   if ( dirp )
   {
      bzero( dirp, sizeof(struct aDirectory) );
   }

   return dirp;
}

/*
 * Free a directory object
 */
void freeDirectory(
   struct aDirectory*      dirp
   )
{
   if ( !dirp )
      return;

   if( dirp->path )
      free( dirp->path );
   if( dirp->items )
      freeDirectoryEntries( dirp->items );
   if( dirp->sortList )
      free( dirp->sortList );

   free( dirp );
}

/*
 * Create a directory entry
 */
struct aDirectoryEntry* newDirectoryEntry( void )
{
   struct aDirectoryEntry* item;

   item = ( struct aDirectoryEntry* ) malloc( sizeof( *item ) );
   if ( item ) 
   {
      bzero( item, sizeof(struct aDirectoryEntry) );
   }

   return item;
}

/*
 * Free a directory entry
 */
void freeDirectoryEntry(
   struct aDirectoryEntry* item
   )
{
   if ( !item )
      return ;

   if( item->name )
      free( item->name );
   if( item->description )
      free( item->description );
   if( item->album )
      free( item->album );
   if( item->artist )
      free( item->artist );
   if( item->genre )
      free( item->genre );
   if( item->title )
      free( item->title );
   if ( item->thumbCache )
      free( item->thumbCache );
   if ( item->helperCmd )
      free ( item->helperCmd );

   free( item );
}

/*
 * Free a list of directory entries
 */
void freeDirectoryEntries(
   struct aDirectoryEntry* item
   )
{
   struct aDirectoryEntry* n;
   struct aDirectoryEntry* p;

   n = item;
   while( n )
   {
      p = n;
      n = n->next;
      freeDirectoryEntry( p );
   }
}

/*
 * Make a copy of a directory entry.  Note that it won't be linked into
 * a directory
 */
struct aDirectoryEntry* copyDirectoryEntry(
   struct aDirectoryEntry* item
   )
{
   struct aDirectoryEntry* n;

   if ( !item )
      return NULL;

   n = newDirectoryEntry();
   if ( n )
   {
      memcpy( n, item, sizeof(struct aDirectoryEntry) );

      if( item->name ) 
      {
	 n->name = strdup( item->name );
      }
      if( item->description ) 
      {
	 n->description = strdup( item->description );
      }
      if( item->album ) 
      {
	 n->album = strdup( item->album );
      }
      if( item->artist )
      {
	 n->artist = strdup( item->artist );
      }
      if( item->genre )
      {
	 n->genre = strdup( item->genre );
      }
      if( item->year )
      {
         n->year = ( char* ) malloc( strlen( item->year ) + 1 );
         strcpy( n->year, item->year );
      }
      if( item->title )
      {
	 n->title = strdup( item->title );
      }
      if ( item->thumbCache )
      {
	 n->thumbCache = strdup( item->thumbCache );
      }

      n->next = NULL;
   }

   return n;
}

/*
 * Copy a list of directory entries
 */
struct aDirectoryEntry* copyDirectoryEntries(
   struct aDirectoryEntry* item
   )
{
   struct aDirectoryEntry* top = NULL;
   struct aDirectoryEntry* bottom = NULL;
   struct aDirectoryEntry* n;

   while( item )
   {
      n = copyDirectoryEntry( item );
      if( !top )
         top = n;
      else
         bottom->next = n;
      bottom = n;

      item = item->next;
   }

   return top;
}

/*
 * Create the hash table for our directory cache
 */
void initDirCache( void )
{
   if( !dirCacheInited )
   {
      dirCacheInited = 1;
      hcreate_r( HASH_SIZE, &hashTable );
   }
}

/*
 * Free a directory cache hash table
 */
void freeDirCache( void )
{
   hdestroy_r( &hashTable );
   dirCacheInited = 0;
}

/*
 * If a directory is in the cache, return its contents
 */
struct aDirectory* getFromDirCache(
   char*                   path
   )
{
   ENTRY                   item;
   ENTRY*                  hentry;
   struct aDirectory*      n;

   item.key = path;
   item.data = NULL;
#if USE_THREADS
   pthread_mutex_lock( &dirCacheMutex );
#endif
   hsearch_r( item, FIND, &hentry, &hashTable );
#if USE_THREADS
   pthread_mutex_unlock( &dirCacheMutex );
#endif
   if( hentry )
   {
      n = ( struct aDirectory* ) hentry->data;
      return n;
   }

   return NULL;
}

/*
 * Get the owner of this file or directory from the cache
 */
struct aDirectory* getOwnerFromDirCache(
   char*                   path
   )
{
   char*                   base;
   struct aDirectory*      dirp;
   char*                   newpath;
   
   newpath = decodeString( path );

#ifdef PHOTO_ORGANIZER
   if( strcmp( newpath, XMPCR_NAME ) == 0 || strcmp( newpath, getVirtualDirName() ) == 0 || strcmp( newpath, getPhotoOrganizerDirName() ) == 0 )
#else /* PHOTO_ORGANIZER */
   if( strcmp( newpath, XMPCR_NAME ) == 0 || strcmp( newpath, getVirtualDirName() ) == 0 )
#endif /* PHOTO_ORGANIZER */
      base = strdup( "/" );
   else
      base = dirBaseName( newpath );
   if( getDebugMode() )
      addToLog( "Get owner of path '%s'; base=%s\n", newpath, base );
   
   dirp = getFromDirCache( base );
   if( !dirp )
   {
      if( strcmp( base, "/" ) != 0 )
         dirp = addToDirCache( base );
   }
   free( base );
   free( newpath );
   return dirp;
}

/*
 * Get a description for a file from the contents of a directory in the cache
 */
struct aDirectoryEntry* getFileFromDirCache(
   char*                   path
   )
{
   struct aDirectory*      dirp;
   char*                   leaf;
   struct aDirectoryEntry* f;

   dirp = getOwnerFromDirCache( path );
   if( !dirp )
      return NULL;
   leaf = dirLeafName( path );
   f = findEntryInDir( dirp, leaf );
   free( leaf );

   return f;
}

/*
 * Add a directory & its contents to the cache, if it's not already there
 */
struct aDirectory* addToDirCache(
   char*                   path
   )
{
   struct aDirectory*      n;
   struct stat             statbuf;
   struct aDirectoryEntry* top = NULL;
   struct aDirectoryEntry* bottom = NULL;
   struct aDirectoryEntry* i;
   DIR*                    dp;
   struct dirent*          dentry;
   char*                   fullpath;
   ENTRY                   item;
   ENTRY*                  hentry;
   int                     inCache = 0;
   char*                   cp;
   int                     rc;

   path = decodeString( path );

   rc = stat( path, &statbuf );
#ifdef PHOTO_ORGANIZER
   if( rc < 0 &&
      strncmp( path, XMPCR_NAME, strlen( XMPCR_NAME ) ) != 0 &&
      strncmp( path, getVirtualDirName(), strlen( getVirtualDirName() ) ) != 0 &&
      strncmp( path, getPhotoOrganizerDirName(), strlen( getPhotoOrganizerDirName() ) ) != 0 )
#else /* PHOTO_ORGANIZER */
   if( rc < 0 &&
      strncmp( path, XMPCR_NAME, strlen( XMPCR_NAME ) ) != 0 &&
      strncmp( path, getVirtualDirName(), strlen( getVirtualDirName() ) ) != 0 )
#endif /* PHOTO_ORGANIZER */
   {
      free( path );
      return NULL;
   }
   n = getFromDirCache( path );
#ifdef  PHOTO_ORGANIZER
   if( n && ( statbuf.st_mtime <= n->modified ||
      strncmp( path, XMPCR_NAME, strlen( XMPCR_NAME ) ) == 0 ||
      strncmp( path, getVirtualDirName(), strlen( getVirtualDirName() ) ) == 0 ||
      strncmp( path, getPhotoOrganizerDirName(), strlen( getPhotoOrganizerDirName() ) ) == 0 ) )
#else /* PHOTO_ORGANIZER */
   if( n && ( statbuf.st_mtime <= n->modified ||
      strncmp( path, XMPCR_NAME, strlen( XMPCR_NAME ) ) == 0 ||
      strncmp( path, getVirtualDirName(), strlen( getVirtualDirName() ) ) == 0 ) )
#endif /* PHOTO_ORGANIZER */
   {
      free( path );
      return n;
   }
   else if( n )
   {
      if( strcmp( n->path, "/" ) == 0 )
         return n;
      inCache = 1;
      freeDirectoryEntries( n->items );
      n->items = NULL;
   }
   if( !n )
      n = newDirectory();
   n->modified = statbuf.st_mtime;
   n->path = strdup( path );

   if( getDebugMode() )
      addToLog( "Add directory to cache: %s\n", n->path );

#ifdef XMRADIO
   if( strcmp( XMPCR_NAME, path ) == 0 )
      top = readXMRadioCategories( n );
   else if( strncmp( XMPCR_NAME, path, strlen( XMPCR_NAME ) ) == 0 )
      top = readXMRadioStations( path, n );
   else
#endif /* XMRADIO */
     if( strcmp( getVirtualDirName(), path ) == 0 )
      top = readVirtualDirectories( n );
   else if( strncmp( getVirtualDirName(), path, strlen( getVirtualDirName() ) ) == 0 )
      top = readVirtualDirContents( path, n );
#ifdef PHOTO_ORGANIZER
   else if( strncmp( getPhotoOrganizerDirName(), path, strlen( getPhotoOrganizerDirName() ) ) == 0 )
      top = readPhotoOrganizerDirContents( path, n );
#endif /* PHOTO_ORGANIZER */
   else if( ( S_ISREG( statbuf.st_mode ) || S_ISLNK( statbuf.st_mode ) ) &&
      ( guessFileType( path ) & FILE_PLAYLIST ) )
   {
      if( getDebugMode() )
         addToLog( "Reading contents of playlist: %s\n", path );

      for( cp = path; *cp; cp++ )
         /* do nothing */;
      while( cp > path && *cp != '.' )
         cp--;
      if( strcasecmp( cp, ".pls" ) == 0 )
         top = readPLSFile( path, n );
      else if( strcasecmp( cp, ".m3u" ) == 0 )
         top = readM3UFile( path, n );
      else if( strcasecmp( cp, ".ram" ) == 0 )
         top = readRAMFile( path, n );
   }
   else
   {
      if( getDebugMode() )
         addToLog( "Reading contents of directory: %s\n", n->path );

      dp = opendir( path );
      if( !dp )
         return n;
   
      while( ( dentry = readdir( dp ) ) != NULL )
      {
         if( dentry->d_name[ 0 ] == '.' )
            continue;

#ifdef IGNORE_FILE
	 if( ignoreFile(dentry->d_name))
	   continue;
#endif /* IGNORE_FILE */
   
         cp = fullPathname( path, dentry->d_name );
         fullpath = decodeString( cp );
         free( cp );
         statbuf.st_mtime = 0;
         statbuf.st_mode = 0;
         stat( fullpath, &statbuf );
         free( fullpath );
         i = newDirectoryEntry();
         i->name = ( char* ) malloc( strlen( dentry->d_name ) + 1 );
         strcpy( i->name, dentry->d_name );
         if( S_ISDIR( statbuf.st_mode ) )
            i->fileType = FILE_FOLDER;
         else if( S_ISREG( statbuf.st_mode ) || S_ISLNK( statbuf.st_mode ) )
         {
            i->fileType = guessFileType( dentry->d_name );
         }
         i->created = statbuf.st_ctime;
         i->modified = statbuf.st_mtime;
         i->dirp = n;
         if( getDebugMode() )
	 {
	    addToLog( "New directory entry: %s\n", dentry->d_name );
	 }

	 cp = malloc( strlen(path) + strlen(dentry->d_name) + strlen( "/.on" ) + 1 );
	 sprintf( cp, "%s/.on%s", path, dentry->d_name );

	 if ( stat( cp, &statbuf ) == 0 )
	 {
	    if ( i->helperCmd )
	       free( i->helperCmd );
	    
	    i->helperCmd = cp;
	    addToLog( "Found helper command %s ", i->helperCmd );
	 }
	 else
	 {
	    free( cp );
	 }

         if( !top )
            top = i;
         else
            bottom->next = i;
         bottom = i;
      }

      closedir( dp );
   }

   n->items = top;
   if( !inCache )
   {
      if( getDebugMode() )
         addToLog( "Add to directory cache: %s\n", n->path );

      item.key = n->path;
      item.data = ( void* ) n;
#if USE_THREADS
      pthread_mutex_lock( &dirCacheMutex );
#endif
      hsearch_r( item, ENTER, &hentry, &hashTable );
#if USE_THREADS
      pthread_mutex_unlock( &dirCacheMutex );
#endif
   }
   free( path );
   return n;
}

/*
 * Add a prebuilt directory entry to our cache.  Typically used only to
 * record our virtual root dir in the cache.
 */
void addDirToDirCache(
   struct aDirectory*         dirp
   )
{
   ENTRY                   item;
   ENTRY*                  hentry;
   int                     rc;

   if( getDebugMode() )
      addToLog( "Add dirp to directory cache: %s\n", dirp->path );

   item.key = dirp->path;
   item.data = ( void* ) dirp;
#if USE_THREADS
   pthread_mutex_lock( &dirCacheMutex );
#endif
   rc = hsearch_r( item, ENTER, &hentry, &hashTable );
#if USE_THREADS
   pthread_mutex_unlock( &dirCacheMutex );
#endif
   if ( !rc )
      addToLog( "dirCache: hash table full\n");
}

/*
 * Find an entry in a directory by name
 */
struct aDirectoryEntry* findEntryInDir(
   struct aDirectory*         dirp,
   char*                      name
   )
{
   struct aDirectoryEntry*    item;
   int                        isRoot;
   char*                      newname;

   if( strcmp( dirp->path, "/" ) == 0 )
      isRoot = 1;
   else
      isRoot = 0;
   if( getDebugMode() )
      addToLog( "Search %s for %s, isRoot=%d\n", dirp->path, name, isRoot );

   if( isRoot && strcmp( name, "/" ) == 0 )
      return dirp->items;

   newname = decodeString( name );
   for( item = dirp->items; item != NULL; item = item->next )
   {
      if( strcmp( item->name, newname ) == 0 ||
         ( isRoot && item->name[ 0 ] == '/' &&
	 strcmp( &item->name[ 1 ], newname ) == 0 ) )
      {
         break;
      }
   }
   free( newname );

   return item;
}

/*
 * Reverse mapping from url to directoryEntry
 *
 * Used to aid QueryItem commands.
 */
struct aDirectoryEntry* getUrlEntry(
   char*                url
   )
{
   ENTRY                item;
   ENTRY*               hentry;

   if( !urlTableInited )
      return NULL;

   item.key = url;
   item.data = NULL;
#if USE_THREADS
   pthread_mutex_lock( &dirCacheMutex );
#endif
   hsearch_r( item, FIND, &hentry, &urlTable );
#if USE_THREADS
   pthread_mutex_unlock( &dirCacheMutex );
#endif
   if( !hentry )
      return NULL;
   return ( struct aDirectoryEntry* ) hentry->data;
}

void saveUrlEntry(
   char*                url,
   struct aDirectoryEntry* entry
   )
{
   ENTRY                item;
   ENTRY*               hentry;

   if( !urlTableInited )
   {
      urlTableInited++;
      hcreate_r( 4000, &urlTable );
   }

   if( getDebugMode() )
   {
      addToLog( "Save url mapping for %s\n", url );
   }

   item.key = strdup( url );
   item.data = entry;
#if USE_THREADS
   pthread_mutex_lock( &dirCacheMutex );
#endif
   hsearch_r( item, ENTER, &hentry, &urlTable );
#if USE_THREADS
   pthread_mutex_unlock( &dirCacheMutex );
#endif
}
