/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * Dave Clemans, April 2003
 *
 * Copyright (C) 2003  Dave Clemans
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * See the file "COPYING" for more details.
 */
#ifndef  INCLUDED_GLOBALS
#define  INCLUDED_GLOBALS

#define  USE_THREADS          1

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#if defined( USE_THREADS )
  #include <pthread.h>
#endif

#define  HTTP_HEADER          "HTTP/1.0 200 Ok\r\n"
#define  HTTP_HEADER_LEN      strlen(HTTP_HEADER)
#define  XML_HEADER           "Content-Type: text/xml\r\n"
#define  XML_HEADER_LEN       strlen(XML_HEADER)

#define  SERVER_VERSION       "0.60"
#define  SERVER_NAME          "byRequest - 'Yet Another Tivo Server'"
#define  SERVER_ORGANIZATION  "The Land Of Mystery"
#define  SERVER_COMMENT       "May 15, 2004"

#define  XML_START            "<?xml version='1.0' encoding='ISO-8859-1' ?>"
#define  XML_END              "<!-- Copyright (c) 2002 TiVo Inc.-->"

#define  SORT_NONE            0
#define  SORT_TYPE            1
#define  SORT_TITLE           2
#define  SORT_CREATED         3
#define  SORT_MODIFIED        4
#define  SORT_RANDOM          5

#define  VIRT_BYARTIST        0
#define  VIRT_BYGENRE         1
#define  VIRT_BYALBUM         2
#define  VIRT_XMROOT          16
#define  VIRT_XMCAT           17

/*
 * For describing "containers" (directories), and their contents
 */
struct aDirectoryEntry
{
   char*                name;
   char*                description;
   int                  folderType;
   int                  fileType;
#define  FILE_FOLDER          0x01
#define  FILE_MPEG            0x02
#define  FILE_JPEG            0x04
#define  FILE_PLAYLIST        0x08
#define  FILE_STREAM          0x10
#define  FILE_CGI             0x20
#define  FILE_XMDIR           0x40
#define  IS_FOLDER( i )       ( ( i )->fileType & FILE_FOLDER )
#define  IS_MPEG( i )         ( ( i )->fileType & FILE_MPEG )
#define  IS_JPEG( i )         ( ( i )->fileType & FILE_JPEG )
#define  IS_PLAYLIST( i )     ( ( i )->fileType & FILE_PLAYLIST )
#define  IS_STREAM( i )       ( ( i )->fileType & FILE_STREAM )
#define  IS_CGI( i )          ( ( i )->fileType & FILE_CGI )
#define  IS_XMDIR( i )        ( ( i )->fileType & FILE_XMDIR )
   struct aDirectory*   dirp;
   time_t               created;
   time_t               modified;
   int                  randomValue;
   char                *helperCmd;

   /* mp3 specific fields */
   int                  infoValid;
   time_t               lastInfoCheck;
   char*                album;
   char*                artist;
   char*                genre;
   char*                title;
   char*		year;
   int                  rate;
   int                  duration;

   /* jpg specific fields */
   int                  rotation;
   char                *thumbCache;
#ifdef PHOTO_ORGANIZER
   struct tm           exif_date_time;
#endif /* PHOTO_ORGANIZER */

   /* Next entry in the list */
   struct aDirectoryEntry* next;
};

struct aDirectory
{
   char*                path;
   struct aDirectoryEntry* items;
   time_t               modified;
   int*                 sortList;
   int                  sortCount;
};

/*
 * For describing parsed parameters from command lines
 */
struct aParameter
{
   char*                name;
   char*                value;
   struct aParameter*   next;
};
struct aParameterList
{
   struct aParameter*   top;
   struct aParameter*   bottom;
};

/*
 * Used to hold text/xml messages as they are built
 */
struct aMessage
{
   char*                msg;
   int                  length;
   char*                ptr;
   int                  size;
};

/*
 * A saved copy of a beacon from another tivo or another server
 */
struct aBeacon
{
   time_t               lastSeen;
   char*                machine;
   char*                identity;
   char*                platform;
   char*                swversion;
   char*                method;
   char*                services;
   struct aBeacon*      next;
};

/*
 * A list of filetypes
 */
struct aFileType
{
   int                  size;
   char**               types;
};

/*
 * The data structures supporting the stream object.
 */
struct anAudioStreamClient
{
   int                     fd;
   int                     duration;
   int                     open;
   time_t                  endingTime;
};

struct anAudioStream
{
   char*                   url;
   int                     isXM;
   int                     fd;
   int                     numClients;
   struct anAudioStreamClient* clients;

   struct anAudioStream*   next;
};

/*
 * The data structure supporting transcoding of streams.
 *
 * Basically, you define a command line that takes as input the desired
 * format (wma, ogg, etc.) and produces a "wav" format on its output.
 */
struct aTranscodingMethod
{
   char*                   cmdLine;
   char*                   cmdLineCopy;
   char*		   fileType;
   char**                  parsedCmd;

   struct aTranscodingMethod* next;
};

/*
 * The data structure supporting the recorder entry.
 */

struct aRecorder
{
   char*		name;
   char*		directory;
   char*		filename;
   char*		containerSize;
};


/*
 * Globally visible functions
 */

/*
 * Ignore files stuff
 */
#define IGNORE_FILE /* turn it on */
extern void initIgnore( char* );
extern int  ignoreFile( char* );
extern char* getIgnoreString (void);

/*
 * For manipulating containers (directories)
 */
extern struct aDirectory* newDirectory( void );
extern void freeDirectory( struct aDirectory* );
extern struct aDirectoryEntry* newDirectoryEntry( void );
extern void freeDirectoryEntry( struct aDirectoryEntry* );
extern void freeDirectoryEntries( struct aDirectoryEntry* );
extern struct aDirectoryEntry* copyDirectoryEntry( struct aDirectoryEntry* );
extern struct aDirectoryEntry* copyDirectoryEntries( struct aDirectoryEntry* );
extern void initDirCache( void );
extern void freeDirCache( void );
extern struct aDirectory* getFromDirCache( char* );
extern struct aDirectory* getOwnerFromDirCache( char* );
extern struct aDirectoryEntry* getFileFromDirCache( char* );
extern struct aDirectory* addToDirCache( char* );
extern struct aDirectoryEntry* findEntryInDir( struct aDirectory*, char* );
extern void addDirToDirCache( struct aDirectory* );
extern int getFiletypeFilter( struct aParameterList* );
extern struct aDirectoryEntry* getFilteredDirectory(
   char*,
   struct aParameterList*
   );
extern struct aDirectoryEntry** sortDirectoryContents(
   struct aDirectoryEntry*,
   struct aDirectory*,
   int*,
   struct aParameterList*
   );
extern struct aDirectoryEntry* getDirectory(
   char*,
   struct aParameterList*
   );
extern struct aDirectoryEntry* getDirectoryEntry( char* );
extern int guessFileType( char* );
extern struct aDirectoryEntry* getUrlEntry( char* );
extern void saveUrlEntry( char*, struct aDirectoryEntry* );

/*
 * For manipulating "virtual" folders
 */
#define VIRTUAL_BYALBUM_NAME     "By album..."
#define VIRTUAL_BYARTIST_NAME    "By artist..."
#define VIRTUAL_BYGENRE_NAME     "By genre..."
extern struct aDirectory* getVirtualDirectory( int, char* );
extern struct aDirectory* getVirtualSubDirectory(
   struct aDirectoryEntry*
   );
extern struct aDirectoryEntry* readVirtualDirectories( struct aDirectory* );
extern struct aDirectoryEntry* readVirtualDirContents( char*, struct aDirectory* );

/*
 * For handling client commands...
 */
extern void processDatagram( int );
extern void processCommand( int );
extern void queryRootContainer( int, struct aParameterList * );
extern void queryContainer( int, struct aParameterList* );
extern void queryServer( int, struct aParameterList* );
extern void queryItem( int, struct aParameterList* );
extern void resetServer( int, struct aParameterList* );
extern void sendFile( int, char*, struct aParameterList* );
extern void sendBeaconMessage( int, struct sockaddr_in*, int, int );
extern int rcvBeaconMessage( char* );

/*
 * For handling audio files...
 */
extern void getMp3Info(
   struct aDirectory*,
   struct aDirectoryEntry*
   );
extern void sendMPEGFile(
   int,
   char*,
   struct aParameterList*
   );
extern void sendStreamFile(
   int,
   char*,
   struct aParameterList*,
   int,
   int
   );

/*
 * For handling image files...
 */
extern void sendJPEGFile(
   int,
   char*,
   struct aParameterList*
   );
extern void sendCGIImage(
   int,
   char*,
   struct aParameterList*
   );

/*
 * For handling playlists
 */
extern struct aDirectoryEntry* readM3UFile( char*, struct aDirectory* );
extern struct aDirectoryEntry* readPLSFile( char*, struct aDirectory* );
extern struct aDirectoryEntry* readRAMFile( char*, struct aDirectory* );

/*
 * For handling configuration info
 */
extern char* getNextWord( char*, char** );
extern void readConfigFile( char* );
extern int getDebugMode( void );
extern void setDebugMode( int );
extern char* getConfigFile( void );
extern void setLogfile( char* );
extern char* getLogfile( void );
extern char* getMpegEntry( int );
extern char* getJpegEntry( int );
extern char* getStreamEntry( int );
extern char* getScriptEntry( int );
extern char* getPlayEntry( int );
extern char* getInternetEntry( int );
extern struct aDirectoryEntry* getRootdir( void );
extern char* getConvertPath( void );
extern char* getCacheSize( void );
extern uint32_t getBcastAddress( void );
extern char * getThumbCacheDir( void );
extern void setThumbCacheDir( char* dir );
extern void setXMRadio( char* );
extern char* getXMRadio( void );
extern void setXMRadioControl( char* );
extern char* getXMRadioControl( void );
extern void setVirtualDirName( char* );
extern char* getVirtualDirName( void );
extern void setDataPort( int );
extern int getDataPort( void );
extern struct aDirectoryEntry* getVirtualDirectories( void );
extern char* getIFName (void);

#ifdef PHOTO_ORGANIZER
extern void setPhotoOrganizerDirName( char * );
extern char *getPhotoOrganizerDirName( void );
extern struct aDirectoryEntry* getPhotoOrganizerDirectories( void );
extern struct aDirectoryEntry* readPhotoOrganizerDirContents( char*, struct aDirectory* );
#endif


/*
 * XM Satellite Radio functions
 */
#define  XMPCR_NAME        "XM Satellite Radio"
#define  XMPCR_SOCK        3877
extern int initXMRadio( void );
extern int changeXMRadioChannel( struct aDirectoryEntry* );
extern struct aDirectoryEntry* readXMRadioCategories( struct aDirectory* );
extern struct aDirectoryEntry* readXMRadioStations( char*, struct aDirectory* );   
extern void shutdownXMRadio( void );
extern int grabXMRadioStream( void );
extern void releaseXMRadioStream( void );
extern void startupXMThread( void );

/*
 * Audio stream functions
 */
struct anAudioStream* audioStreamCreateStream( void );
void audioStreamDestroyStream( struct anAudioStream* );
void audioStreamSetUrl( struct anAudioStream*, char* );
char* audioStreamGetUrl( struct anAudioStream* );
void audioStreamSetFd( struct anAudioStream*, int );
int audioStreamGetFd( struct anAudioStream* );
void audioStreamAddClient( struct anAudioStream*, int, int, int, int );
void audioStreamRemoveClient( struct anAudioStream*, int );
struct anAudioStream* audioStreamFindStream( char* );
void audioStreamProcessStream( struct anAudioStream*, int );

struct aTranscodingMethod* getTranscodingMethods (void);
struct aTranscodingMethod* transcodeMethodCreate( char*, char* );
void transcodeMethodDestroy( struct aTranscodingMethod* );
void transcodeMethodDestroyAll( void );
struct aTranscodingMethod* transcodeMethodFind( char* );
int transcodeMethodStart( struct aTranscodingMethod*, char*, int, char*, int );

/*
 * Recorder functions 
 */
void recorderInit( char*, char*, char*, char* );
void recorderUpdateEvents( void );
void recorderStartThread( void );
void recorderThread( void );
struct aRecorder* getCurrentRecorder (void);

/*
 * Utilities
 */
extern void storeParameter(
   struct aParameterList*,
   char*,
   char*
   );
extern char* findParameter(
   struct aParameterList*,
   char*
   );
extern void freeParameters( struct aParameterList* );
extern struct aParameterList* copyParameters( struct aParameterList* );
extern char* encodeString( char* );
extern char* decodeString( char* );
extern char* fullPathname( char*, char* );
extern struct aMessage* newMessage( void );
extern void freeMessage( struct aMessage* );
extern void appendToMsg( struct aMessage*, char*, ... );
extern void sendXmlMessage( struct aMessage*, int );
extern void addToLog( char*, ... );
extern char* getHostname( void );
extern long getHostid( void );
extern char* getHostaddress( void );
#if USE_THREADS
extern pthread_t newThread( void*(*)(void*), void* );
#endif
extern int isInternetRef( char* );
extern char* isValidFileRef( char* );
extern char* getToolPath( char* );
extern char* getLibPath( char* );
extern char* dirLeafName( char* );
extern char* dirBaseName( char* );
extern void doHelperCmd( struct aDirectoryEntry * );

extern ssize_t writen( int fd, const void *vptr, size_t n );
extern ssize_t readn( int fd, void *vptr, size_t n );

extern void pauseFor( int, int );

#ifdef PLAYLIST_EDITING
 extern void createPeerTempFile (int fd, char *path);
 extern void deletePeerTempFile (int fd);
 extern int modifyPlaylist( int, struct aParameterList*, char * );
#endif

#endif
