/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * Dave Clemans, April 2003
 *
 * Copyright (C) 2003  Dave Clemans
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * See the file "COPYING" for more details.
 */

#include "globals.h"

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include <sys/poll.h>

void sendThumbnail( int fd, char *path, struct aDirectoryEntry *item, int width, int height, int rotation )
{
   int                  rc, i;
   int                  ifd = -1;
   struct pollfd        pollfds[ 1 ];
   char                 buffer[1024];
   char                 buf1[ 32 ], buf2[ 32 ], buf3[ 128 ];
   struct stat          statbuf;


   addToLog( "Caching %s", path );
   if ( item == NULL )
   {
      addToLog( "sendThumbnail: null item \n" );
      write( fd, "404 file not found\r\n", 20 );
      return;
   }


   if( stat( path, &statbuf ) < 0 )
   {
      addToLog( "Unable to find image file %s: %s\n", path );
      write( fd, "404 file not found\r\n", 20 );
      return;
   }
   
   if ( (item->thumbCache == NULL) || rotation || (statbuf.st_mtime > item->modified) )
   {

      if ( item->thumbCache == NULL )
      {
         item->thumbCache = malloc( strlen( getThumbCacheDir() ) + 8 );
         sprintf( item->thumbCache, "%s/XXXXXX", getThumbCacheDir() );

         ifd = mkstemp( item->thumbCache );
         if ( ifd < 0 )
         {
            free(item->thumbCache);
            item->thumbCache = NULL;
            
            addToLog( "Unable to create temp file for image conversion: %s\n",
		      strerror_r( errno, buffer, sizeof buffer ) );
            write( fd, "404 file not found\r\n", 20 );
            return;
         }
      }

      item->modified = statbuf.st_mtime;

      rc = fork();
      if( rc < 0 )
      {
         addToLog( "Unable to create fork for image conversion: %s\n", 
		   strerror_r( errno, buffer, sizeof buffer ) );
         write( fd, "404 file not found\r\n", 20 );
         return;
      }

      if( rc == 0 )
      {

         sprintf( buf1, "%dx%d", width, height );
         sprintf( buf2, "%d", ( item->rotation + rotation ) % 360 );
         sprintf( buf3, "jpeg:%s", item->thumbCache );
         
         if( ( ( item->rotation + rotation ) % 360 ) != 0 )
            execlp( getConvertPath(), "convert", "-size", buf1, "-resize", buf1,
                    "-rotate", buf2, path, buf3, NULL );
         else
            execlp( getConvertPath(), "convert", "-size", buf1, "-resize", buf1,
                    path, buf3, NULL );

         addToLog( "Unable to exec %s: %s\n", getConvertPath(),
                   strerror_r( errno, buffer, sizeof buffer ) );
         exit( -1 );
      }

      wait( NULL );
   }

   if( stat( item->thumbCache, &statbuf ) < 0 )
   {
      addToLog( "Unable to find thumb cache file %s: %s\n", item->thumbCache );
      write( fd, "404 file not found\r\n", 20 );
      return;
   }

   if ( ifd < 0 )
   {
      ifd = open( item->thumbCache, O_RDONLY );
      if ( ifd < 0 )
      {
	 addToLog( "Unable to open thumb cache file: %s\n", item->thumbCache );
	 write( fd, "404 file not found\r\n", 20 );
	 return;
      }
   }
   
   addToLog( "Sending thumbnail image file %s", path );

   write( fd, HTTP_HEADER, HTTP_HEADER_LEN );
   write( fd, "Content-Type: image/jpeg\r\n\r\n", 28 );

   sendfile( fd, ifd, 0, statbuf.st_size , NULL, NULL, 0);

   close( ifd );
   item->rotation = ( rotation + item->rotation ) % 360;
}

/*
 * Send a image file to the client
 */
void sendJPEGFile(
   int                  fd,
   char*                path,
   struct aParameterList* params
   )
{
   int                  i;
   char*                cp;
   struct stat          statbuf;
   struct pollfd        pollfds[ 2 ];
   int                  pipefds[ 2 ];
   int                  rc;
   int                  width;
   int                  height;
   int                  rotation;
   int                  x;
   int                  y;
   char*                tokptr;
   float                pixelShape;
   struct aDirectoryEntry* item;
   char                 buffer[ 1024 ];
   char                 buf1[ 32 ];
   char                 buf2[ 32 ];
   int                  pid;
   struct sigaction     pipeSig;
   int tread=0,twritten=0;

   if( stat( path, &statbuf ) < 0 )
   {
      addToLog( "Unable to find image file %s: %s\n", path,
         strerror_r( errno, buffer, sizeof buffer ) );
      write( fd, "404 file not found\r\n", 20 );
      return;
   }
   item = getFileFromDirCache( path );
   if( item == NULL )
   {
      addToLog( "Unable to find entry for image file %s\n", path );
      write( fd, "404 file not found\r\n", 20 );
      return;
   }

   width = 0;
   cp = findParameter( params, "Width" );
   if( cp != NULL )
      width = atoi( cp );
   height = 0;
   cp = findParameter( params, "Height" );
   if( cp != NULL )
      height = atoi( cp );
   rotation = 0;
   cp = findParameter( params, "Rotation" );
   if( cp != NULL )
      rotation = atoi( cp );
   pixelShape = 1.0;
   cp = findParameter( params, "PixelShape" );
   if( cp != NULL )
   {
      cp = strtok_r( cp, ":", &tokptr );
      x = atoi( cp );
      cp = strtok_r( cp, ":", &tokptr );
      y = atoi( cp );
      pixelShape = ( ( double ) x ) / ( ( double ) y );
      if( y > x )
         width /= pixelShape;
      else if( x > y )
         height *= pixelShape;
   }

   if ( getThumbCacheDir() != NULL )
   {
      /* force a new thumbnail if the rotation changes */
      if ( item->thumbCache && rotation )
      {
         unlink( item->thumbCache );
         free( item->thumbCache );
         item->thumbCache = NULL;
      }

      if ( (width==88) && (height==60) )
      {
         sendThumbnail( fd, path, item, width, height, rotation );
         return;
      }      
   }

   rc = pipe( pipefds );
   if( rc < 0 )
   {
      addToLog( "Unable to create pipe: %s\n",
         strerror_r( errno, buffer, sizeof buffer ) );
      write( fd, "404 file not found\r\n", 20 );
      return;
   }

   pid = fork();
   if( pid < 0 )
   {
      addToLog( "Unable to create fork for image conversion: %s\n",
         strerror_r( errno, buffer, sizeof buffer ) );
      write( fd, "404 file not found\r\n", 20 );
      return;
   }

   if( pid == 0 )
   {
      pipeSig.sa_handler = SIG_DFL;
      sigemptyset( &pipeSig.sa_mask );
      pipeSig.sa_flags = 0;
      sigaction( SIGPIPE, &pipeSig, NULL );

      rc = open( "/dev/null", O_RDWR );
      dup2( rc, 0 );
      dup2( rc, 2 );
      dup2( pipefds[ 1 ], 1 );
      close( rc );
      close( pipefds[ 0 ] );
      close( pipefds[ 1 ] );

      sprintf( buf1, "%dx%d", width, height );
      sprintf( buf2, "%d", ( item->rotation + rotation ) % 360 );
      if( ( ( item->rotation + rotation ) % 360 ) != 0 )
         execlp( getConvertPath(), "convert", "-size", buf1, "-resize", buf1,
            "-rotate", buf2, path, "jpeg:-", NULL );
      else
         execlp( getConvertPath(), "convert", "-size", buf1, "-resize", buf1,
            path, "jpeg:-", NULL );

      addToLog( "Unable to exec %s: %s\n", getConvertPath(),
         strerror_r( errno, buffer, sizeof buffer ) );
      exit( -1 );
   }

   close( pipefds[ 1 ] );

   addToLog( "Sending image file %s, length=%d\n", path, statbuf.st_size );

   write( fd, HTTP_HEADER, HTTP_HEADER_LEN );
#if 0
   /* We really don't know the size */
   sprintf( buffer, "Content-Length: %d\r\n", statbuf.st_size );
   write( fd, buffer, strlen( buffer ) );
#endif
   write( fd, "Content-Type: image/jpeg\r\n\r\n", 28 );
   
   pollfds[0].fd = pipefds[0];
   pollfds[1].fd = fd;

   pollfds[0].events = POLLIN;
   pollfds[1].events = 0;
   
   while (1)
   {
      pollfds[0].revents = 0;
      pollfds[1].revents = 0;

      rc = poll( pollfds, 2, 60*1000 );
      if ( rc == 0 )
      {
         addToLog( "sendJPEGFile(%s) convert timeout ... cleaning up process", path );
         kill( pid, SIGTERM );
         break;
      }

      if ( pollfds[1].revents & POLLHUP )
      {
         addToLog( "sendJPEGFile(%s) socket hungup", path );
         kill( pid, SIGTERM );
         break;
      }

      if ( pollfds[1].revents & (POLLOUT | POLLERR) )
      {
         i = writen( fd, buffer, i );
	 //addToLog( "sendJPEGFILE(%s), writen() %d %x ", path, i, pollfds[1].revents );
         if ( i <= 0 )
	 {
            break;
	 }
	 twritten += i;

	 pollfds[0].fd = pipefds[0];
         pollfds[0].events = POLLIN;

         pollfds[1].events = 0;
      }

      if (  pollfds[0].revents & (POLLHUP | POLLIN | POLLERR) )
      {
         i = readn( pipefds[0], buffer, sizeof(buffer) );
	 //addToLog( "sendJPEGFILE(%s), readn() %d %x ", path, i, pollfds[0].revents );
         if ( i <= 0 )
	 {
            break;
	 }
	 tread += i;

	 pollfds[0].fd = -1; /* igonore read() until we have finished writing */
         pollfds[0].events = 0;

         pollfds[1].events = POLLOUT;
      }

   }

   waitpid( -1, NULL, WNOHANG );
   addToLog( "Finished sending image file %s, length=%d (%d,%d)", path, statbuf.st_size, tread,twritten );
   kill( -pid, SIGTERM );
   kill( -pid, SIGKILL );

   close( pipefds[ 0 ] );

   item->rotation = ( rotation + item->rotation ) % 360;
}


/*
 * Send a 
 */
void sendCGIImage(
   int                  fd,
   char*                path,
   struct aParameterList* params
   )
{
   int                  i;
   char*                cp;
   struct stat          statbuf;
   struct pollfd        pollfds[2];
   int                  p1fds[2], p2fds[2];
   int                  rc;
   int                  width = 0, height = 0;
   int                  x = 0, y = 0;
   char*                tokptr;
   float                pixelShape = 1.0;
   struct aDirectoryEntry* item;
   char                 buffer[1024];
   char                 buf1[32];
   int                  pid[2];


   if( stat( path, &statbuf ) < 0 )
   {
      addToLog( "Unable to find image file %s: %s\n", path,
         strerror_r( errno, buffer, sizeof buffer ) );
      write( fd, "404 file not found\r\n", 20 );
      return;
   }

   cp = findParameter( params, "Width" );
   if( cp != NULL )
      width = atoi( cp );

   cp = findParameter( params, "Height" );
   if( cp != NULL )
      height = atoi( cp );

   cp = findParameter( params, "PixelShape" );
   if( cp != NULL )
   {
      cp = strtok_r( cp, ":", &tokptr );
      x = atoi( cp );
      cp = strtok_r( cp, ":", &tokptr );
      y = atoi( cp );

      pixelShape = ( ( double ) x ) / ( ( double ) y );

      if( y > x )        width /= pixelShape;
      else if( x > y )   height *= pixelShape;
   }

   rc = pipe(p1fds) || pipe(p2fds);
   if( rc < 0 )
   {
      addToLog( "Unable to create pipe: %s\n",
         strerror_r( errno, buffer, sizeof buffer ) );
      write( fd, "404 file not found\r\n", 20 );
      return;
   }

   pid[0] = fork();
   if( pid[0] < 0 )
   {
      addToLog( "Unable to create fork for image conversion: %s\n",
         strerror_r( errno, buffer, sizeof buffer ) );
      write( fd, "404 file not found\r\n", 20 );
      return;
   }

   if( pid[0] == 0 )
   {
      char *cp;

      close( p1fds[0] );

      rc = open( "/dev/null", O_RDWR );
      dup2( rc, fileno(stdin) );
      dup2( rc, fileno(stderr) );
      close( rc );

      dup2( p1fds[1], fileno(stdout) );
      close( p1fds[1] );

      cp = strrchr( path, '/' );
      if ( (cp[0] == 0) || (cp[1] == 0) )
	 exit(-1);
      
      cp++;

      execlp( path, cp, NULL );
      exit(-1);
   }

   close( p1fds[1] );

   pid[1] = fork();
   if( pid[0] < 0 )
   {
      addToLog( "Unable to create fork for image conversion: %s\n",
         strerror_r( errno, buffer, sizeof buffer ) );
      write( fd, "404 file not found\r\n", 20 );
      return;
   }

   if ( pid[1] == 0 )
   {
      close( p2fds[0] );

      dup2( p1fds[0], fileno(stdin ) );
      close( p1fds[0] );

      dup2( p2fds[1], fileno(stdout) );
      close( p2fds[1] );

      rc = open( "/dev/null", O_RDWR );
      dup2( rc, fileno(stderr) );
      close( rc );

      sprintf( buf1, "%dx%d", width, height );
      execlp( getConvertPath(), "convert", "-size", buf1, "-resize", buf1, "jpeg:-", "jpeg:-", NULL );

      exit( -1 );
   }

   close( p2fds[1] );

   addToLog( "Sending dynamic image file %s, length=%d\n", path, statbuf.st_size );

   write( fd, HTTP_HEADER, HTTP_HEADER_LEN );
   write( fd, "Content-Type: image/jpeg\r\n\r\n", 28 );


   pollfds[0].fd = p2fds[0];
   pollfds[1].fd = fd;

   pollfds[0].events = POLLIN;
   pollfds[1].events = 0;
   
   while (1)
   {
      pollfds[0].revents = 0;
      pollfds[1].revents = 0;

      rc = poll( pollfds, 2, 60*1000 );
      if ( rc == 0 )
      {
	 addToLog( "sendCGIFile(%s) convert timeout ... cleaning up process", path );
         kill( pid[0], SIGTERM );
         kill( pid[1], SIGTERM );
         break;
      }

      if ( pollfds[1].revents & POLLHUP )
      {
         kill( pid[0], SIGTERM );
         kill( pid[1], SIGTERM );
         break;
      }

      if ( pollfds[1].revents & (POLLOUT | POLLERR) )
      {
         i = writen( fd, buffer, i );
         if ( i <= 0 )
            break;

	 addToLog( "sendCGIImage(%s), writen() %d %x ", path, i, pollfds[1].revents );

	 pollfds[0].fd = p2fds[0];
         pollfds[0].events = POLLIN;

         pollfds[1].events = 0;
      }

      if (  pollfds[0].revents & (POLLHUP | POLLIN | POLLERR) )
      {
         i = readn( p2fds[0], buffer, sizeof(buffer) );
         if ( i <= 0 )
            break;

	 addToLog( "sendCGIImage(%s), readn() %d %x ", path, i, pollfds[0].revents );

	 pollfds[0].fd = -1; /* igonore read() until we have finished writing */
         pollfds[0].events = 0;

         pollfds[1].events = POLLOUT;
      }
   }

   waitpid( -1, NULL, WNOHANG );

   close( p1fds[ 0 ] );
   close( p2fds[ 0 ] );

   addToLog( "sendCGIFile(%s) done", path );
}
