/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * Dave Clemans, April 2003
 *
 * Copyright (C) 2003  Dave Clemans
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * See the file "COPYING" for more details.
 */

#include "globals.h"

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <search.h>
#include <time.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <signal.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/poll.h>
#include <netdb.h>
#include <stdio.h>

static void shutdownServer( int );

/*
 * Main server loop
 */
int main(
   int                  argc,
   char*                argv[]
   )
{
   struct sockaddr_in   beacon;
   struct sockaddr_in   server;
   struct sockaddr_in   bcast;
   struct sockaddr_in   client;
   struct in_addr       serv_address;
   struct in_addr       bind_address;
   struct in_addr       bcast_address;
   int                  sock;
   int                  serv_sock;
   struct pollfd        pollfds[ 2 ];
   struct aDirectoryEntry* dp;
   struct aParameterList* params;
   int                  rc;
   char*                cp;
   int                  fd;
   int                  on;
   FILE*                fp;
   socklen_t            len;
   time_t               currentTime;
   time_t               beaconSent;
   int                  stateCounter;
   char*                configfile = NULL;
   time_t               configTime;
   int                  debug = 0;
   int                  preload = 0;
   struct sigaction     shutdownSig;
   struct sigaction     pipeSig;
   struct stat          statbuf;

   shutdownSig.sa_handler = shutdownServer;
   sigemptyset( &shutdownSig.sa_mask );
   shutdownSig.sa_flags = 0;
   sigaction( SIGINT, &shutdownSig, NULL );
   sigaction( SIGQUIT, &shutdownSig, NULL );
   sigaction( SIGTERM, &shutdownSig, NULL );
   sigaction( SIGHUP, &shutdownSig, NULL );

   pipeSig.sa_handler = SIG_IGN;
   sigemptyset( &pipeSig.sa_mask );
   pipeSig.sa_flags = 0;
   sigaction( SIGPIPE, &pipeSig, NULL );

   initDirCache();

   for( rc = 1; rc < argc; rc++ )
   {
      if( argv[ rc ][ 0 ] != '-' )
         configfile = argv[ rc ];
      else
      {
         cp = argv[ rc ];
         if( strcasecmp( "-debug", cp ) == 0 )
         {
            debug++;
            setDebugMode( debug );
         }
         else if( strcasecmp( "-preload", cp ) == 0 )
         {
            preload++;
         }
      }
   }
   if( configfile == NULL )
   {
      fprintf( stderr, "byRequest: Missing config file parameter.\n" );
      exit( 1 );
   }
   if( !debug )
   {
      /* Run in background, disconnected from terminal */
      rc = fork();
      switch ( rc )
      {
         case -1:
            perror("fork");
            exit(1);
            break;
         case 0:
            break;   /* child */
         default:
            exit(0);  /* parent */
      }
      close(0);
      close(1);
      close(2);
      setsid();
   }

   stat( configfile, &statbuf );
   configTime = statbuf.st_mtime;
   readConfigFile( configfile );

   addToLog( "%s server, version %s\n\tfrom %s on %s as pid %d\n", SERVER_NAME,
      SERVER_VERSION, SERVER_ORGANIZATION, SERVER_COMMENT, getpid() );

   if( preload )
   {
      addToLog( "Preloading directory structure into server...\n" );

      params = ( struct aParameterList* ) malloc( sizeof( *params ) );
      params->top = NULL;
      params->bottom = NULL;
      storeParameter( params, "Recurse", "Yes" );
      dp = getFilteredDirectory( "/", params );
      freeDirectoryEntries( dp );
      freeParameters( params );
      free( params );

      addToLog( "Directory structure preloaded\n" );
   }

   beacon.sin_family = AF_INET;
   bind_address.s_addr = htonl( INADDR_ANY );
   beacon.sin_addr = bind_address;
   beacon.sin_port = htons( 2190 );

   bcast.sin_family = AF_INET;
   bcast_address.s_addr = htonl( getBcastAddress() );
   bcast.sin_addr = bcast_address;
   bcast.sin_port = htons( 2190 );

   if( ( ( sock = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP ) ) == -1 ) )
   {
      addToLog( "Unable to open beacon socket\n" );
      exit( 1 );
   }

   on = 1;
   (void)setsockopt ( sock, SOL_SOCKET, SO_REUSEADDR, (void *)&on, sizeof(on));

   if( bind( sock, ( struct sockaddr* ) &beacon, sizeof( struct sockaddr_in ) ) == -1 )
   {
      addToLog( "Unable to bind address to socket\n" );
      exit( 2 );
   }
   on = 1;
   if( setsockopt( sock, SOL_SOCKET, SO_BROADCAST, ( void* )&on, sizeof( on ) ) < 0 )
   {
      addToLog( "Unable to set broadcast flag on beacon socket\n" );
      exit( 2 );
   }
   
   server.sin_family = AF_INET;
   serv_address.s_addr = htonl( INADDR_ANY );
   server.sin_addr = serv_address;
   server.sin_port = htons( getDataPort() );
   
   if( ( ( serv_sock = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP ) ) == -1 ) )
   {
      addToLog( "Unable to open server socket\n" );
      exit( 3 );
   }

   on = 1;
   (void)setsockopt ( serv_sock, SOL_SOCKET, SO_REUSEADDR, (void *)&on, sizeof(on));

   if( bind( serv_sock, ( struct sockaddr* ) &server, sizeof( struct sockaddr_in ) ) == -1 )
   {
      addToLog( "Unable to bind address to server socket\n" );
      exit( 4 );
   }
   if( listen( serv_sock, 16 ) == -1 )
   {
      addToLog( "Unable to start listening to server socket\n" );
      exit( 5 );
   }

   stateCounter = 0;
   beaconSent = 0;
   for( ;; )
   {
      if( !stat( configfile, &statbuf ) )
      {
         if( statbuf.st_mtime > configTime + 300 )
         {
            configTime = statbuf.st_mtime;
            readConfigFile( configfile );
         }
      }

      /* Make sure any children our threads create are "waited" for */
      wait3( NULL, WNOHANG, NULL );

      currentTime = time( NULL );
      if( ( stateCounter < 10 && ( beaconSent + 5 <= currentTime ) ) ||
         ( stateCounter >= 10 && ( beaconSent + 60 <= currentTime ) ) )
      {
         sendBeaconMessage( sock, &bcast, sizeof( struct sockaddr_in ), 1 );
         beaconSent = currentTime;
         stateCounter++;
      }

      /*
       * Wait for something to happen
       */
      pollfds[ 0 ].fd = sock;
      pollfds[ 0 ].events = POLLIN | POLLPRI | POLLERR | POLLHUP;
      pollfds[ 0 ].revents = 0;
      pollfds[ 1 ].fd = serv_sock;
      pollfds[ 1 ].events = POLLIN | POLLPRI | POLLERR | POLLHUP;
      pollfds[ 1 ].revents = 0;
      rc = poll( pollfds, 2, ( stateCounter < 10 ) ? 5000 : 60000 );
      if( rc == 0 )
         continue;
      if( ( pollfds[ 0 ].revents & POLLHUP ) || ( pollfds[ 1 ].revents & POLLHUP ) )
         break;
      if( pollfds[ 0 ].revents > 0 )
      {
         processDatagram( sock );
      }
      if( pollfds[ 1 ].revents > 0 )
      {
         fd = accept( serv_sock, ( struct sockaddr* ) &client, &rc );
         if( fd < 0 )
         {
            addToLog( "Can't accept server connection?\n" );
            continue;
         }

         processCommand( fd );
      }
   }

#ifdef XMRADIO
   shutdownXMRadio();
#endif /* XMRADIO */
   exit( 0 );
}

/*
 * A basic shutdown signal handler, for services we may want to close down
 * cleanly when the server exits.
 */
static void shutdownServer(
   int                  code
   )
{
#ifdef XMRADIO
   shutdownXMRadio();
#endif /* XMRADIO */

   exit( 1 );
}
