/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * Dave Clemans, April 2003
 *
 * Copyright (C) 2003  Dave Clemans
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * See the file "COPYING" for more details.
 */

#include "globals.h"

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#define __USE_GNU 1
#include <search.h>

#ifdef PHOTO_ORGANIZER

#include <libexif/exif-data.h>

void get_exif( struct aDirectoryEntry *p )
{
   ExifData  *ped;
   char      *path;

   if ( !p )
      return;

   path = (char *)malloc( strlen(p->name) + strlen( p->dirp->path) + 2 );
   if ( !path )
      return;

   sprintf( path, "%s/%s", p->dirp->path, p->name );
   ped = exif_data_new_from_file( path );
   free(path);

   if ( ped )
   {
      ExifEntry *e;
      e = exif_content_get_entry( ped->ifd[EXIF_IFD_0], EXIF_TAG_DATE_TIME );
      if ( !e )
         e = exif_content_get_entry( ped->ifd[EXIF_IFD_0], EXIF_TAG_DATE_TIME );
 
      if ( !e )
         e = exif_content_get_entry( ped->ifd[EXIF_IFD_EXIF], EXIF_TAG_DATE_TIME_ORIGINAL );
 
      if ( !e )
         e = exif_content_get_entry( ped->ifd[EXIF_IFD_EXIF], EXIF_TAG_DATE_TIME_DIGITIZED );
 
      if ( e )
      {
	 e->data[ 4] = 0;
	 e->data[ 7] = 0;
	 e->data[10] = 0;
	 e->data[13] = 0;
	 e->data[16] = 0;
	 
	 p->exif_date_time.tm_year = atoi(e->data      );
	 p->exif_date_time.tm_mon  = atoi(e->data +  5 ) - 1;
	 p->exif_date_time.tm_mday = atoi(e->data +  8 );
	 p->exif_date_time.tm_hour = atoi(e->data + 11 );
	 p->exif_date_time.tm_min  = atoi(e->data + 14 );
	 p->exif_date_time.tm_sec  = atoi(e->data + 17 );
      }                                                                          

      exif_data_unref(ped);
   }
}

char *sMonths[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec", "" };


/*
 * Returns the contents of a virtual sub-directory as a directory list
 */
struct aDirectoryEntry* readPhotoOrganizerDirContents(
   char*                   path,
   struct aDirectory*      dirp
   )
{
   struct aParameterList*  params;
   struct aDirectoryEntry* topItem;
   struct aDirectoryEntry* bottom;
   struct aDirectoryEntry* item;
   struct aDirectoryEntry* realDir;
   struct aDirectoryEntry* ip;
   struct aDirectoryEntry* dp;
   struct aDirectoryEntry* cp;
   struct aDirectoryEntry* vp;
   struct aDirectory*      subdirp;
   struct aDirectory      *yrdirp;
   struct aDirectory      *mondirp;
   char*                   subpath;
   char                    key[64];

   topItem = NULL;

   if( strncmp( path, getPhotoOrganizerDirName(), strlen( getPhotoOrganizerDirName() ) ) )
   {
      addToLog( "Unknown virtual directory type.\n" );
      return NULL;
   }

   params  = (struct aParameterList *)malloc( sizeof( struct aParameterList ) );
   bzero( params, sizeof( struct aParameterList ) );
   storeParameter( params, "Recurse", "Yes" );

   for( vp = getPhotoOrganizerDirectories(); vp; vp = vp->next )
   {
      realDir = getFilteredDirectory( vp->name, params );
      if( realDir == NULL )
      {
         addToLog( "Failed to get virtual directory: %s.\n", vp->name );
         return NULL;
      }

      for( ip = realDir; ip; ip = ip->next )
      {
         if( !IS_JPEG( ip ) )
            continue;

         if( IS_FOLDER( ip ) )
            continue;

	 if ( !ip->exif_date_time.tm_year )
	    get_exif( ip );

	 sprintf( key, "%d", ip->exif_date_time.tm_year );

         for( dp = topItem; dp; dp = dp->next )
         {
	    if ( strcmp( key, dp->name ) == 0 )
	       break;
         }

         if( !dp )
         {
            item = newDirectoryEntry();
            item->dirp = dirp;
            item->name = strdup( key );
            item->description = strdup( key );
            item->fileType = FILE_FOLDER;
   
            if( !topItem )  topItem = item;
            else            bottom->next = item;
            bottom = item;
	 }

	 subpath = ( char* )malloc( strlen( path ) + strlen(key) + 2 );
	 sprintf( subpath, "%s/%s", path, key );
	 yrdirp = getFromDirCache( subpath );
	 if( !yrdirp )
	 {
	    yrdirp = newDirectory();
	    yrdirp->path = strdup( subpath );
	    addDirToDirCache( yrdirp );
	 }
	 free( subpath );

	 subpath = ( char* )malloc( strlen( path ) + strlen(key) + strlen( sMonths[ip->exif_date_time.tm_mon] ) + 3 );
	 sprintf( subpath, "%s/%s/%s", path, key, sMonths[ip->exif_date_time.tm_mon] );
	 mondirp = getFromDirCache( subpath );
	 if( !mondirp )
	 {
            item = newDirectoryEntry();
            item->dirp = yrdirp;
            item->name = strdup( sMonths[ip->exif_date_time.tm_mon] );
            item->description = strdup( sMonths[ip->exif_date_time.tm_mon] );
            item->fileType = FILE_FOLDER;

	    item->next = yrdirp->items;
	    yrdirp->items = item;

	    mondirp = newDirectory();
	    mondirp->path = strdup( subpath );
	    addDirToDirCache( mondirp );
	 }
	 free( subpath );

         cp = copyDirectoryEntry( ip );
         cp->next = mondirp->items;
         mondirp->items = cp;
      }

      freeDirectoryEntries( realDir );
   }

   freeParameters( params );

   return topItem;
}

#endif /* PHOTO_ORGANIZER */
