/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * Dave Clemans, April 2003
 *
 * Copyright (C) 2003  Dave Clemans
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * See the file "COPYING" for more details.
 */

#include "globals.h"

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>

/*
 * Read a .m3u (winamp/xmms) playlist file
 * Currently very minimal; metadata ignored
 */
struct aDirectoryEntry *readM3UFile(
   char*                path,
   struct aDirectory*   dirp
   )
{
   char*                cp;
   char*                newpath;
   FILE*                filep;
   struct aDirectoryEntry *topItem;
   struct aDirectoryEntry *bottom;
   struct aDirectoryEntry *item;
   struct stat          statbuf;
   char                 buffer[ 1024 ];
   char                 title[ 1024 ];

   newpath = decodeString( path );

   filep = fopen( newpath, "r" );
   if( filep == NULL )
   {
      addToLog(  "Can't open playlist file: %s\n", path );
      free( newpath );
      return NULL;
   }

   topItem = NULL;
   title[0] ='\0';
   while( fgets( buffer, sizeof buffer, filep ) != NULL )
   {
      for( cp = buffer; *cp; cp++ )
         if( *cp == '\n' )
            *cp = '\0';
      if( strlen( buffer ) == 0 )
         continue;
      if( buffer[ 0 ] == '#' )
      {
	 if ( strncmp( "#EXTINF:", buffer, strlen( "#EXTINF:" ) ) != 0 )
	    continue;

	 for ( cp = buffer ; *cp && (*cp != ',') ; cp++ )
	    /* NOOP */;

	 if ( *cp == ',' )
	    strcpy( title, cp+1 );

	 continue;
      }

      statbuf.st_mtime = 0;
      statbuf.st_ctime = 0;
      stat( buffer, &statbuf );
      item = newDirectoryEntry();
      item->dirp = dirp;
      item->name = ( char* ) malloc( strlen( buffer ) + 1 );
      strcpy( item->name, buffer );

      if (title[0] == '\0') {
         cp = strrchr (buffer, '/');
         strcpy (title, cp+1);
      }
      item->title = ( char* ) malloc( strlen( title ) + 1 );
      strcpy( item->title, title );
      addToLog(  "adding %s\n", item->title );
      title[0] ='\0';

      item->fileType = guessFileType( buffer );
      item->modified = statbuf.st_mtime;
      item->created = statbuf.st_ctime;

      if( topItem == NULL )
         topItem = item;
      else
         bottom->next = item;
      bottom = item;
   }
   fclose( filep );
   
   free( newpath );

   return topItem;
}

/*
 * Read a .pls playlist file
 * Currently very minimal; most metadata ignored
 */
struct aDirectoryEntry* readPLSFile(
   char*                path,
   struct aDirectory*   dirp
   )
{
   char*                cp;
   char*                newpath;
   FILE*                filep;
   struct aDirectoryEntry *topItem;
   struct aDirectoryEntry *bottom;
   struct aDirectoryEntry *item;
   char*                lastTitle;
   struct stat          statbuf;
   char                 buffer[ 1024 ];
   
   newpath = decodeString( path );

   filep = fopen( path, "r" );
   if( filep == NULL )
   {
      addToLog(  "Can't open playlist file: %s\n", path );
      free( newpath );
      return NULL;
   }

   topItem = NULL;
   lastTitle = NULL;
   while( fgets( buffer, sizeof buffer, filep ) != NULL )
   {
      for( cp = buffer; *cp; cp++ )
         if( *cp == '\n' )
            *cp = '\0';
      if( strlen( buffer ) == 0 )
         continue;
      if( buffer[ 0 ] == '#' || buffer[ 0 ] == '[' )
         continue;

      if( getDebugMode() )
         addToLog( "Playlist .pls file line: %s\n", buffer );

      if( strncasecmp( buffer, "title", 5 ) == 0 )
      {
         cp = &buffer[ 5 ];
         while( *cp && isdigit( ( int ) *cp ) )
            cp++;
         if( *cp == '=' )
         {
            if( lastTitle != NULL )
               free( lastTitle );
            lastTitle = ( char* ) malloc( strlen( &cp[ 1 ] ) + 1 );
            strcpy( lastTitle, &cp[ 1 ] );
         }
         continue;
      }
      else if( strncasecmp( buffer, "file", 4 ) != 0 )
      {
         continue;
      }
      cp = &buffer[ 4 ];
      while( *cp && isdigit( ( int ) *cp ) )
         cp++;
      if( *cp == '=' )
         cp++;

      statbuf.st_mtime = 0;
      statbuf.st_ctime = 0;
      stat( buffer, &statbuf );
      item = newDirectoryEntry();
      item->dirp = dirp;
      item->name = ( char* ) malloc( strlen( cp ) + 1 );
      strcpy( item->name, cp );
      item->fileType = guessFileType( cp );
      item->modified = statbuf.st_mtime;
      item->created = statbuf.st_ctime;
      if( lastTitle != NULL )
      {
         item->title = lastTitle;
         lastTitle = NULL;
      }

      if( topItem == NULL )
         topItem = item;
      else
         bottom->next = item;
      bottom = item;
   }
   fclose( filep );
   
   free( newpath );

   return topItem;
}

/*
 * Read a .ram (RealAudio?) playlist file
 * Currently very minimal; metadata ignored
 */
struct aDirectoryEntry *readRAMFile(
   char*                path,
   struct aDirectory*   dirp
   )
{
   char*                cp;
   char*                newpath;
   FILE*                filep;
   struct aDirectoryEntry *topItem;
   struct aDirectoryEntry *bottom;
   struct aDirectoryEntry *item;
   struct stat          statbuf;
   char                 buffer[ 1024 ];
   
   newpath = decodeString( path );

   filep = fopen( newpath, "r" );
   if( filep == NULL )
   {
      addToLog(  "Can't open playlist file: %s\n", path );
      free( newpath );
      return NULL;
   }

   topItem = NULL;
   while( fgets( buffer, sizeof buffer, filep ) != NULL )
   {
      for( cp = buffer; *cp; cp++ )
         if( *cp == '\n' )
            *cp = '\0';
      if( strlen( buffer ) == 0 )
         continue;
      if( buffer[ 0 ] == '#' || buffer[ 0 ] == '\'' )
         continue;

      statbuf.st_mtime = 0;
      statbuf.st_ctime = 0;
      stat( buffer, &statbuf );
      item = newDirectoryEntry();
      item->dirp = dirp;
      item->name = ( char* ) malloc( strlen( buffer ) + 1 );
      strcpy( item->name, buffer );
      item->fileType = guessFileType( buffer );
      item->modified = statbuf.st_mtime;
      item->created = statbuf.st_ctime;

      if( topItem == NULL )
         topItem = item;
      else
         bottom->next = item;
      bottom = item;
   }
   fclose( filep );
   
   free( newpath );

   return topItem;
}


#ifdef PLAYLIST_EDITING

static char *getPeerTempfileName (int fd)
{
//struct sockaddr name;
//socklen_t len = sizeof(struct sockaddr);
//int retval;
char buf[1024];

	strcpy (buf, "/tmp/byRequest_");
//	retval = getpeername (fd, &name, &len);
//	if (retval == 0)
//		strcat (buf, (char *) name.sa_data);
	return (strdup (buf));
}

void createPeerTempFile (int fd, char *path)
{
char *cp;
FILE *fp;

	cp = getPeerTempfileName (fd);
	fp = fopen (cp, "w");
	fprintf (fp, "%s\n", path);
	fclose (fp);
fprintf (stderr, "Created peerTempFile: '%s'\n", cp);
	free (cp);
}

void deletePeerTempFile (int fd)
{
char *cp;

	cp = getPeerTempfileName (fd);
	unlink (cp);
	free (cp);
}

static char *getPeerName (int fd)
{
char *cp;
static char buffer[2048];
FILE *fp;

	cp = getPeerTempfileName (fd);
	fp = fopen (cp, "r");
	fgets (buffer, 2048, fp);
	fclose (fp);
	free (cp);
	buffer[strlen (buffer) - 1] = 0;
	return buffer;
}


static int hasPeerTempFile (int fd)
{
char *cp;
int retval = 0;

	cp = getPeerTempfileName (fd);
	if (!access (cp, F_OK))
		retval = 1;
	return retval;
}

static void modifyPlaylistAsHTML (int fd, struct aMessage* msg, char *plFilename)
{
   char*                   addr;
   char*		   cp;

   addr = getHostaddress();

   htmlHeader (msg, 0, plFilename);

   if (hasPeerTempFile (fd))	{
      appendToMsg (msg, "<li><a href=\"http://%s:2190/TiVoConnect?Command=QueryContainer", addr);
      appendToMsg (msg, "&Container=%s&Format=text/html", plFilename);
      appendToMsg (msg, "&PlaylistEdit=Add\">");
      appendToMsg (msg, "Add playing file to playlist");
      appendToMsg (msg, "</a></li><br><li><a href=\"http://%s:2190/TiVoConnect?Command=QueryContainer", addr);
      appendToMsg (msg, "&Container=%s&Format=text/html", plFilename);
      appendToMsg (msg, "&PlaylistEdit=Remove\">");
      appendToMsg (msg, "Remove playing file from playlist");
      appendToMsg (msg, "</a><br></li>\n");
   } else {
      appendToMsg (msg, "<li>Add Song: Start a song playing first<br></li>");
      appendToMsg (msg, "<li>Remove Song: Start a song playing first<br></li>");
   }
   appendToMsg (msg, "<li><a href=\"http://%s:2190/TiVoConnect?Command=QueryContainer", addr);
   cp = strrchr (plFilename, '/');
   *cp++ = '\0';
   appendToMsg (msg, "&Container=%s&Format=text/html", plFilename);
   appendToMsg (msg, "&PlaylistEdit=Delete/%s\">", cp);
   appendToMsg (msg, "Delete playlist from server</a><br></li>\n");

   htmlFooter (msg);
   htmlSend (fd, msg);

}


static void modifyPlaylistAsXML (int fd, struct aMessage* msg, char *plFilename, int itemCount)
{
   char*                   addr;
   char*		   cp;
   char*		   title;
   char			   buffer[2048];

   appendToMsg( msg, "%s\n", XML_START );
   appendToMsg( msg, "<TiVoContainer>\n" );
   appendToMsg( msg, " <Details>\n" );
   appendToMsg( msg, " <ContentType>x-container/tivo-music</ContentType>\n");
   appendToMsg( msg, " <SourceFormat>x-container/folder</SourceFormat>\n" );
   appendToMsg( msg, " <TotalDuration>0</TotalDuration>\n");
   appendToMsg( msg, " <TotalItems>3</TotalItems>\n");
   appendToMsg( msg, " <Title>%s</Title>\n", getHostname() );
   appendToMsg( msg, " </Details>\n" );
   appendToMsg( msg, " <ItemStart>0</ItemStart>\n");
   if (itemCount == 0)
      appendToMsg( msg, " <ItemCount>0</ItemCount>\n");
   else {
      appendToMsg( msg, " <ItemCount>3</ItemCount>\n");


      //Add
      if (hasPeerTempFile (fd))	{
	title = "Add playing song to playlist";
	sprintf (buffer, "/TiVoConnect?Command=QueryContainer&Container=%s&PlaylistEdit=Add&Format=text/xml", plFilename);
      } else {
	title = "Add song: Start song you wish to add";
	sprintf (buffer, "/TiVoConnect?Command=QueryContainer&Container=%s&Format=text/xml", plFilename);
      }

      appendToMsg( msg, " <Item>\n" );
      appendToMsg( msg, "  <Details>\n" );
      appendToMsg( msg, "   <ContentType>x-container/folder</ContentType>\n");
      appendToMsg( msg, "   <SourceFormat>x-container/folder</SourceFormat>\n" );
      appendToMsg( msg, "   <Title>%s</Title>\n", title);
      appendToMsg( msg, "  </Details>\n" );
      appendToMsg( msg, "  <Links>\n" );
      appendToMsg( msg, "   <Content>\n" );
      appendToMsg( msg, "    <Url>%s</Url>\n", buffer);
      appendToMsg( msg, "   </Content>\n" );
      appendToMsg( msg, "  </Links>\n" );
      appendToMsg( msg, " </Item>\n" );

      //Remove
      if (hasPeerTempFile (fd))	{
	title = "Remove playing song from playlist";
	sprintf (buffer, "/TiVoConnect?Command=QueryContainer&Container=%s&PlaylistEdit=Remove&Format=text/xml", plFilename);
      } else {
	title = "Remove song: Start song you wish to remove";
	sprintf (buffer, "/TiVoConnect?Command=QueryContainer&Container=%s&Format=text/xml", plFilename);
      }

      appendToMsg( msg, " <Item>\n" );
      appendToMsg( msg, "  <Details>\n" );
      appendToMsg( msg, "   <ContentType>x-container/folder</ContentType>\n");
      appendToMsg( msg, "   <SourceFormat>x-container/folder</SourceFormat>\n" );
      appendToMsg( msg, "   <Title>%s</Title>\n", title);
      appendToMsg( msg, "  </Details>\n" );
      appendToMsg( msg, "  <Links>\n" );
      appendToMsg( msg, "   <Content>\n" );
      appendToMsg( msg, "    <Url>%s</Url>\n", buffer);
      appendToMsg( msg, "   </Content>\n" );
      appendToMsg( msg, "  </Links>\n" );
      appendToMsg( msg, " </Item>\n" );

      // Delete
      appendToMsg( msg, " <Item>\n" );
      appendToMsg( msg, "  <Details>\n" );
      appendToMsg( msg, "   <ContentType>x-container/folder</ContentType>\n");
      appendToMsg( msg, "   <SourceFormat>x-container/folder</SourceFormat>\n" );
//      cp = strrchr (path, '/');
      appendToMsg( msg, "   <Title>Delete playlist from server</Title>\n");
      appendToMsg( msg, "  </Details>\n" );
      appendToMsg( msg, "  <Links>\n" );
      appendToMsg( msg, "   <Content>\n" );
      appendToMsg( msg, "    <Url>/TiVoConnect?Command=QueryContainer");
      cp = strrchr (plFilename, '/');
      *cp++ = '\0';
      appendToMsg (msg, "&Container=%s&Format=text/xml", plFilename);
      appendToMsg (msg, "&PlaylistEdit=Delete/%s\">", cp);
      appendToMsg( msg, "</Url>\n" );
      appendToMsg( msg, "   </Content>\n" );
      appendToMsg( msg, "  </Links>\n" );
      appendToMsg( msg, " </Item>\n" );
   }

   appendToMsg( msg, "</TiVoContainer>\n" );
   appendToMsg( msg, "%s\n", XML_END );
   sendXmlMessage( msg, fd );
}


int modifyPlaylist( int fd, struct aParameterList* params, char * mode)
{
	// only as HTML for the moment
   struct aMessage*        msg;
   FILE *		   fp;
   FILE *	           newfp;
   char 		   buffer[2048];
   char *		   peer;
   char* 		   plFilename;
   char*		   format;
   char*		   itemCountStr;
   int			   itemCount = 0;

   plFilename = findParameter (params, "Container");
   plFilename = decodeString( plFilename );

   if (!strcasecmp (mode, "Add"))	{
	peer = getPeerName (fd);
	fp = fopen (plFilename, "a");
	// should look for dups
	fprintf (fp, "%s\n", peer);
	fflush (fp);
	fclose (fp);
	return 1;
   }

   if (!strcasecmp (mode, "Remove"))	{
	peer = getPeerName (fd);
	fp = fopen (plFilename, "r");
	sprintf (buffer, "%s_", plFilename);
	newfp = fopen (buffer, "w");
	while (!feof (fp))	{
		fgets (buffer, 2048, fp);
		buffer[strlen (buffer) - 1] = 0;
		if (strcmp (peer, buffer))
			fprintf (newfp, "%s\n", buffer);
	}
	fclose (fp);
	fclose (newfp);
	unlink (plFilename);
	sprintf (buffer, "%s_", plFilename);
	rename (buffer, plFilename);
	return 1;
   }

   if (!strncasecmp (mode, "Delete/", 7))	{
	sprintf (buffer, "%s/%s", plFilename, &mode[7]);
	unlink (buffer);
	return 1;
   }

   // modify page
   format = findParameter (params, "Format");
   format = decodeString( format );
   itemCountStr = findParameter (params, "ItemCount");
   if (itemCountStr != NULL)
	itemCount = atoi (itemCountStr);
   msg = newMessage();

   if (format != NULL && !strcasecmp (format, "text/html"))
   	modifyPlaylistAsHTML (fd, msg, plFilename);
   else if (format != NULL && !strcasecmp (format, "text/xml"))
   	modifyPlaylistAsXML (fd, msg, plFilename, itemCount);

   freeMessage( msg );
   return 0;
}

#endif

