/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * This file supports a time-based functionality for automatic recording
 * of shows from an internet radio stream.
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * Dave Clemans, May 2004
 *
 * Copyright (C) 2004  Dave Clemans
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * See the file "COPYING" for more details.
 */
                                                                                
#include "globals.h"

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/poll.h>
#include <time.h>
#include <errno.h>
#include <dirent.h>

#ifdef __APPLE__
#define blkcnt_t	long
#endif

/*
 * The event list to record...
 */
struct aRecorderEvent
{
   char*                url;
   char*                show;
   int                  year;
   int                  month;
   int                  day;
   int                  hour;
   int                  minute;
   int                  weekday;
   int                  duration;
   time_t               nextEvent;

   struct aRecorderEvent* next;
};

static struct aRecorderEvent* firstEvent = NULL;
static struct aRecorderEvent* lastEvent = NULL;
static char*            eventContainer = NULL;
static int              maximumContainerSize;

#if USE_THREADS
static pthread_t        recorderThreadId = 0;
static pthread_mutex_t  recorderMutex = PTHREAD_MUTEX_INITIALIZER;
static int              recorderThreadStarted = 0;
#endif

static struct aRecorder* currentRecorder = NULL;

static void recorderStartStream( struct aRecorderEvent* );

struct aRecorder* getCurrentRecorder (void)
{
	return currentRecorder;
}


/*
 * Read and validate a scheduling file, and start the background recorder
 * thread.  If we see any problems, a message is logged, and the recorder
 * thread is not started.
 *
 * The scheduler file format...
 *    # comment line (ignored )
 *    URL ShowName Duration Year Month Day Hour Minute Weekday
 *       a value of '*' for Year, Month, Day, Hour, Minute is a wildcard
 */
void recorderInit(
   char*		containerName,
   char*                containerDir,
   char*                schedule,
   char*                maxContainerSize
   )
{
   FILE*                fp;
   struct aRecorderEvent* rep;
   char*                cp;
   char*                tokPtr;
   struct stat          statbuf;
   char                 buffer[ 512 ];
   char                 errbuf[ 512 ];

   if (currentRecorder != NULL) {
	free (currentRecorder->name);
	free (currentRecorder->directory);
	free (currentRecorder->filename);
	free (currentRecorder->containerSize);
	free (currentRecorder);
   }
   currentRecorder = (struct aRecorder *) calloc (1, sizeof (struct aRecorder));
   currentRecorder->name = strdup (containerName);
   currentRecorder->directory = strdup (containerDir);
   currentRecorder->filename = strdup (schedule);
   currentRecorder->containerSize = strdup (maxContainerSize);

   if( stat( containerDir, &statbuf ) )
   {
      if( mkdir( containerDir, 0755 ) )
      {
         addToLog( "Recorder directory '%s' not accessible; %s\n",
            strerror_r( errno, errbuf, sizeof errbuf ) );

         return;
      }
   }
   if( !S_ISDIR( statbuf.st_mode ) )
   {
      addToLog( "Recorder directory '%s' is not usable.\n" );

      return;
   }
   if( eventContainer )
      free( eventContainer );
   eventContainer = strdup( containerDir );

   maximumContainerSize = atoi( maxContainerSize );
   if( maximumContainerSize <= 0 )
      maximumContainerSize = 10;

   fp = fopen( schedule, "r" );
   if( !fp )
   {
      addToLog( "Unable to open scheduler file '%s'; %s\n", schedule,
         strerror_r( errno, errbuf, sizeof errbuf ) );

      return;
   }

#if USE_THREADS
   pthread_mutex_lock( &recorderMutex );
#endif

   while( firstEvent )
   {
      rep = firstEvent;
      firstEvent = firstEvent->next;

      if( rep->url )
         free( rep->url );
      if( rep->show )
         free( rep->show );
      free( rep );
   }
   firstEvent = NULL;
   lastEvent = NULL;

   while( fgets( buffer, sizeof buffer, fp ) != NULL )
   {
      if( buffer[ 0 ] == '#' )
         continue;
      if( strlen( buffer ) == 0 || buffer[ 0 ] == '\n' )
         continue;

      rep = ( struct aRecorderEvent* ) malloc( sizeof( *rep ) );
      cp = getNextWord( buffer, &tokPtr );
      rep->url = strdup( cp );
      cp = getNextWord( tokPtr, &tokPtr );
      rep->show = strdup( cp );
      cp = getNextWord( tokPtr, &tokPtr );
      rep->duration = atoi( cp ) * 60;
      cp = getNextWord( tokPtr, &tokPtr );
      if( isdigit( ( int ) *cp ) )
         rep->year = atoi( cp );
      else
         rep->year = -1;
      if( rep->year > 1900 )
         rep->year -= 1900;
      cp = getNextWord( tokPtr, &tokPtr );
      if( isdigit( ( int ) *cp ) )
         rep->month = atoi( cp );
      else
         rep->month = -1;
      if( rep->month > 11 )
         rep->month = -1;
      cp = getNextWord( tokPtr, &tokPtr );
      if( isdigit( ( int ) *cp ) )
         rep->day = atoi( cp );
      else
         rep->day = -1;
      if( rep->day > 31 )
         rep->day = -1;
      cp = getNextWord( tokPtr, &tokPtr );
      if( isdigit( ( int ) *cp ) )
         rep->hour = atoi( cp );
      else
         rep->hour = -1;
      if( rep->hour > 23 )
         rep->hour = -1;
      cp = getNextWord( tokPtr, &tokPtr );
      rep->minute = atoi( cp );
      if( rep->minute > 59 )
         rep->minute = -1;
      rep->weekday = -1;
      cp = getNextWord( tokPtr, &tokPtr );
      if( *cp )
      {
         if( strcmp( cp, "1-5" ) == 0 )
            rep->weekday = 7;
         else if( strcmp( cp, "0,6" ) == 0 )
            rep->weekday = 8;
         else if( isdigit( ( int ) *cp ) )
            rep->weekday = atoi( cp );
         else
         {
            if( strncasecmp( cp, "sunday", 2 ) == 0 )
               rep->weekday = 0;
            else if( strncasecmp( cp, "monday", 2 ) == 0)
               rep->weekday = 1;
            else if( strncasecmp( cp, "tuesday", 2 ) == 0 )
               rep->weekday = 2;
            else if( strncasecmp( cp, "wednesday", 2 ) == 0 )
               rep->weekday = 3;
            else if( strncasecmp( cp, "thursday", 2 ) == 0 )
               rep->weekday = 4;
            else if( strncasecmp( cp, "friday", 2 ) == 0 )
               rep->weekday = 5;
            else if( strncasecmp( cp, "saturday", 2 ) == 0 )
               rep->weekday = 6;
         }
      }
      if( rep->weekday > 8 )
         rep->weekday = -1;
      rep->nextEvent = 0;
      rep->next = NULL;

      if( !firstEvent )
         firstEvent = rep;
      else
         lastEvent->next = rep;
      lastEvent = rep;
   }

#if USE_THREADS
   pthread_mutex_unlock( &recorderMutex );
#endif

   fclose( fp );

   recorderUpdateEvents();

#if USE_THREADS
   recorderStartThread();
#endif
}

/*
 * Update the next event time on our pending scheduling events
 */
void recorderUpdateEvents( void )
{
   struct tm*           timeInfo;
   struct tm            timeInfo_s;
   struct tm            nextInfo_s;
   time_t               current;
   int                  dayIncr;
   int                  daysInMonth;
   struct aRecorderEvent* rep;
   char                 buffer[ 64 ];

   current = time( NULL );
   timeInfo = localtime_r( &current, &timeInfo_s );

#if USE_THREADS
   pthread_mutex_lock( &recorderMutex );
#endif

   for( rep = firstEvent; rep; rep = rep->next )
   {
      nextInfo_s.tm_wday = timeInfo->tm_wday;
      nextInfo_s.tm_yday = timeInfo->tm_yday;
      nextInfo_s.tm_isdst = timeInfo->tm_isdst;
      nextInfo_s.tm_sec = 0;

      if( rep->year >= 0 )
         nextInfo_s.tm_year = rep->year;
      else
         nextInfo_s.tm_year = timeInfo->tm_year;
      if( rep->month >= 0 )
         nextInfo_s.tm_mon = rep->month;
      else
         nextInfo_s.tm_mon = timeInfo->tm_mon;
      if( rep->day >= 0 )
         nextInfo_s.tm_mday = rep->day;
      else
         nextInfo_s.tm_mday = timeInfo->tm_mday;
      if( rep->hour >= 0 )
         nextInfo_s.tm_hour = rep->hour;
      else
         nextInfo_s.tm_hour = timeInfo->tm_hour;
      if( rep->minute >= 0 )
         nextInfo_s.tm_min = rep->minute;
      else
         nextInfo_s.tm_min = timeInfo->tm_min;
      nextInfo_s.tm_sec = 0;

      rep->nextEvent = mktime( &nextInfo_s );

      if( rep->day < 0 )
      {
         if( ( rep->nextEvent < ( current - 10 ) ) ||
            ( rep->weekday >= 0 && rep->weekday != nextInfo_s.tm_wday ) )
         {
            if( rep->weekday < 0 )
               dayIncr = 1;
            else
            {
               if( rep->weekday == 7 )
               {
                  if( nextInfo_s.tm_wday == 5 )
                     dayIncr = ( rep->nextEvent > current ) ? 0 : 3;
                  else if( nextInfo_s.tm_wday == 6 )
                     dayIncr = 2;
                  else if( nextInfo_s.tm_wday == 0 )
                     dayIncr = 1;
		  else
		     dayIncr = ( rep->nextEvent > current ) ? 0 : 1;
               }
               else if( rep->weekday == 8 )
               {
                  if( nextInfo_s.tm_wday == 0 )
                     dayIncr = ( rep->nextEvent > current ) ? 0 : 6;
                  else if( nextInfo_s.tm_wday == 6 )
                     dayIncr = ( rep->nextEvent > current ) ? 0 : 1;
                  else
                  {
                     dayIncr = 6 - nextInfo_s.tm_wday;
                  }
               }
               else
               {
                  if( rep->weekday > nextInfo_s.tm_wday )
                  {
                     dayIncr = rep->weekday - nextInfo_s.tm_wday;
                  }
                  else
                  {
                     dayIncr = 7 - ( nextInfo_s.tm_wday - rep->weekday );
                  }
               }
            }

            switch( timeInfo->tm_mon )
            {
               case 0:
               case 2:
               case 4:
               case 6:
               case 7:
               case 9:
               case 11:
                  daysInMonth = 31;
                  break;
               case 3:
               case 5:
               case 8:
               case 10:
                  daysInMonth = 30;
                  break;
               case 1:
                  daysInMonth = 28;
                  if( ( timeInfo->tm_year % 4 ) == 0 )
                  {
                     if( ( timeInfo->tm_year % 100 ) == 0 )
                     {
                        if( ( timeInfo->tm_year % 400 ) == 0 )
                           daysInMonth = 29;
                     }
                     else
                     {
                        daysInMonth = 29;
                     }
                  }
                  break;
            }

            if( ( nextInfo_s.tm_mday + dayIncr ) < daysInMonth )
               nextInfo_s.tm_mday += dayIncr;
            else
            {
               nextInfo_s.tm_mday += dayIncr - daysInMonth;
               nextInfo_s.tm_mon++;
               if( nextInfo_s.tm_mon >= 12 )
               {
                  nextInfo_s.tm_mon = 0;
                  nextInfo_s.tm_year++;
               }
            }
         }
      }

      rep->nextEvent = mktime( &nextInfo_s );
      if( rep->nextEvent < ( current - 10 ) )
         rep->nextEvent = 0;

      addToLog( "Next recording event for url '%s' scheduled at: %s",
         rep->url, asctime_r( &nextInfo_s, buffer ) );
   }

#if USE_THREADS
   pthread_mutex_unlock( &recorderMutex );
#endif
}

/*
 * Start the recorder thread, if it's not already started...
 */
void recorderStartThread( void )
{
   if( recorderThreadStarted )
   {
      pthread_cancel( recorderThreadId );
   }

   recorderThreadId = newThread( ( void* (*)( void* ) ) recorderThread, NULL );

   recorderThreadStarted++;
}

/*
 * The main recorder thread
 */
void recorderThread( void )
{
   time_t               current;
   time_t               nextEvent;
   struct aRecorderEvent* rep;
   sigset_t             threadMask;

   sigemptyset( &threadMask );
   sigaddset( &threadMask, SIGHUP );
   sigaddset( &threadMask, SIGINT );
   sigaddset( &threadMask, SIGQUIT );
   sigaddset( &threadMask, SIGTERM );
   pthread_sigmask( SIG_BLOCK, &threadMask, NULL );

   while( 1 )
   {
      addToLog( "Recorder thread is checking for recording events...\n" );

      current = time( NULL );
      nextEvent = 0;
#if USE_THREADS
      pthread_mutex_lock( &recorderMutex );
#endif
      for( rep = firstEvent; rep; rep = rep->next )
      {
         if( nextEvent == 0 || nextEvent > rep->nextEvent )
         {
            if( rep->nextEvent > current )
               nextEvent = rep->nextEvent;
         }
      }
#if USE_THREADS
      pthread_mutex_unlock( &recorderMutex );
#endif
      if( nextEvent == 0 )
      {
         pauseFor( 3600, 0 );
         continue;
      }

      addToLog( "Recorder thread sleeping for %d...\n",
         nextEvent - current - 1 );
      pauseFor( nextEvent - current - 1, 0 );

      addToLog( "Recorder thread waking up... check for recording events.\n" );

      current = time( NULL );
#if USE_THREADS
      pthread_mutex_lock( &recorderMutex );
#endif
      for( rep = firstEvent; rep; rep = rep->next )
      {
         if( ( rep->nextEvent > ( current - 90 ) ) &&
            ( rep->nextEvent < ( current + 90 ) ) )
         {
#if USE_THREADS
            newThread( ( void* (*)( void* ) ) recorderStartStream, rep );
#else
            recorderStartStream( rep );
#endif
         }
      }
#if USE_THREADS
      pthread_mutex_unlock( &recorderMutex );
#endif

      pauseFor( 360, 0 );

      recorderUpdateEvents();
   }
}

/*
 * Finally, prepare the stream for recording, and start the thread to 
 * do the recording.
 */
void recorderStartStream(
   struct aRecorderEvent* rep
   )
{
   DIR*                 dp;
   struct dirent*       dentry;
   struct stat          statbuf;
   char*                cp;
   char*                pathname;
   char*                todelete;
   blkcnt_t             numblocks;
   blkcnt_t             maxblocks;
   time_t               oldest;
   int                  duration;
   struct tm*           timeInfo;
   struct tm            timeInfo_s;
   int                  fd;
   char                 errbuf[ 256 ];
   sigset_t             threadMask;

   sigemptyset( &threadMask );
   sigaddset( &threadMask, SIGHUP );
   sigaddset( &threadMask, SIGINT );
   sigaddset( &threadMask, SIGQUIT );
   sigaddset( &threadMask, SIGTERM );
   pthread_sigmask( SIG_BLOCK, &threadMask, NULL );

   addToLog( "Start recording event for url '%s'\n", rep->url );

   pathname = ( char* ) malloc( strlen( eventContainer ) + 1024 );

#if USE_THREADS
   pthread_mutex_lock( &recorderMutex );
#endif

   dp = opendir( eventContainer );
   if( dp )
   {
      numblocks = 0;
      maxblocks = maximumContainerSize * 1024 * 1024 * 1024 / 512;

      while( ( dentry = readdir( dp ) ) != NULL )
      {
         if( dentry->d_name[ 0 ] == '.' )
            continue;

         cp = dentry->d_name;
         while( *cp )
            cp++;
         while( cp > dentry->d_name && *cp != '.' )
            cp--;
         if( strcasecmp( cp, ".mp3" ) == 0 )
         {
            sprintf( pathname, "%s/%s", eventContainer, dentry->d_name );
            if( !stat( pathname, &statbuf ) )
            {
               numblocks += statbuf.st_blocks;
            }
         }
      }

      while( numblocks > maxblocks )
      {
         rewinddir( dp );
         oldest = 0;
         todelete = NULL;
         pathname[ 0 ] = '\0';
         while( ( dentry = readdir( dp ) ) != NULL )
         {
            if( dentry->d_name[ 0 ] == '.' )
               continue;

            cp = dentry->d_name;
            while( *cp )
               cp++;
            while( cp > dentry->d_name && *cp != '.' )
               cp--;
            if( strcasecmp( cp, ".mp3" ) == 0 )
            {
               sprintf( pathname, "%s/%s", eventContainer, dentry->d_name );
               stat( pathname, &statbuf );
               if( oldest == 0 || statbuf.st_mtime < oldest )
               {
                  if( todelete )
                     free( todelete );
                  todelete = strdup( pathname );
                  oldest = statbuf.st_mtime;
               }
            }
         }
         if( todelete )
         {
            stat( todelete, &statbuf );
            unlink( todelete );
            numblocks -= statbuf.st_blocks;
            free( todelete );
            todelete = NULL;
         }
      }

      closedir( dp );
   }

   timeInfo = localtime_r( &rep->nextEvent, &timeInfo_s );
   sprintf( pathname, "%s/%s_%04d%02d%02d%02d%02d.mp3", eventContainer,
      rep->show, timeInfo->tm_year + 1900, timeInfo->tm_mon + 1, timeInfo->tm_mday,
      timeInfo->tm_hour, timeInfo->tm_min );

#if USE_THREADS
   pthread_mutex_unlock( &recorderMutex );
#endif

   fd = open( pathname, O_WRONLY | O_CREAT, 0664 );
   if( fd < 0 )
   {
      addToLog( "Failed to create %s; %s\n", pathname,
         strerror_r( errno, errbuf, sizeof errbuf ) );

      free( pathname );
      return;
   }

   free( pathname );

   sendStreamFile( fd, rep->url, NULL, 0, rep->duration + 10 );
}
