/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * Defines a "pseudo" stream object recording what internet, xmpcr, etc.
 * streams are currently active.  A stream object, once it's created, can be
 * told to stream to all its output clients until ALL of the output fd's
 * are closed, OR the input fd is closed.
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * Dave Clemans, May 2004
 *
 * Copyright (C) 2004  Dave Clemans
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * See the file "COPYING" for more details.
 */

#include "globals.h"

#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <sys/poll.h>

static struct anAudioStream* activeStreams = NULL;

#if USE_THREADS
static pthread_mutex_t     streamMutex = PTHREAD_MUTEX_INITIALIZER;
#endif

/*
 * Creates and returns a new empty stream object.  The stream object
 * is linked into the list of active streams.
 */
struct anAudioStream* audioStreamCreateStream( void )
{
   struct anAudioStream*   stream;

   stream = ( struct anAudioStream* ) malloc( sizeof *stream );
   stream->fd = -1;
   stream->numClients = 0;
   stream->clients = NULL;
   stream->url = NULL;
   stream->isXM = 0;

#if USE_THREADS
   pthread_mutex_lock( &streamMutex );
#endif
   stream->next = activeStreams;
   activeStreams = stream;
#if USE_THREADS
   pthread_mutex_unlock( &streamMutex );
#endif

   return stream;
}

/*
 * Destroys a stream object by de-linking it from the list of active streams,
 * and freeing up all the memory it used.
 */
void audioStreamDestroyStream(
   struct anAudioStream*      stream
   )
{
   struct anAudioStream*      s;
   struct anAudioStream*      sp;
   int                        i;

#if USE_THREADS
   pthread_mutex_lock( &streamMutex );
#endif
   s = activeStreams;
   sp = NULL;
   for( ; s; s = s->next )
   {
      if( s == stream )
         break;
      sp = s;
   }
   if( s )
   {
#ifdef XMRADIO
      if( s->isXM )
         releaseXMRadioStream();
#endif /* XMRADIO */

      if( s->fd >= 0 )
         close( s->fd );
      for( i = 0; i < s->numClients; i++ )
      {
         if( s->clients[ i ].fd >= 0 )
            close( s->clients[ i ].fd );
      }
      if( s->clients )
         free( s->clients );
      s->clients = NULL;
      if( s->url )
         free( s->url );
      s->url = NULL;
      s->fd = -1;

      if( !sp )
         activeStreams = activeStreams->next;
      else
      {
         sp->next = s->next;
      }

      free( s );
   }
#if USE_THREADS
   pthread_mutex_unlock( &streamMutex );
#endif
}

/*
 * Set the url of this audio stream object.
 */
void audioStreamSetUrl(
   struct anAudioStream*   stream,
   char*                   url
   )
{
#if USE_THREADS
   pthread_mutex_lock( &streamMutex );
#endif

   if( stream->url )
      free( stream->url );
   stream->url = strdup( url );

#if USE_THREADS
   pthread_mutex_unlock( &streamMutex );
#endif
}

/*
 * Returns the current url of an audio stream.
 */
char* audioStreamGetUrl(
   struct anAudioStream*   stream
   )
{
   if( !stream->url )
      return "*none*";
   return stream->url;
}

/*
 * Set the input fd of this audio stream object.
 */
void audioStreamSetFd(
   struct anAudioStream*   stream,
   int                     fd
   )
{
#if USE_THREADS
   pthread_mutex_lock( &streamMutex );
#endif

   if( stream->fd >= 0 )
      close( stream->fd );
   stream->fd = fd;

#if USE_THREADS
   pthread_mutex_unlock( &streamMutex );
#endif
}

/*
 * Returns the input fd of this audio stream object.
 */
int audioStreamGetFd(
   struct anAudioStream*   stream
   )
{
   return stream->fd;
}

/*
 * Add a new output client to a stream object.
 */
void audioStreamAddClient(
   struct anAudioStream*   stream,
   int                     client,
   int                     addStreamHeader,
   int                     duration,
   int                     isXM
   )
{
   struct anAudioStreamClient* clients;
   int                     i;

   addToLog( "Add fd %d to stream %s\n", client, audioStreamGetUrl( stream ) );

#if USE_THREADS
   pthread_mutex_lock( &streamMutex );
#endif

   if( !stream->clients )
   {
      stream->clients = ( struct anAudioStreamClient* ) malloc( sizeof( *clients ) );
      stream->clients[ 0 ].fd = client;
      stream->clients[ 0 ].open = 1;
      stream->clients[ 0 ].duration = duration;
      if( duration <= 0 )
         stream->clients[ 0 ].endingTime = 0;
      else
         stream->clients[ 0 ].endingTime = time( NULL ) + duration + 10;
      stream->numClients = 1;
   }
   else
   {
      clients = stream->clients;
      stream->clients = ( struct anAudioStreamClient* )
         malloc( ( stream->numClients + 1 ) * sizeof ( *clients ) );
      for( i = 0; i < stream->numClients; i++ )
      {
         stream->clients[ i ].fd = clients[ i ].fd;
         stream->clients[ i ].duration = clients[ i ].duration;
         stream->clients[ i ].open = clients[ i ].open;
         stream->clients[ i ].endingTime = clients[ i ].endingTime;
      }
      free( clients );
      stream->clients[ stream->numClients ].fd = client;
      stream->clients[ stream->numClients ].open = 1;
      stream->clients[ stream->numClients ].duration = duration;
      if( duration <= 0 )
         stream->clients[ i ].endingTime = 0;
      else
         stream->clients[ i ].endingTime = time( NULL ) + duration + 10;
      stream->numClients++;
   }

   stream->isXM = isXM;

#if USE_THREADS
   pthread_mutex_unlock( &streamMutex );
#endif

   if( addStreamHeader )
   {
      write( client, HTTP_HEADER, HTTP_HEADER_LEN );
      write( client, "Content-Length: 0\r\n", 19 );
      write( client, "Content-Type: audio/mpeg\r\n\r\n", 28 );
   }
}

/*
 * Remove a client output from an audio stream.  If this is the last client,
 * the stream is effectively shut down (but still allocated).
 */
void audioStreamRemoveClient(
   struct anAudioStream*   stream,
   int                     client
   )
{
   int                     clientNum;
   int                     i;
   struct anAudioStreamClient* clients;

   addToLog( "Remove fd %d from stream %s\n", client, audioStreamGetUrl( stream ) );

#if USE_THREADS
   pthread_mutex_lock( &streamMutex );
#endif

   for( clientNum = 0; clientNum < stream->numClients; clientNum++ )
   {
      if( stream->clients[ clientNum ].fd == client )
         break;
   }
   if( clientNum < stream->numClients )
   {
      if( stream->numClients == 1 )
      {
         free( stream->clients );
         stream->clients = NULL;
      }
      else
      {
         clients = ( struct anAudioStreamClient* )
            malloc( ( stream->numClients - 1 ) * sizeof ( *clients ) );
         for( i = 0; i < clientNum; i++ )
         {
            clients[ i ].fd = stream->clients[ i ].fd;
            clients[ i ].open = stream->clients[ i ].open;
            clients[ i ].duration = stream->clients[ i ].duration;
            clients[ i ].endingTime = stream->clients[ i ].endingTime;
         }
         for( i = clientNum + 1; i < stream->numClients; i++ )
         {
            clients[ i - 1 ].fd = stream->clients[ i ].fd;
            clients[ i - 1 ].open = stream->clients[ i ].open;
            clients[ i - 1 ].duration = stream->clients[ i ].duration;
            clients[ i - 1 ].endingTime = stream->clients[ i ].endingTime;
         }
         free( stream->clients );
         stream->clients = clients;
      }
      stream->numClients--;
      close( client );
   }

#if USE_THREADS
   pthread_mutex_unlock( &streamMutex );
#endif
}

/*
 * Returns the number of output fds in this stream.
 */
int audioStreamGetNumOutputs(
   struct anAudioStream*   stream
   )
{
   return stream->numClients;
}

/*
 * Is a stream already open on this url?
 */
struct anAudioStream* audioStreamFindStream(
   char*                   url
   )
{
   struct anAudioStream*   stream;

#if USE_THREADS
   pthread_mutex_lock( &streamMutex );
#endif

   for( stream = activeStreams; stream; stream = stream->next )
   {
      if( strcasecmp( stream->url, url ) == 0 )
         break;
   }

#if USE_THREADS
   pthread_mutex_unlock( &streamMutex );
#endif

   return stream;
}

/*
 * Process an audio stream... Copy data from the input fd, to the output fds,
 * with parameterized buffering.
 */
void audioStreamProcessStream(
   struct anAudioStream*   stream,
   int                     rate
   )
{
   int                     i;
   int                     rc;
   long                    bytesRead;
   long                    bytesSent;
   struct pollfd*          pollFds;
   int                     atEof;
   int                     prevSize;
   int                     anyWritten;
   int                     numBuffers;
   time_t                  current;
   char                    errbuf[ 256 ];

   struct buffer
   {
      char                 buf[ 1024 ];      /* Should fit in a single packet */
      int                  size;
      struct buffer*       next;
   };

   struct buffer*          freeq;
   struct buffer*          outq;
   struct buffer*          outqBottom;
   struct buffer*          b;
   struct buffer*          iobufs;

   addToLog( "Sending stream file '%s'\n", audioStreamGetUrl( stream ) );

   bytesRead = 0;
   bytesSent = 0;
   numBuffers = atoi( getCacheSize() );
   if( numBuffers <= 0 )
      numBuffers = 16;
   if( rate == 0 && numBuffers < 64 )
      numBuffers = 64;
   iobufs = ( struct buffer* ) malloc( numBuffers * sizeof( struct buffer ) );
   outq = NULL;
   freeq = NULL;
   for( rc = 0; rc < numBuffers; rc++ )
   {
      iobufs[ rc ].next = freeq;
      freeq = &iobufs[ rc ];
   }

   atEof = 0;
   prevSize = 0;

   pollFds = NULL;

   for( ; ; )
   {
      if( atEof && outq == NULL )
         break;

      current = time( NULL );

#if USE_THREADS
      pthread_mutex_lock( &streamMutex );
#endif

      if( prevSize != stream->numClients + 1 || !pollFds )
      {
         prevSize = stream->numClients + 1;
         if( pollFds )
            free( pollFds );
         pollFds = ( struct pollfd* ) malloc( prevSize *
            sizeof( *pollFds ) );
      }

      pollFds[ 0 ].fd = audioStreamGetFd( stream );
      pollFds[ 0 ].events = POLLIN | POLLPRI | POLLHUP;
      pollFds[ 0 ].revents = 0;

      for( i = 0; i < stream->numClients; i++ )
      {
         pollFds[ i + 1 ].fd = stream->clients[ i ].fd;
         pollFds[ i + 1 ].events = POLLOUT | POLLHUP;
         pollFds[ i + 1 ].revents = 0;
      }

      if( outq )
      {
         anyWritten = 0;
         bytesSent += outq->size;
         for( i = 1; i <= stream->numClients; i++ )
         {
            if( stream->clients[ i - 1 ].duration > 0 &&
               time( NULL ) >= stream->clients[ i - 1 ].endingTime )
            {
               addToLog( "Client recording duration has expired.\n" );
               stream->clients[ i - 1 ].open = 0;
            }
            rc = poll( &pollFds[ i ], 1, 0 );
            if( rc )
            {
               anyWritten++;
               if( pollFds[ i ].revents & POLLHUP )
               {
                  addToLog( "Client has closed the connection.\n" );
                  stream->clients[ i - 1 ].open = 0;
               }
               else
               {
                  rc = write( pollFds[ i ].fd, outq->buf, outq->size );
                  if( rc != outq->size )
                  {
                     addToLog( "Write error in output stream '%s'; rc=%d, %s\n",
                        audioStreamGetUrl( stream ),
                        rc, strerror_r( errno, errbuf, sizeof errbuf ) );
                     stream->clients[ i - 1 ].open = 0; 
                  }
               }
            }
         }
         if( anyWritten )
         {
            b = outq;
            outq = outq->next;
            b->next = freeq;
            freeq = b;
         }
      }
#if USE_THREADS
      pthread_mutex_unlock( &streamMutex );
#endif

      /* Check for closed output streams... */
      rc = 1;
      while( rc )
      {
         rc = 0;
         for( i = 0; i < stream->numClients; i++ )
         {
            if( !stream->clients[ i ].open )
            {
               audioStreamRemoveClient( stream, stream->clients[ i ].fd );
               rc = 1;
               break;
            }
         }
      }

      if( stream->numClients <= 0 )
      {
         addToLog( "All clients have been removed from the stream.\n" );
         break;
      }

      if( freeq && !atEof )
      {
         rc = poll( &pollFds[ 0 ], 1, 10000 );
         if( rc )
         {
            do
            {
               if( pollFds[ 0 ].revents & POLLHUP )
               {
                  addToLog( "End of stream signaled in '%s'\n",
                     audioStreamGetUrl( stream ) );
                  atEof = 1;
               }
               else if( pollFds[ 0 ].revents & ( POLLIN | POLLPRI ) )
               {
                  rc = read( pollFds[ 0 ].fd, freeq->buf, sizeof freeq->buf );
                  if( rc <= 0 )
                  {
                     if( rc < 0 )
                     {
                        addToLog( "Read error signaled in stream '%s': rc=%d, %s\n",
                           audioStreamGetUrl( stream ), rc,
                           strerror_r( errno, errbuf, sizeof errbuf ) );
                     }
                     addToLog( "Reached end of stream '%s'\n",
                        audioStreamGetUrl( stream ) );
                     atEof = 1;
                  }
                  else
                  {
                     bytesRead += rc;
                     freeq->size = rc;
                     b = freeq;
                     freeq = freeq->next;
                     b->next = NULL;
                     if( outq == NULL )
                        outq = b;
                     else
                        outqBottom->next = b;
                     outqBottom = b;
                  }
               }
               rc = poll( &pollFds[ 0 ], 1, 0 );
            } while( rc && freeq && !atEof );
         }
      }
   }
   if( pollFds )
      free( pollFds );
   free( iobufs );

   addToLog( "Finished MPEG stream, bytes read=%ld, bytes sent=%ld: %s\n", 
      bytesRead, bytesSent,
      strerror_r( errno, errbuf, sizeof errbuf ) );
}
