/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * Dave Clemans, April 2003
 *
 * Copyright (C) 2003  Dave Clemans
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * See the file "COPYING" for more details.
 */

#include "globals.h"

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/poll.h>
#include <netdb.h>

/*
 * A simple packages for storing and retrieving parameters to commands
 */
void storeParameter(
   struct aParameterList* list,
   char*                name,
   char*                value
   )
{
   struct aParameter*   param;

   param = ( struct aParameter* ) malloc( sizeof( *param ) );
   param->name = ( char* ) malloc( strlen( name ) + 1 );
   strcpy( param->name, name );
   param->value = ( char* ) malloc( strlen( value ) + 1 );
   strcpy( param->value, value );
   param->next = NULL;
   
   if( list->top == NULL )
      list->top = param;
   else
      list->bottom->next = param;
   list->bottom = param;
}

char* findParameter(
   struct aParameterList* list,
   char*                name
   )
{
   struct aParameter*   param;
      
   for( param = list->top; param != NULL; param = param->next )
   {
      if( strcasecmp( param->name, name ) == 0 )
         return param->value;
   }
   
   return NULL;
}

void freeParameters(
   struct aParameterList* list
   )
{
   struct aParameter*   curr;
   struct aParameter*   next;
      
   curr = list->top;
   while( curr != NULL )
   {
      next = curr;
      curr = curr->next;
      free( next->name );
      free( next->value );
   }
   
   list->top = NULL;
   list->bottom = NULL;
}

struct aParameterList* copyParameters(
   struct aParameterList* oldList
   )
{
   struct aParameter*   lp;
   struct aParameter*   np;
   struct aParameterList* newList;

   newList = ( struct aParameterList* ) malloc( sizeof( *newList ) );
   newList->top = NULL;
   newList->bottom = NULL;

   for( lp = oldList->top; lp != NULL; lp = lp->next )
   {
      np = ( struct aParameter* ) malloc( sizeof( *np ) );
      np->name = NULL;
      np->value = NULL;
      np->next = NULL;
      if( lp->name )
      {
         np->name = ( char* ) malloc( strlen( lp->name ) + 1 );
         strcpy( np->name, lp->name );
      }
      if( lp->value )
      {
         np->value = ( char* ) malloc( strlen( lp->value ) + 1 );
         strcpy( np->value, lp->value );
      }
      if( newList->top == NULL )
         newList->top = np;
      else
         newList->bottom->next = np;
      newList->bottom = np;
   }

   return newList;
}

/*
 * Build the full pathname of a file
 */
char* fullPathname(
   char*                path,
   char*                name
   )
{
   int                  len;
   char*                fullpath;

   if( strcmp( path, "/" ) == 0 && name[ 0 ] == '/' )
   {
      fullpath = ( char* ) malloc( strlen( name ) + 1 );
      strcpy( fullpath, name );
      return fullpath;
   }
   else
   {
      len = strlen( path ) + strlen( name ) + 2;
      fullpath = ( char* ) malloc( len );
      sprintf( fullpath, "%s/%s", path, name );
   }

   return fullpath;
}

/*
 * Create a new message descriptor
 */
struct aMessage* newMessage( void )
{
   struct aMessage*     newmsg;

   newmsg = ( struct aMessage* ) malloc( sizeof( *newmsg ) );
   newmsg->msg = ( char* ) malloc( 1024 );
   newmsg->length = 1024;
   newmsg->ptr = newmsg->msg;
   newmsg->size = 0;

   return newmsg;
}

/*
 * Free up a message descriptor
 */
void freeMessage(
   struct aMessage*     msg
   )
{
   free( msg->msg );
   free( msg );
}

/*
 * Append info to our current "under construction" message
 */
void appendToMsg(
   struct aMessage*     msg,
   char*                format,
   ...
   )
{
   char*                newmsg;
   va_list              ap;

   va_start( ap, format );

   if( msg->size + 512 > msg->length )
   {
      newmsg = ( char* ) malloc( 2*msg->length );
      if( msg->msg )
      {
         strcpy( newmsg, msg->msg );
         free( msg->msg );
      }
      msg->msg = newmsg;
      msg->length *= 2;
   }

   vsprintf( &msg->msg[ msg->size ], format, ap );
   msg->size += strlen( &msg->msg[msg->size] );

   va_end( ap );
}

typedef void (*sigHandler)(int);

/*
 * Send the current xml tivo message
 */
void sendXmlMessage(
   struct aMessage*     msg,
   int                  fd
   )
{
   char                 buffer[ 256 ];

   write( fd, HTTP_HEADER, HTTP_HEADER_LEN );
   sprintf( buffer, "Content-Length: %d\r\n", msg->size );
   write( fd, buffer, strlen( buffer ) );
   write( fd, XML_HEADER, XML_HEADER_LEN );
   write( fd, "\r\n", 2 );
   write( fd, msg->msg, msg->size );
}

/*
 * Log something so that what we're doing is visible...
 */
void addToLog(
   char*                format,
   ...
   )
{
   va_list              ap;
   time_t               currentTime;
   char*                currentDate;
   char*                cp;
   static FILE*         logfp = NULL;
   static char*         logfile = NULL;

   if( logfile == NULL )
   {
      cp = getLogfile();
      if( cp != NULL ) {
	      logfile = ( char* ) malloc( strlen( cp ) + 1 );
	      strcpy( logfile, cp );
      }
   }
   else
   {
      cp = getLogfile();
      if( cp == NULL )
      {
         free( logfile );
         logfile = NULL;
         if( logfp != NULL )
            fclose( logfp );
         return;
      }
      if( strcmp( logfile, cp ) != 0 )
      {
         if( logfp != NULL )
            fclose( logfp );
         free( logfile );
         logfile = ( char* ) malloc( strlen( cp ) + 1 );
         strcpy( logfile, cp );
         logfp = NULL;
      }
   }

   if( logfp == NULL && logfile != NULL )
   {
      unlink( logfile );
      logfp = fopen( logfile, "w" );
   }

   if( logfp != NULL || getDebugMode() )
   {
      currentTime = time( NULL );
      currentDate = asctime( localtime( &currentTime ) );
      for( cp = currentDate; *cp; cp++ )
      {
         if( *cp == '\n' )
            *cp = '\0';
      }
      for( cp = currentDate; *cp; cp++ )
         /* do nothing */;
      cp -= 5;    /* strip off the year */
      *cp = '\0';

      if( logfp != NULL )
      {
         fprintf( logfp, "\n%s: ", currentDate );

         va_start( ap, format );
         vfprintf( logfp, format, ap );
         va_end( ap );
         fflush( logfp );
      }

      if( getDebugMode() )
         {
         fprintf( stderr, "\n%s: ", currentDate );

         va_start( ap, format );
         vfprintf( stderr, format, ap );
         va_end( ap );
      }
   }
}

/*
 * Return the hostname of this host
 */
char* getHostname( void )
{
   static char          hostname[ 256 ];
   static int           haveName = 0;
   char*                cp;

   if( !haveName )
   {
      gethostname( hostname, sizeof hostname - 1 );
      for( cp = hostname; *cp; cp++ )
         if( *cp == '.' )
            *cp = '\0';
      haveName = 1;
   }

   return hostname;
}

/*
 * Return the "id" of this host (for us, ip address)
 */
long getHostid( void )
{
   struct hostent*      host;
   int                  i;
   static long          hostid;
   static int           haveId = 0;

   if( !haveId )
   {
      hostid = 0;
      host = gethostbyname( getHostname() );
      for( i = 0; host && i < host->h_length; i++ )
      {
         hostid = hostid << 8;
         hostid += host->h_addr_list[ 0 ][ i ];
      }
   }

   return hostid;
}

/*
 * Return the "dotted" numerical address of this host
 */
char* getHostaddress( void )
{
   struct hostent*      host;
   int                  i;
   char*                addr;
   char                 buffer[ 64 ];
   char                 buf[ 16 ];

   buffer[ 0 ] = '\0';
   host = gethostbyname( getHostname() );
   for( i = 0; host && i < host->h_length; i++ )
   {
      sprintf( buf, "%d", host->h_addr_list[ 0 ][ i ] & 0xFF );
      if( buffer[ 0 ] != '\0' )
         strcat( buffer, "." );
      strcat( buffer, buf );
   }

   addr = ( char* ) malloc( strlen( buffer ) + 1 );
   strcpy( addr, buffer );

   return addr;
}

#if USE_THREADS
static pthread_attr_t   threadAttr;
static int              threadInited = 0;
#define NUM_THREADS     32
static pthread_t        thread[ NUM_THREADS ];
static int              currentThread = 0;

/*
 * A simplified wrapper for creating threads
 */
pthread_t newThread(
   void*                (*entryPoint)(void*),
   void*                argument
   )
{
   pthread_t            threadId;
   char                 errbuf[ 256 ];

   if( !threadInited )
   {
      if( pthread_attr_init( &threadAttr ) != 0 )
         addToLog( "pthread_attr_init failed!\n" );
      if( pthread_attr_setdetachstate( &threadAttr, PTHREAD_CREATE_DETACHED ) )
         addToLog( "pthread_attr_setdetachstate failed!\n" );
      threadInited = 1;
   }

   if( pthread_create( &thread[ currentThread ], &threadAttr,
      ( void*(*)(void*) )entryPoint, ( void* ) argument ) != 0 )
   {   
      addToLog( "pthread_create failed!; %s\n",
         strerror_r( errno, errbuf, sizeof errbuf ) );
   }
   threadId = thread[ currentThread ];
   currentThread++;
   if( currentThread >= NUM_THREADS )
      currentThread = 0;

   return threadId;
}
#endif

/*
 * Is this a path to an Internet stream?
 */
int isInternetRef(
   char*                path
   )
{
   int                  i;
   char*                cp;

   if( *path == '/' )
      path++;
   if( strncmp( path, "localhost/", 10 ) == 0 )
      path += 10;
   for( i = 0; ( cp = getInternetEntry( i ) ) != NULL; i++ )
   {
      if( strncasecmp( cp, path, strlen( cp ) ) == 0 &&
         strncmp( &path[ strlen( cp ) ], "://", 3 ) == 0 )
      {
         return 1;
      }
   }
   return 0;
}

/*
 * Is this either a file in our exported structure, or an Internet stream?
 */
char* isValidFileRef(
   char*                path
   )
{
   int                  i;
   char*                cp;
   char*                e;
   char                 savec;
   char*                newpath;
   struct aDirectoryEntry* rootdir;

   if( strcmp( path, "/" ) == 0 )
   {
      newpath = ( char* ) malloc( 2 );
      strcpy( newpath, "/" );
      return newpath;
   }

   newpath = NULL;
   for( rootdir = getRootdir(); rootdir != NULL; rootdir = rootdir->next )
   {
      e = rootdir->name;
      if( strncmp( path, e, strlen( e ) ) == 0 )
      {
         if( path[ strlen( e ) ] == '/' ||
            strlen( path ) == strlen( e ) )
            break;
      }
   }
   if( rootdir == NULL )
   {
      if( path[ 0 ] == '/' )
      {
         for( cp = &path[ 1 ]; *cp && *cp != '/'; cp++ )
            /* do nothing */;
         savec = *cp;
         *cp = '\0';
         for( rootdir = getRootdir(); rootdir != NULL; rootdir = rootdir->next )
         {
            if( strcasecmp( &path[ 1 ], rootdir->description ) == 0 )
            {
               newpath = ( char* ) malloc( strlen( &cp[ 1 ] ) + 2 +
                  strlen( rootdir->name ) );
               sprintf( newpath, "%s/%s", rootdir->name,
                  ( savec == '/' ) ? &cp[ 1 ] : cp );
               path = newpath;
               break;
            }
         }
         *cp = savec;
      }
   }
   if( rootdir == NULL && !isInternetRef( path ) )
      return NULL;

   if( newpath == NULL )
   {
      newpath = ( char* ) malloc( strlen( path ) + 1 );
      strcpy( newpath, path );
   }

   return newpath;
}

/*
 * Gets the pathname of an executable, looking in the PATH environment variable
 */
char* getToolPath(
   char*                   tool
   )
{
   char*                   pathVar;
   char*                   pathTmp;
   char*                   tokptr;
   char*                   dir;
   char*                   path;
   char*                   xpath;

   pathVar = getenv( "PATH" );
   if( pathVar == NULL )
      return NULL;
   pathTmp = ( char* ) malloc( strlen( pathVar ) + 1 );
   strcpy( pathTmp, pathVar );

   dir = strtok_r( pathTmp, ":", &tokptr );
   while( dir != NULL )
   {
      if( strlen( dir ) == 0 )
         path = fullPathname( ".", tool );
      else
         path = fullPathname( dir, tool );
      xpath = decodeString( path );
      if( access( xpath, X_OK) == 0 )
      {
         free( pathTmp );
         free( path );
         return xpath;
      }

      free( path );
      free( xpath );

      dir = strtok_r( NULL, ":", &tokptr );
   }

   free( pathTmp );
   return NULL;
}

/*
 * Gets the pathname of a library, looking in the LD_LIBRARY_PATH environment variable
 */
char* getLibPath(
   char*                   lib
   )
{
   char*                   pathVar;
   char*                   pathTmp;
   char*                   tokptr;
   char*                   dir;
   char*                   path;

   pathVar = getenv( "LD_LIBRARY_PATH" );
   if( pathVar == NULL )
      return NULL;
   pathTmp = ( char* ) malloc( strlen( pathVar ) + 1 );
   strcpy( pathTmp, pathVar );

   dir = strtok_r( pathTmp, ":", &tokptr );
   while( dir != NULL )
   {
      if( strlen( dir ) == 0 )
         path = fullPathname( ".", lib );
      else
         path = fullPathname( dir, lib );
      if( access( path, R_OK) == 0 )
      {
         free( pathTmp );
         return path;
      }

      free( path );

      dir = strtok_r( NULL, ":", &tokptr );
   }

   free( pathTmp );
   return NULL;
}

/*
 * Get the leaf name from a complete file pathname
 */
char* dirLeafName(
   char*                   path
   )
{
   int                     foundSomething = 0;
   char*                   endP;
   char*                   startP;
   char*                   leaf;

   for( endP = path; *endP; endP++ )
      /* do nothing */;
   if( endP > path )
      endP--;
   if( endP > path && *endP == '/' )
      endP--;

   // Lets see whether we have the special words Artist: or
   // Genre: in the name right after a /... in that case, we know that
   // the file name starts right there as we are in byArtist or byGenre
   // mode. This avoids issues with names that contain /.
   for( startP = endP; startP > path; startP-- ) {
      if ( *startP == '/' && (foundSomething == 12 || foundSomething == 13) ) {
          break;
      }
      if ( *startP == '/' && foundSomething == 0) {
          foundSomething++;
          continue;
      }
      if ( *startP == '.' && foundSomething == 1) {
          foundSomething++;
          continue;
      }
      if ( *startP == '.' && foundSomething == 2) {
          foundSomething++;
          continue;
      }
      if ( *startP == '.' && foundSomething == 3) {
          foundSomething++;
          continue;
      }
      if ( (*startP == 'm' || *startP == 't' || *startP == 'e')
            && foundSomething == 4) {
          foundSomething++;
          continue;
      }
      if ( (*startP == 'u' || *startP == 's' || *startP == 'r')
            && foundSomething == 5) {
          foundSomething++;
          continue;
      }
      if ( (*startP == 'b' || *startP == 'i' || *startP == 'n')
            && foundSomething == 6) {
          foundSomething++;
          continue;
      }
      if ( (*startP == 'l' || *startP == 't' || *startP == 'e')
            && foundSomething == 7) {
          foundSomething++;
          continue;
      }
      if ( (*startP == 'a' || *startP == 'r' || *startP == 'g')
            && foundSomething == 8) {
          foundSomething++;
          continue;
      }
      if ( (*startP == ' ' || *startP == 'a') && foundSomething == 9) {
          foundSomething++;
          continue;
      }
      if ( (*startP == 'y' || *startP == ' ') && foundSomething == 10) {
          foundSomething++;
          continue;
      }
      if ( (*startP == 'B' || *startP == 'y') && foundSomething == 11) {
          foundSomething++;
          continue;
      }
      if ( *startP == 'B' && foundSomething == 12) {
          foundSomething++;
          continue;
      }
      foundSomething = 0;
   }

   if (foundSomething == 0)
   {
      for( startP = endP; startP > path && *startP != '/'; startP-- )
         /* do nothing */;
   }
   else 
   {
       startP += foundSomething;
   }

   if( startP > path && *startP == '/' )
      startP++;
   leaf = ( char* ) malloc( endP - startP + 2 );
   strncpy( leaf, startP, endP - startP + 1 );
   leaf[ endP - startP + 1 ] = '\0';

   return leaf;
}

/*
 * Get the directory portion of a complete file pathname
 */
char* dirBaseName(
   char*                   path
   )
{
   char*                   endP;
   char*                   base;

   for( endP = path; *endP; endP++ )
      /* do nothing */;
   if( endP > path )
      endP--;
   if( endP > path && *endP == '/' )
      endP--;
   while( endP > path && *endP != '/' )
      endP--;
   if( endP > path && *endP == '/' )
      endP--;
   base = ( char* ) malloc( endP - path + 2 );
   if( endP == path && *endP == '/' )
      strcpy( base, "/" );
   else if( endP == path )
      strcpy( base, "." );
   else
   {
      strncpy( base, path, endP - path + 1 );
      base[ endP - path + 1 ] = '\0';
   }

   return base;
}


/*
 *
 */
void doHelperCmd( struct aDirectoryEntry *item )
{
   struct stat             statbuf;
   int                     pid;
   char                    buffer[1024];
   int                     rc;
   char                   *cp;
   int                    status;


   if ( item == NULL )
      return;

   if ( item->helperCmd == NULL )
      return;

   if ( stat( item->helperCmd, &statbuf ) < 0 )
      return;

   if( !S_ISREG( statbuf.st_mode ) && !S_ISLNK( statbuf.st_mode ) )
      return;

   cp = strrchr( item->helperCmd, '/' );
   if ( (cp[0] == 0) || (cp[1] == 0) )
      return;

   cp++;

   pid = fork();
   if( pid < 0 )
   {
      addToLog( "Unable to fork for helper func %s : %s\n", item->helperCmd,
         strerror_r( errno, buffer, sizeof buffer ) );
      return;
   }

   if( pid == 0 )
   {
#if 0
      rc = open( "/dev/null", O_RDWR );
      dup2( rc, 0 );
      dup2( rc, 1 );
      dup2( rc, 2 );
      close( rc );
#endif
      addToLog( "Execute helper %s for %s ", item->helperCmd, item->name );

      execlp( item->helperCmd, cp, NULL );
   }

   wait( &status );
   addToLog( "Helper %s returned %d", item->name, status );

}

ssize_t writen( int fd, const void *vptr, size_t n )
{
   size_t nleft = n;
   ssize_t nwritten;
   const char *ptr = vptr;

   while( nleft > 0 )
   {
      if( (nwritten = write( fd, ptr, nleft)) <= 0 )
      {
	 if( errno == EINTR )
	    nwritten = 0;
	 else
	    return (-1);
      }
      nleft -= nwritten;
      ptr += nwritten;
   }
 
   return( n );
}

ssize_t readn( int fd, void *vptr, size_t n )
{
   size_t nleft = n;
   ssize_t nread;
   char *ptr = vptr;

   while( nleft > 0 )
   {
      if( (nread = read( fd, ptr, nleft)) < 0 )
      {
	 if( errno == EINTR )
	    nread = 0;
	 else
	    return (-1);
      }
      else if ( nread == 0 )
	 break;

      nleft -= nread;
      ptr += nread;
   }
   
   return( n-nleft );
}

/*
 * Pause for a short amount of time.  The time is given as the number of
 * seconds, and the number of nanoseconds.
 */
void pauseFor(
   int                  seconds,
   int                  nanoseconds
   )
{
#ifdef USE_USLEEP
   unsigned long        microseconds;

   microseconds = ( seconds * 1000000 ) + ( nanoseconds / 1000 );
   usleep( microseconds );
#else
   struct timespec      delay;

   delay.tv_sec = seconds;
   delay.tv_nsec = nanoseconds;
   nanosleep( &delay, NULL );
#endif
}


/*
 * The bodies of the escape and unescape functions are directly
 * copied from the curl library. They fall under the following copyright:
 *    Copyright (C) 1998 - 2004, Daniel Stenberg, <daniel@haxx.se>, et al.
 *
 * The old implementation of those functions was not fully functional. It
 * resulted in & signs being produced as the result of the encoding. Given
 * that the curl folks have been tweaking this code for years, this seems
 * to be the prident course of action. At the same time, we don't want
 * to introduce a dependency on curl, as this app may run on systems
 * without curl.
 */

#define ishex(in) ((in >= 'a' && in <= 'f') || \
                   (in >= 'A' && in <= 'F') || \
                   (in >= '0' && in <= '9'))

/*
 * Generate an encoded version of a string that escapes special characters
 * Typically this string will be sent as part of a xml or http string
 * Each call will overwrite the previous result
 */
char* encodeString(
   char*                string
   )
{
  size_t alloc = strlen(string)+1;
  char *ns;
  char *testing_ptr = NULL;
  unsigned char in;
  size_t newlen = alloc;
  int strindex=0;
  size_t length;

  ns = malloc(alloc);
  if(!ns)
    return NULL;

  length = alloc-1;
  while(length--) {
    in = *string;
    if(!(in >= 'a' && in <= 'z') &&
       !(in >= 'A' && in <= 'Z') &&
       !(in >= '0' && in <= '9')) {
      /* encode it */
      newlen += 2; /* the size grows with two, since this'll become a %XX */
      if(newlen > alloc) {
        alloc *= 2;
        testing_ptr = realloc(ns, alloc);
        if(!testing_ptr) {
          free( ns );
          return NULL;
        }
        else {
          ns = testing_ptr;
        }
      }
      snprintf(&ns[strindex], 4, "%%%02X", in);

      strindex+=3;
    }
    else {
      /* just copy this */
      ns[strindex++]=in;
    }
    string++;
  }
  ns[strindex]=0; /* terminate it */
  return ns;
}

/*
 * Generate an decoded version of a string that escapes special characters
 * Typically this string will be sent as part of a xml or http string
 * Each call will overwrite the previous result
 */
char* decodeString(
   char*                string
   )
{
  if( !string )
    return NULL;

  int alloc = (int)strlen(string)+1;
  char *ns = malloc(alloc);
  unsigned char in;
  int strindex=0;
  long hex;

  if( !ns )
    return NULL;

  while(--alloc > 0) {
    in = *string;
    if(('%' == in) && ishex(string[1]) && ishex(string[2])) {
      /* this is two hexadecimal digits following a '%' */
      char hexstr[3];
      char *ptr;
      hexstr[0] = string[1];
      hexstr[1] = string[2];
      hexstr[2] = 0;

      hex = strtol(hexstr, &ptr, 16);

      in = (unsigned char)hex; /* this long is never bigger than 255 anyway */
      string+=2;
      alloc-=2;
    }

    ns[strindex++] = in;
    string++;
  }
  ns[strindex]=0; /* terminate it */
  return ns;
}

