/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * Dave Clemans, April 2003
 *
 * Copyright (C) 2003  Dave Clemans
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * See the file "COPYING" for more details.
 */

#include "globals.h"

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#define __USE_GNU 1
#include <search.h>

/*
 * Returns the known virtual directories as a directory list
 */
struct aDirectoryEntry* readVirtualDirectories(
   struct aDirectory*      dirp
   )
{
   struct aDirectoryEntry* topItem;
   struct aDirectoryEntry* bottom;
   struct aDirectoryEntry* item;

   topItem = NULL;

   item = newDirectoryEntry();
   item->dirp = dirp;
   item->name = strdup( VIRTUAL_BYALBUM_NAME );
   item->title = strdup( VIRTUAL_BYALBUM_NAME );
   item->modified = 0;
   item->created = 0;
   item->fileType = FILE_FOLDER;
   item->next = NULL;

   topItem = item;
   bottom = item;

   item = newDirectoryEntry();
   item->dirp = dirp;
   item->name = strdup( VIRTUAL_BYARTIST_NAME );
   item->title = strdup( VIRTUAL_BYARTIST_NAME );
   item->modified = 0;
   item->created = 0;
   item->fileType = FILE_FOLDER;
   item->next = NULL;

   bottom->next = item;
   bottom = item;

   item = newDirectoryEntry();
   item->dirp = dirp;
   item->name = strdup( VIRTUAL_BYGENRE_NAME );
   item->title = strdup( VIRTUAL_BYGENRE_NAME );
   item->modified = 0;
   item->created = 0;
   item->fileType = FILE_FOLDER;
   item->next = NULL;

   bottom->next = item;
   bottom = item;

   return topItem;
}

/*
 * Returns the contents of a virtual sub-directory as a directory list
 */
struct aDirectoryEntry* readVirtualDirContents(
   char*                   path,
   struct aDirectory*      dirp
   )
{
   struct aDirectoryEntry* topItem;
   struct aDirectoryEntry* bottom;
   struct aDirectoryEntry* item;
   int                     virtType;
   struct aParameterList*  params;
   struct aDirectoryEntry* realDir;
   struct aDirectoryEntry* ip;
   struct aDirectoryEntry* dp;
   struct aDirectoryEntry* cp;
   struct aDirectoryEntry* vp;
   char*                   key;
   struct aDirectory*      subdirp;
   char*                   subpath;

   topItem = NULL;
   virtType = -1;

   if( strncmp( path, getVirtualDirName(), strlen( getVirtualDirName() ) ) == 0 )
   {
      if( strcmp( &path[ strlen( path ) - strlen( VIRTUAL_BYALBUM_NAME ) ],
         VIRTUAL_BYALBUM_NAME ) == 0 )
      {
         virtType = VIRT_BYALBUM;
      }
      else if( strcmp( &path[ strlen( path ) - strlen( VIRTUAL_BYARTIST_NAME ) ],
         VIRTUAL_BYARTIST_NAME ) == 0 )
      {
         virtType = VIRT_BYARTIST;
      }
      else if( strcmp( &path[ strlen( path ) - strlen( VIRTUAL_BYGENRE_NAME ) ],
         VIRTUAL_BYGENRE_NAME ) == 0 )
      {
         virtType = VIRT_BYGENRE;
      }
      else
      {
         addToLog( "Unknown virtual subdirectory type.\n" );
      }
   }
   else
   {
      addToLog( "Unknown virtual directory type.\n" );
   }

   if( virtType == -1 )
      return NULL;

   params = ( struct aParameterList* ) malloc( sizeof( *params ) );
   params->top = NULL;
   params->bottom = NULL;
   storeParameter( params, "Recurse", "Yes" );

   for( vp = getVirtualDirectories(); vp; vp = vp->next )
   {
      realDir = getFilteredDirectory( vp->name, params );
      if( realDir == NULL )
      {
         addToLog( "Failed to get virtual directory: %s.\n", vp->name );
         return NULL;
      }
   
      for( ip = realDir; ip; ip = ip->next )
      {
         if( !IS_MPEG( ip ) )
            continue;
         if( IS_FOLDER( ip ) )
            continue;
         key = NULL;
         switch( virtType )
         {
            case VIRT_BYALBUM:
               key = ip->album;
               break;
            case VIRT_BYARTIST:
               key = ip->artist;
               break;
            case VIRT_BYGENRE:
               key = ip->genre;
               break;
         }
         if( !key )
            continue;
         if( strlen( key ) == 0 )
            key = "Other...";

         for( dp = topItem; dp; dp = dp->next )
         {
            if( strcasecmp( dp->name, key ) == 0 )
               break;
         }
         if( !dp )
         {
            item = newDirectoryEntry();
            item->dirp = dirp;
            item->name = strdup( key );
            item->description = strdup( key );
            item->modified = 0;
            item->created = 0;
            item->fileType = FILE_FOLDER;
            item->next = NULL;
   
            if( !topItem )
               topItem = item;
            else
               bottom->next = item;
            bottom = item;
	 }
          
	 subpath = ( char* ) malloc( strlen( path ) + strlen( key ) + 2 );
	 sprintf( subpath, "%s/%s", path, key );
	 subdirp = getFromDirCache( subpath );
	 if( !subdirp )
	 {
	    subdirp = newDirectory();
	    subdirp->path = strdup( subpath );
	    subdirp->items = NULL;
	    addDirToDirCache( subdirp );
	 }
	 free( subpath );

         cp = copyDirectoryEntry( ip );
         cp->next = subdirp->items;
         subdirp->items = cp;
      }

      freeDirectoryEntries( realDir );
   }
   freeParameters( params );

   return topItem;
}
