/*
 * Linux/C based server for TiVo Home Media Option protocol
 *
 * Based on version 1.5.1 of
 *    "TiVo Connect Automatic Machine; Discovery Protocol Specification"
 * Based on version 1.1.0 of
 *    "TiVo Home Media Option; Music and Photos Server Protocol Specification"
 *
 * XM PCR Radio Protocol information comes from the "OpenXM" package;
 *    http://xmpcr.sourceforge.net
 *
 * Dave Clemans, April 2004
 *
 * Copyright (C) 2004  Dave Clemans
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * See the file "COPYING" for more details.
 */

#include "globals.h"

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <termios.h>

/* Whether to use the XMDAEMON protocol to control the XM PCR radio */
/*#define  USE_XMDAEMON   1*/

#ifdef USE_XMDAEMON
static char             xmpcrPacket[ 32768 ];
#else
static char             xmpcrPacket[ 1024 ];
#endif

#ifdef USE_XMDAEMON
static FILE*            xmpcrStream = NULL;
#else
static int              xmpcrStream = -1;
#endif

static int              xmpcrPoweredUp = 0;

/* Service id of channel tuned to after power on */
static int              xmpcrLastChannel = 1;

/* Current channel tuned to */
static int              xmpcrCurrentChannel = 1;
 
#if USE_THREADS
static pthread_mutex_t  xmpcrMutex = PTHREAD_MUTEX_INITIALIZER;
static int              xmpcrInUse = 0;
static time_t           xmpcrUsed = 0;
#endif

struct xmRadioStation
{
   int                  channel;
   int                  serviceid;
   char*                name;
   char*                category;
   char*                artist;
   char*                title;
   
   struct xmRadioStation* next;
};

struct xmRadioCategory
{
   char*                name;
   
   struct xmRadioStation* stations;
   struct xmRadioStation* stationsLast;
   
   struct xmRadioCategory* next;
   struct xmRadioCategory* last;
};

static struct xmRadioCategory* categories = NULL;
static struct xmRadioCategory* categoriesLast = NULL;

#ifndef USE_XMDAEMON

/*
 * The XM PCR radio is controlled via a serial protocol, sent over a USB
 * port using a FTDI USB serial port adapter.
 *
 * Basic XM PCR radio command structure...
 *
 *  # 5AA500 = START COMMAND STRING
 *  # 0 05 = Command Length
 *  # 1 00 = POWER ON COMMAND
 *  # 2 10 = Channel Label Size (8,10, or 16)
 *  # 3 10 = Channel Category Size (8, 10, or 16)
 *  # 4 24 = Artist & Title Size (8, 10, or 16)
 *  # 5 01 = Radio Type (1 = Power Not Constant; 0 = Hardwired Power)
 *  # EDED = END COMMAND STRING
 *  # Returns 33/40 Bytes; 33 = Not Activated.  40 = Activated.
 *       
 *  # 5AA500 = START COMMAND STRING
 *  # 0 02 = Command Length
 *  # 1 01 = POWER OFF COMMAND
 *  # 5 00 = Power Off Type (1 = Sleep Mode; 0 = Off Mode)
 *  # EDED = END COMMAND STRING     
 *
 *  # 5AA500 = START COMMAND STRING
 *  # 0 01 = Command Length
 *  # 1 31 = RADIO ID COMMAND
 *  # EDED = END COMMAND STRING
 *
 *  # 5AA500 = START COMMAND STRING
 *  # 0 02 = Command Length
 *  # 1 13 = MUTE COMMAND
 *  # 2 01 = Mute On/Off (1 = On; 0 = Off)
 *  # Returns 10 Bytes
 *
 *  # 5AA500 = START COMMAND STRING
 *  # 0 04 = Command Length
 *  # 1 25 = CHANNEL LIST COMMAND
 *  # 2 08 = Selection Method (8: Label Channel Select)
 *  #                         (9: Label Channel Next)
 *  # 3 XX = Channel Number (hex)
 *  # 4 00 = Program Type (?)
 *  # EDED = END COMMAND STRING
 *  # Returns 83 Bytes
 *
 *  # 5AA500 = START COMMAND STRING
 *  # 0 01 = Command Length
 *  # 1 43 = TECH COMMAND
 *  # EDED = END COMMAND STRING
 *  # Returns 32 Bytes
 *
 *  # 5AA500 = START COMMAND STRING
 *  # 0 06 = Command Length
 *  # 1 10 = CHANNEL CHANGE COMMAND
 *  # 2 02 = Selection Method (Select Method?)
 *  # 3 XX = Channel Number (hex)
 *  # 4 00 = Unknown (0: Audio; 1: Data)
 *  # 5 00 = Program Type
 *  # 6 01 = Routing (1: Audio Port)
 *  # EDED = END COMMAND STRING
 *  # Returns 12 Bytes
 *
 *  # 5AA500 = START COMMAND STRING
 *  # 0 04 = Command Length
 *  # 1 25 = CHANNEL LIST COMMAND
 *  # 2 09 = ?
 *  # 3 00 = ?
 *  # 4 00 = ?
 *  # EDED = END COMMAND STRING
 *  # Returns 83 Bytes
 *
 */
static unsigned char xmPowerOn[] =        { 0x5a, 0xa5, 0x00, 0x05, 0x00, 0x10, 0x10, 0x24, 0x01, 0xED, 0xED };
static unsigned char xmPowerOff[] =       { 0x5a, 0xa5, 0x00, 0x02, 0x01, 0x00, 0xED, 0xED };
static unsigned char xmOneChannel[] =     { 0x5a, 0xa5, 0x00, 0x04, 0x25, 0x08, 0xFF, 0x00, 0xED, 0xED };
static unsigned char xmAllChannels[] =    { 0x5a, 0xa5, 0x00, 0x04, 0x25, 0x09, 0x00, 0x00, 0xED, 0xED };
static unsigned char xmCurrChannel[] =    { 0x5a, 0xa5, 0x00, 0x04, 0x25, 0x07, 0x00, 0x00, 0xED, 0xED };
static unsigned char xmChangeChannel[] =  { 0x5a, 0xa5, 0x00, 0x06, 0x10, 0x02, 0xFF, 0x00, 0x00, 0x01, 0xED, 0xED };
static unsigned char xmTechSpecs[] =      { 0x5a, 0xa5, 0x00, 0x01, 0x43, 0xED, 0xED };
static unsigned char xmMuteRadio[] =      { 0x5a, 0xa5, 0x00, 0x02, 0x13, 0xFF, 0xED, 0xED };

#endif

static int powerupXM( void );
static int powerdownXM( void );
static int changeXMChannel( int );

#if USE_THREADS
static void xmpcrThread( void );
#endif

#ifndef USE_XMDAEMON
/*
 * Get a response from the XM PCR radio to a command.
 */
static int getXMResponse(
   int                  fd
   )
{
   int                  rc;
   int                  len;


   do
   {
      rc = readn( fd, xmpcrPacket, 1 );
      if( rc != 1 )
         return 0;
   } while( ( ( ( int ) xmpcrPacket[ 0 ] ) & 0xff ) != 0x5a );
   rc = readn( fd, &xmpcrPacket[ 1 ], 3 );
   if( rc != 3 )
      return 0;
   len = ( ( ( ( int ) xmpcrPacket[ 2 ] ) << 8 ) & 0xff ) |
      ( ( ( int ) xmpcrPacket[ 3 ] ) & 0xff );
   if( getDebugMode() )
      addToLog( "Get packet from xmpcr, datalen = %d, total len = %d\n",
         len, len + 6 );
   rc = readn( fd, &xmpcrPacket[ 4 ], len + 2 );
   if( rc != len + 2 )
      addToLog( "Unexpected length from xmpcr; rc = %d; wanted %d\n", rc, len + 2 );
   return rc;
}

/*
 * Send a command to the XM PCR radio
 */
static int sendXMCommand(
   int                  fd,
   char*                command,
   int                  size
   )
{
   int                  rc;

   tcflush( fd, TCIOFLUSH );
   rc = writen( fd, command, size );
   if( rc != size )
   {
      addToLog( "Unexpected write length from send cmd; rc = %d; expected = %d\n",
         rc, size );
      return 0;
   }
   tcdrain( fd );

   return rc;
}

/*
 * Dump out a response packet from the XM PCR
 */
#define TOHEX(v) ((v)<10)?((v)+'0'):((v)+'a'-10)

static void printXMPacket(
   int                  len
   )
{
   char*                packet;
   int                  i;
   int                  ctr;

   packet = ( char* ) malloc( 2 * ( len + 7 ) );
   ctr = 0;
   for( i = 0; i < 4; i++ )
   {
      packet[ ctr++ ] = TOHEX( ( xmpcrPacket[ i ] >> 4 ) & 0x0f );
      packet[ ctr++ ] = TOHEX( xmpcrPacket[ i ] & 0x0f );
   }
   for( i = 0; i < len; i++ )
   {
      packet[ ctr++ ] = TOHEX( ( xmpcrPacket[ 4 + i ] >> 4 ) & 0x0f );
      packet[ ctr++ ] = TOHEX( xmpcrPacket[ 4 + i ] & 0x0f );
   }
   for( i = 0; i < 2; i++ )
   {
      packet[ ctr++ ] = TOHEX( ( xmpcrPacket[ 4 + len + i ] >> 4 ) & 0x0f );
      packet[ ctr++ ] = TOHEX( xmpcrPacket[ 4 + len + i ] & 0x0f );
   }
   packet[ ctr ] = '\0';

   addToLog( "Response from XMPCR: %s\n", packet );

   free( packet );
}
#endif

/*
 * Power up the XM PCR radio
 */
static int powerupXM( void )
{
   int                  rc = 0;

#ifdef USE_XMDAEMON
   if( !xmpcrStream )
      return 0;
#else
   if( xmpcrStream < 0 )
      return 0;
#endif

   addToLog( "Powering up XM PCR Satellite Radio...\n" );
#if USE_THREADS
   pthread_mutex_lock( &xmpcrMutex );
#endif

#ifdef USE_XMDAEMON
   fprintf( xmpcrStream, "xmPOW:ON0:XXX\n" );
   fflush( xmpcrStream );

   pauseFor( 1, 0 );   
   fgets( xmpcrPacket, sizeof xmpcrPacket, xmpcrStream );
#else
   rc = sendXMCommand( xmpcrStream, xmPowerOn, sizeof xmPowerOn );
   if( rc != sizeof xmPowerOn )
   {
      addToLog( "XM Power on command failed; rc=%d\n", rc );
      return 0;
   }

   do
   {
      rc = getXMResponse( xmpcrStream );
      if( getDebugMode() )
         printXMPacket( rc );
      if( ( ( ( int ) xmpcrPacket[ 4 ] ) & 0xff ) == 0x81 )
      {
         pauseFor( 2, 0 );
      }
   } while( ( ( ( int ) xmpcrPacket[ 4 ] ) & 0xff ) != 0x80 );
   if( rc == 29 )
   {
      pauseFor( 5, 0 );

      rc = sendXMCommand( xmpcrStream, xmPowerOn, sizeof xmPowerOn );
      if( rc != sizeof xmPowerOn )
      {
         addToLog( "XM Power on command failed; rc=%d\n", rc );
         return 0;
      }
      do
      {
         rc = getXMResponse( xmpcrStream );
         if( getDebugMode() )
            printXMPacket( rc );
      } while( ( ( ( int ) xmpcrPacket[ 4 ] ) & 0xff ) != 0x80 );
      if( rc == 29 )
         addToLog( "XM PCR Radio not activated?\n" );
   }
#endif

   pauseFor( 1, 0 );

   xmpcrPoweredUp = 1;
   xmpcrUsed = time( NULL );

#if USE_THREADS
   pthread_mutex_unlock( &xmpcrMutex );
#endif

   return 1;
}

/*
 * Power down the XM PCR radio
 */
static int powerdownXM( void )
{
   int                  rc = 0;

   addToLog( "Powering down the XM PCR Satellite Radio...\n" );
#if USE_THREADS
   pthread_mutex_lock( &xmpcrMutex );
#endif

#ifdef USE_XMDAEMON
   fprintf( xmpcrStream, "xmPOW:OFF:XXX\n" );
   fflush( xmpcrStream );

   pauseFor( 1, 0 );
   fgets( xmpcrPacket, sizeof xmpcrPacket, xmpcrStream );
#else
   if( xmpcrStream < 0 )
      return 0;
   sendXMCommand( xmpcrStream, xmPowerOff, sizeof xmPowerOff );
   rc = getXMResponse( xmpcrStream );
   if( getDebugMode() )
      printXMPacket( rc );
#endif

   pauseFor( 1, 0 );

   xmpcrPoweredUp = 0;

#if USE_THREADS
   pthread_mutex_unlock( &xmpcrMutex );
#endif

   return 1;
}

/*
 * Change the channel on the XM PCR radio
 */
static int changeXMChannel(
   int                  channel
   )
{
   int                  rc = 0;

#if USE_THREADS
   pthread_mutex_lock( &xmpcrMutex );
#endif

#ifdef USE_XMDAEMON
   sprintf( xmpcrPacket, "xmCHA:NEW:%03d\n", channel );
   fprintf( xmpcrStream, xmpcrPacket );
   fflush( xmpcrStream );

   fgets( xmpcrPacket, sizeof xmpcrPacket, xmpcrStream );
#else
   xmChangeChannel[ 6 ] = channel;
   rc = sendXMCommand( xmpcrStream, xmChangeChannel, sizeof xmChangeChannel );
   if( rc != sizeof xmChangeChannel )
      return 0;
   rc = getXMResponse( xmpcrStream );
   if( getDebugMode() )
      printXMPacket( rc );
#endif

   xmpcrUsed = time( NULL );
   addToLog( "XM PCR radio channel changed to %d\n", channel );

#if USE_THREADS
   pthread_mutex_unlock( &xmpcrMutex );
#endif

   return rc;
}

#ifndef USE_XMDAEMON
/*
 * Open and initialize the serial link to the XM PCR radio
 */
static int openXMRadio(
   char*                serialPort
   )
{
   int                  fd;
   int                  rc;
   struct termios       serialAttr;
   char                 radioid[ 9 ];

   fd = open( serialPort, O_RDWR );
   if( fd < 0 )
      return fd;

   tcgetattr( fd, &serialAttr );
   cfmakeraw( &serialAttr );
   cfsetispeed( &serialAttr, B9600 );
   cfsetospeed( &serialAttr, B9600 );
   serialAttr.c_iflag |= IGNPAR;
   serialAttr.c_cflag &= ~CSIZE;
   serialAttr.c_cflag |= CS8;
   serialAttr.c_cflag &= ~CSTOPB;
   serialAttr.c_cflag &= ~PARENB;
   tcsetattr( fd, TCSANOW, &serialAttr );

   tcflush( fd, TCIOFLUSH );

   xmpcrStream = fd;

   /* Turn on the receiver */
   rc = powerupXM();
   if( !rc )
      return 0;

   strncpy( radioid, &xmpcrPacket[ 23 ], 8 );
   radioid[ 8 ] = '\0';
   addToLog( "XM PCR Radio id = %s\n", radioid );
   xmpcrLastChannel = ( ( int ) xmpcrPacket[ 13 ] ) & 0xff;
   if( xmpcrLastChannel == 0 )
      xmpcrLastChannel = 1;

   return fd;
}
#endif

#ifndef USE_XMDAEMON
/*
 * Strip trailing spaces off a string
 */
static void stripTrailingSpaces(
   char*                string
   )
{
   char*                cp;

   cp = string;
   while( *cp )
      cp++;
   cp--;
   while( cp > string && isspace( ( int ) *cp ) )
   {
      *cp = '\0';
      cp--;
   }
}

/*
 * Pull info fields out of a channel response from the XM PCR
 */
static void parseXMChannelInfo(
   int*                 channel,
   int*                 serviceid,
   char*                station,          /* assumed to be 17 bytes */
   char*                category,         /* assumed to be 17 bytes */
   char*                artist,           /* assumed to be 17 bytes */
   char*                song              /* assumed to be 17 bytes */
   )
{
   station[ 16 ] = '\0';
   strncpy( station, &xmpcrPacket[ 10 ], 16 );
   stripTrailingSpaces( station );
   category[ 16 ] = '\0';
   strncpy( category, &xmpcrPacket[ 28 ], 16 );
   stripTrailingSpaces( category );
   artist[ 16 ] = '\0';
   strncpy( artist, &xmpcrPacket[ 45 ], 16 );
   stripTrailingSpaces( artist );
   song[ 16 ] = '\0';
   strncpy( song, &xmpcrPacket[ 61 ], 16 );
   stripTrailingSpaces( song );
   *serviceid = ( ( int ) xmpcrPacket[ 8 ] ) & 0xff;
   *channel = ( ( int ) xmpcrPacket[ 7 ] ) & 0xff;
}

/*
 * Get info about all XM PCR radio channels
 */
static int getXMRadioChannels(
   int                  fd
   )
{
   int                  rc;
   int                  channels = 0;

   char                 station[ 17 ];
   char                 category[ 17 ];
   char                 artist[ 17 ];
   char                 song[ 17 ];
   int                  channel;
   int                  serviceid;
   struct xmRadioStation* s;
   struct xmRadioCategory *c;

   if( getDebugMode() )
      addToLog( "Trying to find all XM radio channels.\n" );

   channel = 0;
   for( ; ; )
   {
      pauseFor( 0, 10000000 );

      xmAllChannels[ 6 ] = channel;
      rc = sendXMCommand( fd, xmAllChannels, sizeof xmAllChannels );
      if( rc != sizeof xmAllChannels )
      {
         addToLog( "Channel info command failed; rc=%d\n", rc );
         break;
      }
      do
      {
         rc = getXMResponse( fd );
         if( getDebugMode() )
            printXMPacket( rc );
      } while ( ( ( ( int ) xmpcrPacket[ 4 ] ) & 0xff ) != 0xa5 );

      parseXMChannelInfo( &channel, &serviceid, station, category, artist, song );
      if( channel == 1 )
      {
         strcpy( station, "XM Preview" );
         strcpy( category, "XM Preview" );
      }
      if( !*station && category )
      {
         strcpy( station, "<unknown>" );
      }

      if( !channel )
         break;

      addToLog( "Channel %d; Serviceid %d; station=%s, category=%s, artist=%s, song=%s\n",
         channel, serviceid, station, category, artist, song );
      if( *station && *category && channel != 0 )
      {
         s = ( struct xmRadioStation* ) malloc( sizeof( *s ) );
         s->next = NULL;
         s->channel = channel;
         s->serviceid = serviceid;
         s->name = strdup( station );
         s->category = strdup( category );
         s->artist = strdup( artist );
         s->title = strdup( song );

         for( c = categories; c; c = c->next )
         {
            if( strcmp( c->name, s->category ) == 0 )
               break;
         }
         if( !c )
         {
            c = ( struct xmRadioCategory* ) malloc( sizeof( *c ) );
            c->next = NULL;
            c->last = NULL;
            c->name = strdup( s->category );
            c->stations = NULL;
            c->stationsLast = NULL;
            
            if( !categories )
               categories = c;
            else
               categoriesLast->next = c;
            categoriesLast = c;
         }
         
         if( !c->stations )
            c->stations = s;
         else
            c->stationsLast->next = s;
         c->stationsLast = s;

         channels++;
      }
   }

   for( c = categories; c; c = c->next )
   {
      for( s = c->stations; s; s = s->next )
      {
         if( !s->name || strcmp( s->name, "<unknown>" ) == 0 )
         {
            pauseFor( 0, 10000000 );

            free( s->name );
            xmOneChannel[ 6 ] = s->channel;
            rc = sendXMCommand( fd, xmOneChannel, sizeof( xmOneChannel ) );
            if( rc != sizeof xmOneChannel )
            {
               addToLog( "Channel info command failed; rc = %d\n", rc );
               continue;
            }
            do
            {
               rc = getXMResponse( fd );
               if( getDebugMode() )
                  printXMPacket( rc );
            } while( ( ( ( int ) xmpcrPacket[ 4 ] ) & 0xff ) != 0xa5 );
            parseXMChannelInfo( &channel, &serviceid, station, category, artist, song );
            if( *station )
               s->name = strdup( station );
            else
               s->name = strdup( "<unknown>" );
         }
      }
   }

   if( getDebugMode() )
      addToLog( "Found %d XM radio channels.\n", channels );
}
#endif

/*
 * Initialize the XM PCR Radio, and get it's list of channels 
 */
int initXMRadio( void )
{
#ifdef USE_XMDAEMON
   struct sockaddr_in   stream;
   struct hostent*      host;
   int                  xmfd;
   char*                cp;
#endif
   int                  rc;
   struct xmRadioStation* s;
   struct xmRadioCategory* c;
      
   categories = NULL;
   categoriesLast = NULL;

#ifdef USE_XMDAEMON
   host = gethostbyname( "localhost" );
   stream.sin_family = AF_INET;
   memcpy( &stream.sin_addr, host->h_addr, host->h_length );
   stream.sin_port = htons( XMPCR_SOCK );
   xmfd = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP );
   if( xmfd < 0 )
      return 0;
   rc = connect( xmfd, ( struct sockaddr* ) &stream, sizeof( struct sockaddr_in ) );
   if( rc < 0 )
   {
      close( xmfd );
      return 0;
   }
   xmpcrStream = fdopen( xmfd, "r+" );
   if( !xmpcrStream )
   {
      close( xmfd );
      return 0;
   }

   rc = powerupXM();
   if( !rc )
      return 0;
#else
   xmpcrStream = openXMRadio( getXMRadioControl() );
#endif

#ifndef USE_XMDAEMON
   getXMRadioChannels( xmpcrStream );

   if( xmpcrLastChannel >= 1 && xmpcrLastChannel < 256 )
   {
      s = NULL;
      for( c = categories; c; c = c->next )
      {
         for( s = c->stations; s; s = s->next )
         {
            if( s->serviceid == xmpcrLastChannel )
               break;
         }
         if( s )
            break;
      }
      if( s )
      {
         changeXMChannel( s->channel );
      }
   }
#else
   fprintf( xmpcrStream, "xmCHA:GET:XXX\n" );
   fflush( xmpcrStream );

   pauseFor( 7, 0 );
   fgets( xmpcrPacket, sizeof xmpcrPacket, xmpcrStream );
   
   for( cp = xmpcrPacket; *cp; cp++ )
   {
      if( *cp == '|' || *cp == '~' )
      {
         *cp = '\0';
      }
   }
   
   cp = xmpcrPacket;
   while( *cp )
         cp++;
   cp++;

   while( *cp )
   {
      s = ( struct xmRadioStation* ) malloc( sizeof( *s ) );
      s->next = NULL;
      s->channel = atoi( cp );
      while( *cp )
         cp++;
      cp++;
      s->serviceid = atoi( cp );
      while( *cp )
         cp++;
      cp++;      
      s->name = strdup( cp );
      while( *cp )
         cp++;
      cp++;
      s->category = strdup( cp );
      while( *cp )
         cp++;
      cp++;
      s->artist = strdup( cp );
      while( *cp )
         cp++;
      cp++;
      s->title = strdup( cp );
      while( *cp )
         cp++;
      cp++;
      
      for( c = categories; c; c = c->next )
      {
         if( strcmp( c->name, s->category ) == 0 )
            break;
      }
      if( !c )
      {
         c = ( struct xmRadioCategory* ) malloc( sizeof( *c ) );
         c->next = NULL;
         c->last = NULL;
         c->name = strdup( s->category );
         c->stations = NULL;
         c->stationsLast = NULL;
         
         if( !categories )
            categories = c;
         else
            categoriesLast->next = c;
         categoriesLast = c;
      }
      
      if( !c->stations )
         c->stations = s;
      else
         c->stationsLast->next = s;
      c->stationsLast = s;
   }
#endif

   if( !categories )
      return 0;

   if( getDebugMode() )
   {
      for( c = categories; c; c = c->next )
      {
         addToLog( "%s\n", c->name );
         for( s = c->stations; s; s = s->next )
         {
            addToLog( "\t%-4d %-16s %-16s %-16s\n",
               s->channel, s->name, s->artist, s->title );
         }
      }
   }

   return 1;
}

/*
 * Switch the XM PCR radio to a different channel
 */
int changeXMRadioChannel(
   struct aDirectoryEntry* item
   )
{
   char*                cp;
   int                  channel;

   if( !xmpcrPoweredUp )
      powerupXM();

   if( !item )
      return 0;
   cp = item->name;
   while( *cp && *cp != ' ' )
      cp++;
   if( !*cp )
      return 0;
   cp++;
   channel = atoi( cp );

   if( !xmpcrStream )
      return 0;

   changeXMChannel( channel );

   xmpcrCurrentChannel = channel;

   return 1;
}

/*
 * Returns the known XM Satelite Radio categories as a directory list
 */
struct aDirectoryEntry *readXMRadioCategories(
   struct aDirectory*      dirp
   )
{
   struct aDirectoryEntry* topItem;
   struct aDirectoryEntry* bottom;
   struct aDirectoryEntry* item;
   struct xmRadioCategory* c;
      
   topItem = NULL;
   for( c = categories; c; c = c->next )
   {
      item = newDirectoryEntry();
      item->dirp = dirp;
      item->name = ( char* ) malloc( strlen( c->name ) + 1 );
      strcpy( item->name, c->name );
      item->title = ( char* ) malloc( strlen( c->name ) + 1 );
      strcpy( item->title, c->name );
      item->modified = 0;
      item->created = 0;
      item->fileType = FILE_FOLDER | FILE_XMDIR;
      
      if( topItem == NULL )
         topItem = item;
      else
         bottom->next = item;
      bottom = item;
   }
   
   return topItem;
}

/*
 * Returns the known XM Satellite Radio stations in a category as a directory list
 */
struct aDirectoryEntry* readXMRadioStations(
   char*                   path,
   struct aDirectory*      dirp
   )
{
   char*                   cp;
   struct aDirectoryEntry* topItem;
   struct aDirectoryEntry* bottom;
   struct aDirectoryEntry* item;
   struct xmRadioCategory* c;
   struct xmRadioStation*  s;
      
   cp = strrchr( path, '/' );
   if( cp == NULL )
   {
      addToLog( "Can't find category name: %s\n", path );
      return NULL;
   }
   for( c = categories; c; c = c->next )
   {
     addToLog("Checking category %s\n",c->name);
      if( strcmp( c->name, &cp[ 1 ] ) == 0 )
         break;
   }
   if( c == NULL )
   {
      addToLog( "Can't find category in list: %s\n", &cp[ 1 ] );
      return NULL;
   }
      
   topItem = NULL;
   for( s = c->stations; s; s = s->next )
   {
      item = newDirectoryEntry();
      item->dirp = dirp;
      item->name = ( char* ) malloc( 32 );
      sprintf( item->name, "Channel %d", s->channel );
      item->title = strdup( s->name );
      item->modified = 0;
      item->created = 0;
      item->fileType = FILE_STREAM | FILE_XMDIR;
      
      if( topItem == NULL )
         topItem = item;
      else
         bottom->next = item;
      bottom = item;
   }
   
   return topItem;
}

/*
 * Shut down (turn off) our interface to the XM Satellite Radio.
 */
void shutdownXMRadio( void )
{
   powerdownXM();
}

#if USE_THREADS
/*
 * Try to grab control of the xmpcr; if successful return 1; else return 0
 */
static int grabXMRadio( void )
{
   int                  rc = 0;

   pthread_mutex_lock( &xmpcrMutex );
   if( xmpcrInUse )
      rc = 0;
   else
   {
      xmpcrInUse = 1;
      rc = 1;
   }

   pthread_mutex_unlock( &xmpcrMutex );   

   return rc;
}

/*
 * Release control of the xmpcr
 */
static void releaseXMRadio( void )
{
   xmpcrInUse = 0;
}

/*
 * Thread function to power down xm radio when not in use.
 *
 * Eventually may do other things...
 */
static void xmpcrThread( void )
{
   time_t               current;
   sigset_t             threadMask;

   sigemptyset( &threadMask );
   sigaddset( &threadMask, SIGHUP );
   sigaddset( &threadMask, SIGINT );
   sigaddset( &threadMask, SIGQUIT );
   sigaddset( &threadMask, SIGTERM );
   pthread_sigmask( SIG_BLOCK, &threadMask, NULL );

   addToLog( "XM PCR idle thread starting up...\n" );

   while( 1 )
   {
      pauseFor( 60, 0 );

      current = time( NULL );
      if( getDebugMode() )
         addToLog( "In XM PCR idle thread; xmpcrUsed=%ld, current=%ld\n",
            xmpcrUsed, current );

      if( grabXMRadio() )
      {
         if( xmpcrPoweredUp && ( ( xmpcrUsed + 1200 ) < current ) )
         {
            addToLog( "Letting XM PCR Satellite Radio go to sleep...\n" );
            powerdownXM();
         }
         releaseXMRadio();
      }
   }

   addToLog( "XM PCR idle thread exiting...\n" );
}

/*
 * Start up the XM PCR idle thread...
 */
void startupXMThread( void )
{
   int                  rc;

#ifdef USE_XMDAEMON
   if( !xmpcrStream )
      return;
#else
   if( xmpcrStream < 0 )
      return;
#endif

   xmpcrUsed = time( NULL );
   newThread( ( void* (*)( void* ) ) xmpcrThread, NULL );
}
#endif

/*
 * Try to grab the XM radio for a tivo stream...
 */
int grabXMRadioStream( void )
{
   if( !grabXMRadio() )
      return 0;
   xmpcrUsed = time( NULL );

   return 1;
}

/*
 * Release the XM radio for somebody else...
 */
void releaseXMRadioStream( void )
{
   xmpcrUsed = time( NULL );

   releaseXMRadio();
}
